/* 
 * \file        AdaptingImageSize.cpp
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \version     1.0
 * 
 * \SECTION     
 * 
 * It resizes a fingerprint image to the standar dimensions used in every method
 * 
 */


#include "MemUtil.hpp"
#include "ImgUtil.h"

#include "AdaptingImageSize.h"


AdaptingImageSize::AdaptingImageSize(int h, int w){
    hStd = h;
    wStd = w;
    
    hOrig = 0;
    wOrig = 0;
    
    hProp = 0;
    wProp = 0;
    imagen = 0;
}

AdaptingImageSize::AdaptingImageSize(int hStd , int wStd, unsigned char *imagen, int h, int w){
    this->hStd = hStd;
    this->wStd = wStd;
    
    hOrig = 0;
    wOrig = 0;
    
    hProp = 0;
    wProp = 0;
    
    this->imagen = MemUtil::create2DMatrix<unsigned char>(h, w);
    
    for (int i = 0; i < h; i++){
        for (int j = 0; j < w; j++){
            this->imagen[i][j] = imagen[i * w + j];
        }
    }
    
}

AdaptingImageSize::~AdaptingImageSize(){
    
    MemUtil::freeMatrix(&imagen);
}

unsigned char *AdaptingImageSize::imageToVector(unsigned char **img, int h, int w){
    unsigned char *vector =  MemUtil::createVector<unsigned char>(h * w);
    
    for (int i = 0; i < h; i++){
        for (int j = 0; j < w; j++){
            vector[i * w + j] = img[i][j];
        }
    }
    
    return vector;
}

unsigned char **AdaptingImageSize::reSize(unsigned char **image, int h, int w){
    hOrig = h;
    wOrig = w;
    
    unsigned char **imgAux = proportionImage(image, hOrig, wOrig);
    
    unsigned char **imgFinal = imageSizing(imgAux, hProp, wProp);
    
    MemUtil::freeMatrix(&imgAux);
    
    return imgFinal;
}

unsigned char **AdaptingImageSize::proportionImage(unsigned char **img, int h, int w){
//cout << "proportion Image" << endl;
    float x = (float) wStd / (float)hStd;
    
    float y = (float) w / (float)h;
    unsigned char **imgProp;
    

    if (x > y)
        imgProp = fillWidth(img, h, w, x);
    else if (x < y)
        imgProp = fillHeight(img, h, w, x);
    else{
        imgProp = MemUtil::copy2DMatrix(img, h, w);
        hProp = h;
        wProp = w;
    }
    
    return imgProp;
}

unsigned char **AdaptingImageSize::fillWidth(unsigned char **img, int h, int w, float x){
    
    int auxW = x * h;
     
    int halfW = (auxW - w) / 2;
    
    hProp = h;
    wProp = 2 * halfW + w;
    
    unsigned char **newImg = MemUtil::create2DMatrix<unsigned char>(hProp, wProp);
    
    for (int j = 0; j < halfW; j++){
        for (int i = 0; i < h; i++){
            newImg[i][j] = (unsigned char)colorBackground;
        }
    }
    for (int j = halfW; j < halfW + w; j++){
        for (int i = 0; i < h; i++){
            newImg[i][j] = img[i][j - halfW];
        }
    }
    for (int j = halfW + w; j < wProp; j++){
        for (int i = 0; i < h; i++){
            newImg[i][j] = (unsigned char)colorBackground;
        }
    }
        
    return newImg;
}

unsigned char **AdaptingImageSize::fillHeight(unsigned char **img, int h, int w, float x){

    int auxH = w / x;
  
    int halfH = (auxH - h) / 2;

    hProp = 2 * halfH + h;
    wProp = w;

    unsigned char **newImg = MemUtil::create2DMatrix<unsigned char>(hProp, wProp);

    for (int i = 0; i < halfH; i++){
        for (int j = 0; j < w; j++){
            newImg[i][j] = (unsigned char)colorBackground;
        }
    }
    for (int i = halfH; i < halfH + h; i++){
        for (int j = 0; j < w; j++){
            newImg[i][j] = img[i - halfH][j];
        }
    }
    for (int i = halfH + h; i < hProp; i++){
        for (int j = 0; j < w; j++){
            newImg[i][j] = (unsigned char)colorBackground;
        }
    }
        
    return newImg;
}

unsigned char **AdaptingImageSize::imageSizing(unsigned char **img, int h, int w){
//cout << "image Sizing" << endl;      
    float rel = (float) h / (float)hStd;
    float indexI;
    float indexJ;
    unsigned char **imgFin = MemUtil::create2DMatrix<unsigned char>(hStd, wStd);
    
    int *vecinos = MemUtil::createVector<int>(8, -1);
    float *pesos = MemUtil::createVector<float>(4);
    for (int i = 0; i < hStd; i++){
        for (int j = 0; j < wStd; j++){
            indexI = i * rel;
            indexJ = j * rel;
            extractNeighbours(indexI, indexJ, h, w, vecinos);
            extractWeights(indexI, indexJ, vecinos, pesos);
            
            //
            imgFin[i][j] = calculateValue(img, vecinos, pesos); //calculateValue(img, vecinos, pesos);
            //
            
            for (int i = 0; i < 8; i++)
                vecinos[i] = -1;
            
            memset(pesos, 0 , 4 * sizeof(float));
            
        }
    }
            MemUtil::freeMatrix(&vecinos);
            MemUtil::freeMatrix(&pesos);
    return imgFin;
}


void AdaptingImageSize::extractNeighbours(float indexI, float indexJ, int h, int w, int *vecinos){
        
    int indI = (int)indexI;
    int indJ = (int)indexJ;
    
    if ( (indI < 0) || (indJ < 0) || (indI >= h) || (indJ >= w)) {
        //return;
    }
    else{
        vecinos[0] = ((int)indI);
        vecinos[1] = ((int)indJ);

        if ((indI + 1 < h) && (indJ + 1 < w)) {
            vecinos[2] = ((int)indI + 1);
            vecinos[3] = ((int)indJ + 1);
        }
        if (indJ + 1 < w){
            vecinos[4] = ((int)indI);
            vecinos[5] = ((int)indJ + 1);
        }
        if (indI + 1 < h){
            vecinos[6] = ((int)indI + 1);
            vecinos[7] = ((int)indJ);
        }
    }
}

void AdaptingImageSize::extractWeights(float indexI, float indexJ, int *vecinos, float *pesos){
    
    float total = 0;
    
    if (vecinos[0] == -1) {
        pesos[0] = 0;
        return;
    }
    
    for (int i = 0; i < 4; i++){  
        if (vecinos[i * 2] != -1) {
            float dist = sqrt((indexI - vecinos[i * 2])*(indexI - vecinos[i * 2])+(indexJ - vecinos[i * 2 + 1])*(indexJ - vecinos[i * 2 + 1]));       
            if (dist == 0)
                dist = 1;
            total += 1./dist; 
            pesos[i] = 1. / dist;
        }
    }  
  
    //Normalized weights
    for (int i = 0; i < 4; i++)
        if (vecinos[i * 2] != -1) {
            pesos[i] = pesos[i]/total;
        }
       
    //return pesos;
}

unsigned char AdaptingImageSize::calculateValue(unsigned char **img, int *vecinos, float *pesos){

    float media = 0;
    
    if (vecinos[0] == -1)
        return colorBackground;
    
    for (int i = 0; i < 4; i++){
        if (vecinos[2 * i] != -1)
            media += ((int) img[vecinos[2 * i]][vecinos[2 * i + 1]]) * pesos[i];
    }
    
    unsigned char out = (unsigned char) round(media);
    return out;
}