 
 /* 
 * \file        CovarianceBasedCoreDetector.cpp
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of CovarianceBasedCoreDetector class
 *
 * 
 */


#include "MemUtil.hpp"
#include "MathUtil.h"
#include "ImgUtil.h"

#include "CovarianceBasedCoreDetector.h"


CovarianceBasedCoreDetector::CovarianceBasedCoreDetector() {
}

CovarianceBasedCoreDetector::CovarianceBasedCoreDetector(float **xCosImage, float **ySinImage, int h, int w) {
    
    //Copying the dimensions
    this->orH=h;
    this->orW=w;
    
    //Declaring the vars
    xCos = MemUtil::copy2DMatrix<float>(xCosImage, orH,orW);
    ySin = MemUtil::copy2DMatrix<float>(ySinImage, orH,orW);
    
    this->segmentation = 0;
    this->maxEigValMap = 0;
    this->binEigValMap = 0;
    this->labelMap = 0;
    this->regionCentroids = 0;
    this->candidateCore = 0;
}

CovarianceBasedCoreDetector::CovarianceBasedCoreDetector(float **xCosImage, float **ySinImage, bool **segmentation, int h, int w) {
    
    //Copying the dimensions
    this->orH=h;
    this->orW=w;
    
    //Declaring the vars
    xCos = MemUtil::copy2DMatrix<float>(xCosImage, orH,orW);
    ySin = MemUtil::copy2DMatrix<float>(ySinImage, orH,orW);
	
    // this->segmentation = MemUtil::copy2DMatrix<bool>(segmentation, orH, orW);
    bool **segAux = ImgUtil::imErode(segmentation, orH, orW);
    this->segmentation = ImgUtil::imErode(segAux, orH, orW);    
    MemUtil::freeMatrix(&segAux);        
        
    if (ImgUtil::DEBUG) {
        string fileNameGray = "segmentedCovariance";
        ImgUtil::writeGrayScaleImage(this->segmentation, orH, orW, fileNameGray);
    }
    
    this->maxEigValMap = 0;
    this->binEigValMap = 0;
    this->labelMap = 0;
    this->regionCentroids = 0;
    this->candidateCore = 0;
}

CovarianceBasedCoreDetector::~CovarianceBasedCoreDetector() {
    MemUtil::freeMatrix(&this->xCos);
    MemUtil::freeMatrix(&this->ySin);
    MemUtil::freeMatrix(&this->maxEigValMap);
    MemUtil::freeMatrix(&this->binEigValMap);
    MemUtil::freeMatrix(&this->labelMap);
    MemUtil::freeMatrix(&this->regionCentroids);
    MemUtil::freeMatrix(&this->candidateCore);
}


void CovarianceBasedCoreDetector::computeCore(int marginSize){
    
    int *values = MemUtil::createVector<int>(4, marginSize);
    
    computeCore(values);
    
    MemUtil::freeMatrix(&values);
}

void CovarianceBasedCoreDetector::computeCore(int *fourMarginSizes){
 
    createMaxEigenvalueMap(covarianceWindowHalfsizeShah,fourMarginSizes);
    
    binarizeMaxEigenvalueMap(eigenvaluesThresholdShah);
            
    if (ImgUtil::DEBUG){
        string filename="EigMap";
        ImgUtil::writeGrayScaleImage(this->maxEigValMap,this->orH,this->orW,filename,255);
        
        float **auxImage=MemUtil::create2DMatrix<float>(this->orH,this->orW);
        for (int i=0; i<this->orH; i++){
            for (int j=0; j<this->orW; j++){
                auxImage[i][j]=(float)this->binEigValMap[i][j];
            }
        }
        
        string filename2="BinEigMap";
        ImgUtil::writeGrayScaleImage(auxImage,this->orH,this->orW,filename2,255);
        MemUtil::freeMatrix(&auxImage);
    }
            
    labelCandidatesMap();
    
    calculateRegionCentroids();
            
    
    selectBestCandidateRegion();
    
    if (ImgUtil::DEBUG){
        cout << "CovarianceBasedCoreDetector: The final core is ["<<candidateCore[0]<<","<<candidateCore[1]<<"]"<<endl;
    }
    
}

void CovarianceBasedCoreDetector::calculateRegionCentroids(){
    
    this->regionCentroids=MemUtil::create2DMatrix<float>(maxLabel,2);
    
    for (int lab=1; lab<=this->maxLabel; lab++){
        
        int accI=0;
        int accJ=0;
        int numElements=0;
        
        
        
        for (int i=0; i<this->orH; i++){
            for (int j=0; j<this->orW; j++){
             
                if (this->labelMap[i][j]==lab){
                    accI+=i;
                    accJ+=j;
                    numElements+=1;
                }
                
            }
        }
        
        this->regionCentroids[lab-1][0]=((float) accI)/((float) numElements);
        this->regionCentroids[lab-1][1]=((float) accJ)/((float) numElements);
        
    }
    
}

void CovarianceBasedCoreDetector::selectBestCandidateRegion(){
    
    float centerI= ((float) this->orH)/2.;
    float centerJ= ((float) this->orW)/2.;
    
    float minSqrDist = ((centerI-regionCentroids[0][0])*(centerI-regionCentroids[0][0]))+((centerJ-regionCentroids[0][1])*(centerJ-regionCentroids[0][1]));

    int bestCandidateLabel=1;
    
    for (int lab=2; lab<=this->maxLabel; lab++){
        
        float thisSqrDist=((centerI-regionCentroids[lab-1][0])*(centerI-regionCentroids[lab-1][0]))+((centerJ-regionCentroids[lab-1][1])*(centerJ-regionCentroids[lab-1][1]));
        
        if (thisSqrDist<minSqrDist){
            minSqrDist=thisSqrDist;
            bestCandidateLabel=lab;
        }
    }
    
    candidateCore=MemUtil::createVector<float>(2);
    candidateCore[0]=regionCentroids[bestCandidateLabel-1][0];
    candidateCore[1]=regionCentroids[bestCandidateLabel-1][1];

    
}

void CovarianceBasedCoreDetector::labelCandidatesMap(){
    
    this->labelMap=MemUtil::create2DMatrix<int>(orH,orW);
    
    int label=0;
    
    for (int i = 0; i < this->orH; i++) {
        for (int j = 0; j < this->orW; j++) {
         
            if ((this->binEigValMap[i][j]==1)&& (this->labelMap[i][j]==0)){
                
                label+=1;
                this->labelMap[i][j]=label;
                fillRegion(i,j,label);
            }
        }
    }
    this->maxLabel=label;
    
}

void CovarianceBasedCoreDetector::fillRegion(int posI,int posJ,int label){
    
    for (int i = max(posI-1, 0); i <= min(posI+1, orH-1); i++) {
        for (int j = max(posJ-1, 0); j <= min(posJ+1, orW-1); j++) {
            if ((this->binEigValMap[i][j]==1)&& (this->labelMap[i][j]==0)){
                this->labelMap[i][j]=label;
                fillRegion(i,j,label);   
            }
        }
    }
    return;
    
}

void CovarianceBasedCoreDetector::createMaxEigenvalueMap(int halfWindowSize,int *marginSize){
    
    //first (and last) two rows (and cols) are assesed a 0 value.
    this->maxEigValMap=MemUtil::create2DMatrix<float>(this->orH,this->orW, (float) 0);
    
    //first (and last) two rows (and cols) are assesed a 0 value.
    //The order of marginSize is [N S E W]
    
    int marginNorth=marginSize[0];//To make things clearer to the reader
    int marginSouth=marginSize[1];//To make things clearer to the reader
    int marginEast=marginSize[2];//To make things clearer to the reader
    int marginWest=marginSize[3];//To make things clearer to the reader
    
    for (int i=marginNorth+1; i<this->orH-marginSouth;i++){
        
        int fromI=max(0,i-halfWindowSize);
        int toI=min(this->orH-1,i+halfWindowSize);
        
        for (int j = marginWest+1; j < orW-marginEast; j++){
            if (segmentation[i][j]){
                int fromJ=max(0,j-halfWindowSize);
                int toJ=min(this->orW-1,j+halfWindowSize);

                int numOfElementsInTheWindow=((1+toJ-fromJ)*(1+toI-fromI));

                float** localVectorList=MemUtil::create2DMatrix<float>(numOfElementsInTheWindow,2);

                int vectorCounter=0;
                for (int u=fromI;u<=toI;u++){
                    for (int v=fromJ;v<=toJ;v++){

                        localVectorList[vectorCounter][0]=this->xCos[u][v];
                        localVectorList[vectorCounter][1]=this->ySin[u][v];
                        vectorCounter++;

                    }
                }

                float **covMatrix=MathUtil::makeCovariancesMatrix(localVectorList,numOfElementsInTheWindow,2);

                float *eigenvalues=MathUtil::getEigenvalues(covMatrix,2);

                maxEigValMap[i][j]=max(abs(eigenvalues[0]),abs(eigenvalues[1]));

                //cout << "Eig value "<< maxEigValMap[i][j] << endl;
                MemUtil::freeMatrix(&localVectorList);
                MemUtil::freeMatrix(&covMatrix);
                MemUtil::freeMatrix(&eigenvalues);
                //delete[] eigenvalues;
            }
        }
    }
        
    
}

void CovarianceBasedCoreDetector::binarizeMaxEigenvalueMap(float normalizedPercentage){
    
    //Retrieve the maximum absolute value of an eigenvalue in the map
    float maxAbsEigenval=abs(this->maxEigValMap[0][0]);
    for (int i=0; i<this->orH;i++){
        
        for (int j = 0; j < this->orW; j++){
            
            if (abs(maxEigValMap[i][j])>maxAbsEigenval){
                maxAbsEigenval=abs(maxEigValMap[i][j]);
            }
        }
    }
    
    
    
    //Create a binary boolean map with those positions w/ an eigenvalue which abs-val is over the threshold
    this->binEigValMap=MemUtil::create2DMatrix<int>(this->orH,this->orW);
    float threshold=normalizedPercentage*maxAbsEigenval;
    
    //cout << "Max Eig value "<< maxAbsEigenval <<"threshold, "<<threshold<< endl;
    
    for (int i=0; i<this->orH;i++){
        
        for (int j = 0; j < this->orW; j++){
            
            if (abs(maxEigValMap[i][j])>=threshold){
                this->binEigValMap[i][j]=1;
            }else{
                this->binEigValMap[i][j]=0;
            }
        }
    }
    
    
    
}


float *CovarianceBasedCoreDetector::getCore(){   
    float *core = MemUtil::createVector<float>(2);
    core[0] = candidateCore[0];
    core[1] = candidateCore[1];
    
    return core;
}

