/* 
 * \file        DirectionImagePartitioningCapelli.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of DirectionImagePartitioningCapelli (Capelli99) class
 *
 * 
 */


#include "MemUtil.hpp"
#include "FingerprintImg.h"


#include "DirectionImagePartitioningCapelli.h"
#include "OrientationEnhancementCappelli.h"
#include "ParametersCappelli.h"


DirectionImagePartitioningCapelli::DirectionImagePartitioningCapelli() {
        
    archMask = 0;
    leftLoopMask = 0;
    rightLoopMask = 0;
    tentedArchMask = 0;
    whorlMask = 0;
    
    orH = -1;
    orW = -1;
    ySin = 0;
    xCos = 0;
}

void DirectionImagePartitioningCapelli::initialize(float **xCos, float **ySin, int orH, int orW) {
    
    this->orH = orH;
    this->orW = orW;
    this->xCos = MemUtil::copy2DMatrix<float>(xCos, orH, orW);
    this->ySin = MemUtil::copy2DMatrix<float>(ySin, orH, orW);

    MaskVertex::setOrientationMapSize(orH, orW);
    
    createArchMask();
    createLeftLoopMask();
    createRightLoopMask();
    createTentedArchMask();
    createWhorlMask();
    
}


DirectionImagePartitioningCapelli::~DirectionImagePartitioningCapelli() {
    
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    
    delete archMask;
    delete leftLoopMask;
    delete rightLoopMask;
    delete tentedArchMask;
    delete whorlMask;
    
}


void DirectionImagePartitioningCapelli::createLeftLoopMask() {

    MaskVertex *p1 = new MaskVertex(0, orH - 1, fix); // 28 alto 30 ancho
    MaskVertex *p2 = new MaskVertex(round(p1p2lrtwCappelli * orW) - 1, orH - 1, dependentX);
    MaskVertex *p3 = new MaskVertex(orW / 2, orH - 1, dependentX);
    MaskVertex *p4 = new MaskVertex(p3->getX() + round(p2p3lrtwCappelli * orW), orH - 1, dependentX);
    MaskVertex *p5 = new MaskVertex(orW - 1, orH - 1, fix);
    MaskVertex *p6 = new MaskVertex(orW / 2, (orH - 1) - round(p3p6lrtCappelli * orH), mobile, 
                                        round(p6mobilityXlrCappelli * orW), round(p6mobilityYlrCappelli * orH));
    MaskVertex *p7 = new MaskVertex(0, round(orH * p7p10lrtCappelli) - 1, dependentY);
    MaskVertex *p8 = new MaskVertex(round(p10p11leftCappelli * orW) - 1, p7->getY(), mobile, 
                                        round(p8mobilityXlrCappelli * orW), round(p8mobilityYlrCappelli * orH));
    MaskVertex *p9 = new MaskVertex(orW - 1, p7->getY(), dependentY);

    MaskVertex *p10 = new MaskVertex(0, 0, fix);
    MaskVertex *p11 = new MaskVertex(p8->getX(), 0, dependentX);
    MaskVertex *p12 = new MaskVertex(orW - 1, 0, fix);

    p6->addDependent(p2);
    p6->addDependent(p3);
    p6->addDependent(p4);

    p8->addDependent(p7);
    p8->addDependent(p9);
    p8->addDependent(p11);
    
    p2->setReferencePoint(p6);
    p3->setReferencePoint(p6);
    p4->setReferencePoint(p6);
    
    p7->setReferencePoint(p8);
    p9->setReferencePoint(p8);
    p11->setReferencePoint(p8);

    vector<MaskVertex *> vertices;
    vertices.push_back(p1);
    vertices.push_back(p2);
    vertices.push_back(p3);
    vertices.push_back(p4);
    vertices.push_back(p5);
    vertices.push_back(p6);
    vertices.push_back(p7);
    vertices.push_back(p8);
    vertices.push_back(p9);
    vertices.push_back(p10);
    vertices.push_back(p11);
    vertices.push_back(p12);

    MaskRegion *P1 = new MaskRegion(p1, p2, p6);
    P1->addVertex(p8);
    P1->addVertex(p7);

    MaskRegion *P2 = new MaskRegion(p2, p3, p6);
    MaskRegion *P3 = new MaskRegion(p3, p4, p6);
    MaskRegion *P4 = new MaskRegion(p4, p5, p9);
    P4->addVertex(p8);
    P4->addVertex(p6);

    MaskRegion *P5 = new MaskRegion(p8, p9, p12);
    P5->addVertex(p11);

    MaskRegion *P6 = new MaskRegion(p7, p8, p11);
    P6->addVertex(p10);

    vector<MaskRegion *> regions;
    regions.push_back(P1);
    regions.push_back(P2);
    regions.push_back(P3);
    regions.push_back(P4);
    regions.push_back(P5);
    regions.push_back(P6);


    MaskRegionsRelation *R1 = new MaskRegionsRelation(P1, P6, relationP1P6leftCappelli * PI / 180.);

    vector<MaskRegionsRelation *> relations;
    relations.push_back(R1);

    leftLoopMask = new DynamicMask(vertices, regions, relations);

    if (ImgUtil::DEBUG) {
        leftLoopMask->fillRegions();
        string fileName = "leftLoopMask";
        ImgUtil::writeGrayScaleImage(leftLoopMask->getRegionMap(), orH, orW, fileName, 255. / (float)regions.size());
    }
}

void DirectionImagePartitioningCapelli::createRightLoopMask() {
    MaskVertex *p1 = new MaskVertex(0, orH - 1, fix); // 28 alto 30 ancho
    MaskVertex *p2 = new MaskVertex(round(p1p2lrtwCappelli * orW) - 1, orH - 1, dependentX);
    MaskVertex *p3 = new MaskVertex(orW / 2, orH - 1, dependentX);
    MaskVertex *p4 = new MaskVertex(p3->getX() + round(p2p3lrtwCappelli * orW), orH - 1, dependentX);
    MaskVertex *p5 = new MaskVertex(orW - 1, orH - 1, fix);
    MaskVertex *p6 = new MaskVertex(orW / 2, (orH - 1) - round(p3p6lrtCappelli * orH), mobile, 
                                        round(p6mobilityXlrCappelli * orW), round(p6mobilityYlrCappelli * orH));
    MaskVertex *p7 = new MaskVertex(0, round(orH * p7p10lrtCappelli) - 1, dependentY);
    MaskVertex *p8 = new MaskVertex(round(p10p11rightCappelli * orW) - 1, p7->getY(), mobile, 
                                        round(p8mobilityXlrCappelli * orW), round(p8mobilityYlrCappelli * orH));
    MaskVertex *p9 = new MaskVertex(orW - 1, p7->getY(), dependentY);

    MaskVertex *p10 = new MaskVertex(0, 0, fix);
    MaskVertex *p11 = new MaskVertex(p8->getX(), 0, dependentX);
    MaskVertex *p12 = new MaskVertex(orW - 1, 0, fix);

    p6->addDependent(p2);
    p6->addDependent(p3);
    p6->addDependent(p4);

    p8->addDependent(p7);
    p8->addDependent(p9);
    p8->addDependent(p11);
    
    p2->setReferencePoint(p6);
    p3->setReferencePoint(p6);
    p4->setReferencePoint(p6);
    
    p7->setReferencePoint(p8);
    p9->setReferencePoint(p8);
    p11->setReferencePoint(p8);

    vector<MaskVertex *> vertices;
    vertices.push_back(p1);
    vertices.push_back(p2);
    vertices.push_back(p3);
    vertices.push_back(p4);
    vertices.push_back(p5);
    vertices.push_back(p6);
    vertices.push_back(p7);
    vertices.push_back(p8);
    vertices.push_back(p9);
    vertices.push_back(p10);
    vertices.push_back(p11);
    vertices.push_back(p12);

    
    MaskRegion *P1 = new MaskRegion(p1, p2, p6);
    P1->addVertex(p8);
    P1->addVertex(p7);

    MaskRegion *P2 = new MaskRegion(p2, p3, p6);
    MaskRegion *P3 = new MaskRegion(p3, p4, p6);
    MaskRegion *P4 = new MaskRegion(p4, p5, p9);
    P4->addVertex(p8);
    P4->addVertex(p6);

    MaskRegion *P5 = new MaskRegion(p8, p9, p12);
    P5->addVertex(p11);

    MaskRegion *P6 = new MaskRegion(p7, p8, p11);
    P6->addVertex(p10);

    vector<MaskRegion *> regions;
    regions.push_back(P1);
    regions.push_back(P2);
    regions.push_back(P3);
    regions.push_back(P4);
    regions.push_back(P5);
    regions.push_back(P6);
    
    

    MaskRegionsRelation *R1 = new MaskRegionsRelation(P4, P5, relationP4P5rightCappelli * PI / 180.);

    vector<MaskRegionsRelation *> relations;
    relations.push_back(R1);

    rightLoopMask = new DynamicMask(vertices, regions, relations);

    if (ImgUtil::DEBUG) {
        rightLoopMask->fillRegions();
        string fileName = "rightLoopMask";
        ImgUtil::writeGrayScaleImage(rightLoopMask->getRegionMap(), orH, orW, fileName, 255. / (float)regions.size());
    }
}

void DirectionImagePartitioningCapelli::createArchMask() {
    

    MaskVertex *p1 = new MaskVertex(0, orH - 1, fix); // OK
    MaskVertex *p2 = new MaskVertex(round(p1p2archCappelli * orW) - 1, orH - 1, fix);
    MaskVertex *p3 = new MaskVertex(p2->getX() + round(p2p3archCappelli * orW) + 1, orH - 1, fix);
    MaskVertex *p4 = new MaskVertex(orW - 1, orH - 1, fix); // OK
    MaskVertex *p5 = new MaskVertex(round(lefttop5archCappelli * orW) - 1, (orH - 1) - round(toptop5archCappelli * orH), fix);
    MaskVertex *p6 = new MaskVertex(p5->getX() + round( (1. - 2. * lefttop5archCappelli) * orW)+1, p5->getY(), fix);
    MaskVertex *p7 = new MaskVertex(p5->getX(), p5->getY() - round(p5p7archCappelli * orH), fix);
    MaskVertex *p8 = new MaskVertex(p6->getX(), p7->getY(), fix);
    MaskVertex *p9 = new MaskVertex(0, round(p9p11archCappelli * orH) - 1, fix);

    MaskVertex *p10 = new MaskVertex(orW - 1, p9->getY(), fix);
    MaskVertex *p11 = new MaskVertex(0, 0, fix); // OK
    MaskVertex *p12 = new MaskVertex(orW - 1, 0, fix); // OK

    p2->setReferencePoint(p5);
    p3->setReferencePoint(p6);
    
    p9->setReferencePoint(p7);
    p10->setReferencePoint(p8);
        
    vector<MaskVertex *> vertices;
    vertices.push_back(p1);
    vertices.push_back(p2);
    vertices.push_back(p3);
    vertices.push_back(p4);
    vertices.push_back(p5);
    vertices.push_back(p6);
    vertices.push_back(p7);
    vertices.push_back(p8);
    vertices.push_back(p9);
    vertices.push_back(p10);
    vertices.push_back(p11);
    vertices.push_back(p12);

    MaskRegion *P1 = new MaskRegion(p1, p2, p5);
    P1->addVertex(p7);
    P1->addVertex(p9);

    MaskRegion *P2 = new MaskRegion(p2, p3, p6);
    P2->addVertex(p8);
    P2->addVertex(p10);
    P2->addVertex(p12);
    P2->addVertex(p11);
    P2->addVertex(p9);
    P2->addVertex(p7);
    P2->addVertex(p5);
    
    MaskRegion *P3 = new MaskRegion(p3, p4, p10);
    P3->addVertex(p8);
    P3->addVertex(p6);
    

    vector<MaskRegion *> regions;
    regions.push_back(P1);
    regions.push_back(P2);
    regions.push_back(P3);

    MaskRegionsRelation *R1 = new MaskRegionsRelation(P1, P2, relationP1P2archCappelli * PI / 180.);
    MaskRegionsRelation *R2 = new MaskRegionsRelation(P2, P3, relationP2P3archCappelli * PI / 180.);
       
    vector<MaskRegionsRelation *> relations;
    relations.push_back(R1);
    relations.push_back(R2);

    archMask = new DynamicMask(vertices, regions, relations);

    if (ImgUtil::DEBUG) {
        archMask->fillRegions();
        string fileName = "archMask";
        ImgUtil::writeGrayScaleImage(archMask->getRegionMap(), orH, orW, fileName, 255. / (float)regions.size());
    }
    
}

void DirectionImagePartitioningCapelli::createTentedArchMask() {
    
    MaskVertex *p1 = new MaskVertex(0, orH - 1, fix); 
    MaskVertex *p2 = new MaskVertex(round(p1p2lrtwCappelli * orW) - 1, orH - 1, fix);
    MaskVertex *p3 = new MaskVertex(orW / 2, orH - 1, fix);
    MaskVertex *p4 = new MaskVertex(p3->getX() + round(p2p3lrtwCappelli * orW), orH - 1, fix);
    MaskVertex *p5 = new MaskVertex(orW - 1, orH - 1, fix);
    MaskVertex *p6 = new MaskVertex(orW / 2, (orH - 1) - round(p3p6lrtCappelli * orH), fix);
    MaskVertex *p7 = new MaskVertex(0, round(orH * p7p10lrtCappelli) - 1, dependentY);
    MaskVertex *p8 = new MaskVertex(orW / 2, p7->getY(), mobile, 
                                        round(p8mobilityXtCappelli * orW), round(p8mobilityYtCappelli * orH));
    MaskVertex *p9 = new MaskVertex(orW - 1, p7->getY(), dependentY);

    MaskVertex *p10 = new MaskVertex(0, 0, fix);
    MaskVertex *p11 = new MaskVertex(p8->getX(), 0, dependentX);
    MaskVertex *p12 = new MaskVertex(orW - 1, 0, fix);

    p8->addDependent(p7);
    p8->addDependent(p9);
    p8->addDependent(p11);
    
    p2->setReferencePoint(p6);
    p3->setReferencePoint(p6);
    p4->setReferencePoint(p6);
    
    p7->setReferencePoint(p8);
    p9->setReferencePoint(p8);
    p11->setReferencePoint(p8);

    vector<MaskVertex *> vertices;
    vertices.push_back(p1);
    vertices.push_back(p2);
    vertices.push_back(p3);
    vertices.push_back(p4);
    vertices.push_back(p5);
    vertices.push_back(p6);
    vertices.push_back(p7);
    vertices.push_back(p8);
    vertices.push_back(p9);
    vertices.push_back(p10);
    vertices.push_back(p11);
    vertices.push_back(p12);

    MaskRegion *P1 = new MaskRegion(p1, p2, p6);
    P1->addVertex(p8);
    P1->addVertex(p7);

    MaskRegion *P2 = new MaskRegion(p2, p3, p6);
    MaskRegion *P3 = new MaskRegion(p3, p4, p6);
    MaskRegion *P4 = new MaskRegion(p4, p5, p9);
    P4->addVertex(p8);
    P4->addVertex(p6);

    MaskRegion *P5 = new MaskRegion(p8, p9, p12);
    P5->addVertex(p11);

    MaskRegion *P6 = new MaskRegion(p7, p8, p11);
    P6->addVertex(p10);

    vector<MaskRegion *> regions;
    regions.push_back(P1);
    regions.push_back(P2);
    regions.push_back(P3);
    regions.push_back(P4);
    regions.push_back(P5);
    regions.push_back(P6);


    MaskRegionsRelation *R1 = new MaskRegionsRelation(P1, P4, relationP1P4tentedCappelli * PI / 180.);

    vector<MaskRegionsRelation *> relations;
    relations.push_back(R1);

    tentedArchMask = new DynamicMask(vertices, regions, relations);

    if (ImgUtil::DEBUG) {
        tentedArchMask->fillRegions();
        string fileName = "tentedArchMask";
        ImgUtil::writeGrayScaleImage(tentedArchMask->getRegionMap(), orH, orW, fileName, 255. / (float)regions.size());
    }
    
}

void DirectionImagePartitioningCapelli::createWhorlMask() {
    
    MaskVertex *p1 = new MaskVertex(0, orH - 1, fix); // 28 alto 30 ancho
    MaskVertex *p2 = new MaskVertex(round(p1p2lrtwCappelli * orW) - 1, orH - 1, dependentX);
    MaskVertex *p3 = new MaskVertex(orW / 2, orH - 1, dependentX);
    MaskVertex *p4 = new MaskVertex(p3->getX() + round(p2p3lrtwCappelli * orW), orH - 1, dependentX);
    MaskVertex *p5 = new MaskVertex(orW - 1, orH - 1, fix);
    MaskVertex *p6 = new MaskVertex(orW / 2, (orH - 1) - round(p3p6whorlCappelli * orH), mobile, 
                                        round(p6mobilityXwhorlCappelli * orW), round(p6mobilityYwhorlCappelli * orH));
    MaskVertex *p7 = new MaskVertex(p2->getX(), round(orH * (bottomtop9Cappelli + p7p9whorlCappelli)) - 1, dependentXY);
    MaskVertex *p8 = new MaskVertex(p4->getX(), p7->getY(), mobile, dependentXY);
    
    MaskVertex *p9 = new MaskVertex(p2->getX(), round(orH * bottomtop9Cappelli) - 1, mobile,
                                        round(p9mobilityXwhorlCappelli * orW), round(p9mobilityYwhorlCappelli * orH));
    MaskVertex *p10 = new MaskVertex(p6->getX(), p9->getY(), mobile,
                                        round(p10mobilityXwhorlCappelli * orW), round(p10mobilityYwhorlCappelli * orH));
    MaskVertex *p11 = new MaskVertex(p8->getX(), p9->getY(), mobile,
                                        round(p11mobilityXwhorlCappelli * orW), round(p11mobilityYwhorlCappelli * orH));
    
    MaskVertex *p12 = new MaskVertex(0, round(p12p14whorlCappelli * orH) - 1, dependentY);
    MaskVertex *p13 = new MaskVertex(orW - 1, p12->getY(), dependentY);
    
    MaskVertex *p14 = new MaskVertex(0, 0, fix);
    MaskVertex *p15 = new MaskVertex(p10->getX(), 0, dependentX);
    MaskVertex *p16 = new MaskVertex(orW - 1, 0, fix);

    p6->addDependent(p7);
    p6->addDependent(p8);

    p9->addDependent(p7);
    p9->addDependent(p12);
    
    p10->addDependent(p15);
    
    p11->addDependent(p8);
    p11->addDependent(p13);
    
    p2->setReferencePoint(p6);
    p3->setReferencePoint(p6);
    p4->setReferencePoint(p6);
    
    p12->setReferencePoint(p9);
    p13->setReferencePoint(p11);
    p15->setReferencePoint(p10);

    vector<MaskVertex *> vertices;
    vertices.push_back(p1);
    vertices.push_back(p2);
    vertices.push_back(p3);
    vertices.push_back(p4);
    vertices.push_back(p5);
    vertices.push_back(p6);
    vertices.push_back(p7);
    vertices.push_back(p8);
    vertices.push_back(p9);
    vertices.push_back(p10);
    vertices.push_back(p11);
    vertices.push_back(p12);
    vertices.push_back(p13);
    vertices.push_back(p14);
    vertices.push_back(p15);
    vertices.push_back(p16);

    MaskRegion *P1 = new MaskRegion(p1, p2, p6);
    P1->addVertex(p7);
    P1->addVertex(p9);
    P1->addVertex(p12);

    MaskRegion *P2 = new MaskRegion(p2, p3, p6);
    MaskRegion *P3 = new MaskRegion(p3, p4, p6);
    MaskRegion *P4 = new MaskRegion(p4, p5, p13);
    P4->addVertex(p11);
    P4->addVertex(p8);
    P4->addVertex(p6);

    MaskRegion *P5 = new MaskRegion(p11, p13, p16);
    P5->addVertex(p15);

    MaskRegion *P6 = new MaskRegion(p10, p11, p15);
    MaskRegion *P7 = new MaskRegion(p9, p10, p15);
    MaskRegion *P8 = new MaskRegion(p9, p15, p14);
    P8->addVertex(p12);
    
    MaskRegion *P9 = new MaskRegion(p6, p8, p11);
    P9->addVertex(p10);
    P9->addVertex(p9);
    P9->addVertex(p7);

    vector<MaskRegion *> regions;
    regions.push_back(P1);
    regions.push_back(P2);
    regions.push_back(P3);
    regions.push_back(P4);
    regions.push_back(P5);
    regions.push_back(P6);
    regions.push_back(P7);
    regions.push_back(P8);
    regions.push_back(P9);


    MaskRegionsRelation *R1 = new MaskRegionsRelation(P1, P9, relationP1P9whorlCappelli * PI / 180.);
    MaskRegionsRelation *R2 = new MaskRegionsRelation(P4, P9, relationP4P9whorlCappelli * PI / 180.);

    vector<MaskRegionsRelation *> relations;
    relations.push_back(R1);
    relations.push_back(R2);

    whorlMask = new DynamicMask(vertices, regions, relations);

    if (ImgUtil::DEBUG) {
        whorlMask->fillRegions();
        
        string fileName = "whorlMask";
        ImgUtil::writeGrayScaleImage(whorlMask->getRegionMap(), orH, orW, fileName, 255. / (float)regions.size());
    }
}



float *DirectionImagePartitioningCapelli::fitMasks() {
    
    float *costs = MemUtil::createVector<float>(5);
    
    costs[0] = archMask->optimizeMask(xCos, ySin);    
    costs[1] = leftLoopMask->optimizeMask(xCos, ySin);
    costs[2] = rightLoopMask->optimizeMask(xCos, ySin);
    costs[3] = tentedArchMask->optimizeMask(xCos, ySin);
    costs[4] = whorlMask->optimizeMask(xCos, ySin);
       
    dimFV = 5;
    
    float total = 0;
    for (int i = 0; i < 5; i++)
        total += costs[i];
    
    float min = 1;
    int prediction = -1;
    for (int i = 0; i < 5; i++) {
        costs[i] /= total;
        if (costs[i] < min) {
            min = costs[i];
            prediction = i;
        }
        if (ImgUtil::DEBUG)
            cout << costs[i] << " ";
    }
    
    if (ImgUtil::DEBUG) {
        if (prediction == 0)
            cout << "A" << endl;
        else if (prediction == 1)
            cout << "L" << endl;
        else if (prediction == 2)
            cout << "R" << endl;
        else if (prediction == 3)
            cout << "T" << endl;
        else if (prediction == 4)
            cout << "W" << endl;
        
        archMask->fillRegions();
        leftLoopMask->fillRegions();
        rightLoopMask->fillRegions();
        tentedArchMask->fillRegions();
        whorlMask->fillRegions();

        string fileName = "whorlMaskAfter";
        ImgUtil::writeGrayScaleImage(whorlMask->getRegionMap(), orH, orW, fileName, 255. / (float)whorlMask->getNRegions());
        fileName = "archMaskAfter";
        ImgUtil::writeGrayScaleImage(archMask->getRegionMap(), orH, orW, fileName, 255. / (float)archMask->getNRegions());
        fileName = "leftMaskAfter";
        ImgUtil::writeGrayScaleImage(leftLoopMask->getRegionMap(), orH, orW, fileName, 255. / (float)leftLoopMask->getNRegions());
        fileName = "rightMaskAfter";
        ImgUtil::writeGrayScaleImage(rightLoopMask->getRegionMap(), orH, orW, fileName, 255. / (float)rightLoopMask->getNRegions());
        fileName = "tentedMaskAfter";
        ImgUtil::writeGrayScaleImage(tentedArchMask->getRegionMap(), orH, orW, fileName, 255. / (float)tentedArchMask->getNRegions());
    }
    
    return costs;
    
}


float *DirectionImagePartitioningCapelli::run(string file, PCASYSManager *pcasys) {
    FingerprintImg fi;
    
    if (ImgUtil::fAdjust != 1){    

        fi.setPCASYSManager(pcasys);        fi.readFingerprintImage(file);       
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    fi.enhance();
    fi.computeOrientations();
    if (ImgUtil::DEBUG)
        fi.writeOrientationImage("orMapPCASYS", ORIENTATION);

    OrientationEnhancementCappelli *iec = new OrientationEnhancementCappelli(fi.getYSin(), fi.getXCos(), fi.getOrH(), fi.getOrW());
    iec->enhance(sigmaCappelli, lambdaCappelli);

    initialize(iec->getcos2Theta(), iec->getSin2Theta(), fi.getOrH(), fi.getOrW());
  
    delete iec;
    
    return fitMasks();
}