/* 
 * \file        FingerCode.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of FingerCode class
 *
 * 
 */


#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "FingerprintImg.h"

#include "FingerCode.h"
#include "NewPoincareSingularPoints.h"
#include "CovarianceBasedCoreDetector.h"

#include "ParametersPoincare.h"
#include "ParametersJain.h"
#include "ParametersLeung.h"

FingerCode::FingerCode() {
    fingerMasc = 0;
    blockMasc = 0;
    
    hFi = 0;
    wFi = 0;
   
    gaborSupportX = 0;
    gaborSupportY = 0;
    
    squareSize = 0;
    
    displacedCenter[0] = 0;
    displacedCenter[1] = 0;
   
    halfSquare = 0;    
    iniRowOffset = 0;
    iniColOffset = 0;
    endRowOffset = 0;
    endColOffset = 0;
    
    fcLabels = 0;
    pixelsPerLabel = 0;
    
    h = 0;
    w = 0;
    
    fcLabels = 0;
    
    fingerImg = 0;    
    
    fingerImgPadded = 0;
    
    fc = 0;
    filteredImages = 0;    
}

FingerCode::FingerCode(int squareSize, float **image, int hImg, int wImg, int coreX, int coreY, int centerDisplacement, int gSupX, int gSupY) {
    int i, j;
    fingerMasc = 0;
    blockMasc = 0;
    
    hFi = hImg;
    wFi = wImg;
   
    gaborSupportX = gSupX;
    gaborSupportY = gSupY;
    
    this->squareSize = squareSize;
   
    // 1- Calculating the displaced center
    //
    displacedCenter[0] = coreY + centerDisplacement;
    displacedCenter[1] = coreX;
   
    //
    // 2- Calculating Fingercode dimensions
    //
    halfSquare = (squareSize - 1) / 2;    
    iniRowOffset = (displacedCenter[0] - halfSquare);
    iniColOffset = (displacedCenter[1] - halfSquare);
    endRowOffset = (displacedCenter[0] + halfSquare);
    endColOffset = (displacedCenter[1] + halfSquare);
    
    if (ImgUtil::DEBUG) {
        cout << "initialRowOffset = " << iniRowOffset << endl;
        cout << "initialColumnOffset = " << iniColOffset << endl;
        cout << "endRowOffset = " << endRowOffset << endl;
        cout << "endColumnOffset = " << endColOffset << endl;
    }
    
    // fc = 0;
    fcLabels = 0;
    pixelsPerLabel = 0;
    
    h = endRowOffset - iniRowOffset + 1;
    w = endColOffset - iniColOffset + 1;
    
    // Image where each pixel takes the number of fingercode tile it belongs to (or -1 if its not part of the fingercode region)
    fcLabels = MemUtil::create2DMatrix<int>(h,w,-1);
    
    fingerImg = MemUtil::create2DMatrix<float>(h,w,255);    
    // fingerImg initialization
    for (i = 0; i < h; i++)
        for (j = 0; j < w; j++)
            if ( (i + iniRowOffset >= 0) && (i + iniRowOffset < hFi) && 
                    (j + iniColOffset >= 0) && (j + iniColOffset < wFi))
                fingerImg[i][j] = (int) image[i + iniRowOffset][j + iniColOffset];       

    
    int gSupX1 = (gaborSupportX - 1);
    int gSupY1 = (gaborSupportY - 1);
    int gSup2X = gSupX1 / 2;
    int gSup2Y = gSupY1 / 2;
    
    fingerImgPadded = MemUtil::create2DMatrix<float>(h+gSupX1,w+gSupY1, colorBackground);    
    // fingerImgPadded initialization
    for (i = 0; i < h + gSupX1; i++)
        for (j = 0; j < w + gSupY1; j++)
            if ( (i + iniRowOffset - gSup2X >= 0) && (i + iniRowOffset < hFi) &&
                    (j + iniColOffset - gSup2Y >= 0) && (j + iniColOffset < wFi))
                fingerImgPadded[i][j] = image[i + iniRowOffset - gSup2X][j + iniColOffset - gSup2Y];  

    
    displacedCenter[0] = (h - 1) / 2;
    displacedCenter[1] = (w - 1) / 2;
    
    if (ImgUtil::DEBUG) {
        string fileName3  = "FingerImgFC";
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName3, 1);
        fileName3  = "FingerImgPaddedFC";
        ImgUtil::writeGrayScaleImage(fingerImgPadded, h + gSupX1, w + gSupY1, fileName3, 1);
    }
    
    fc = 0;
    filteredImages = 0;
}

FingerCode::~FingerCode() {

    MemUtil::freeMatrix(&fcLabels);

    MemUtil::freeMatrix(&pixelsPerLabel);

    MemUtil::freeMatrix(&filteredImages);
    
    MemUtil::freeMatrix(&blockMasc);
    
    MemUtil::freeMatrix(&fingerImg);
    
    MemUtil::freeMatrix(&fingerImgPadded);

    MemUtil::freeMatrix(&fingerMasc);

    MemUtil::freeMatrix(&fc);
}

void FingerCode::initializeFC(int squareSize, float **image, int hImg, int wImg, int coreX, int coreY, int centerDisplacement, int gSupX, int gSupY){
    int i, j;
    fingerMasc = 0;
    blockMasc = 0;
    
    hFi = hImg;
    wFi = wImg;
   
    gaborSupportX = gSupX;
    gaborSupportY = gSupY;
    
    this->squareSize = squareSize;
   
    // 1- Calculating the displaced center
    //
    displacedCenter[0] = coreY + centerDisplacement;
    displacedCenter[1] = coreX;
   
    //
    // 2- Calculating Fingercode dimensions
    //
    halfSquare = (squareSize - 1) / 2;    
    iniRowOffset = (displacedCenter[0] - halfSquare);
    iniColOffset = (displacedCenter[1] - halfSquare);
    endRowOffset = (displacedCenter[0] + halfSquare);
    endColOffset = (displacedCenter[1] + halfSquare);
    
    if (ImgUtil::DEBUG) {
        cout << "initialRowOffset = " << iniRowOffset << endl;
        cout << "initialColumnOffset = " << iniColOffset << endl;
        cout << "endRowOffset = " << endRowOffset << endl;
        cout << "endColumnOffset = " << endColOffset << endl;
    }
    
    // fc = 0;
    fcLabels = 0;
    pixelsPerLabel = 0;
    
    h = endRowOffset - iniRowOffset + 1;
    w = endColOffset - iniColOffset + 1;
    
    // Image where each pixel takes the number of fingercode tile it belongs to (or -1 if its not part of the fingercode region)
    fcLabels = MemUtil::create2DMatrix<int>(h,w,-1);
    
    fingerImg = MemUtil::create2DMatrix<float>(h,w,255);    
    // fingerImg initialization
    for (i = 0; i < h; i++)
        for (j = 0; j < w; j++)
            if ( (i + iniRowOffset >= 0) && (i + iniRowOffset < hFi) && 
                    (j + iniColOffset >= 0) && (j + iniColOffset < wFi))
                fingerImg[i][j] = (int) image[i + iniRowOffset][j + iniColOffset];       

    
    int gSupX1 = (gaborSupportX - 1);
    int gSupY1 = (gaborSupportY - 1);
    int gSup2X = gSupX1 / 2;
    int gSup2Y = gSupY1 / 2;
    
    fingerImgPadded = MemUtil::create2DMatrix<float>(h+gSupX1,w+gSupY1, colorBackground);    
    // fingerImgPadded initialization
    for (i = 0; i < h + gSupX1; i++)
        for (j = 0; j < w + gSupY1; j++)
            if ( (i + iniRowOffset - gSup2X >= 0) && (i + iniRowOffset < hFi) &&
                    (j + iniColOffset - gSup2Y >= 0) && (j + iniColOffset < wFi))
                fingerImgPadded[i][j] = image[i + iniRowOffset - gSup2X][j + iniColOffset - gSup2Y];  

    
    displacedCenter[0] = (h - 1) / 2;
    displacedCenter[1] = (w - 1) / 2;
    
    if (ImgUtil::DEBUG) {
        string fileName3  = "FingerImg";
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName3, 1);
        fileName3  = "FingerImgPadded";
        ImgUtil::writeGrayScaleImage(fingerImgPadded, h + gSupX1, w + gSupY1, fileName3, 1);
    }
    
    fc = 0;
    filteredImages = 0;
}

float *FingerCode::run(string file, PCASYSManager *pcasys, bool segmentar){

    FingerprintImg fi;
    bool **mascaraAux;
    
    if (ImgUtil::fAdjust != 1){  
        fi.setPCASYSManager(pcasys);    
        fi.readFingerprintImage(file);     
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    if (segmentar){
        return run(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());   
    }
    else{
        fi.enhance();
        fi.computeOrientations();
        
        mascaraAux = MemUtil::create2DMatrix<bool>(fi.getOrH(),fi.getOrW(),true);
//        // Metemos una máscara a para ajustar las orientaciones
        NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(fi.getSegmentedImageFloat(),  fi.getH(), fi.getW());
        npsp->setOrH(fi.getOrH());
        npsp->setOrW(fi.getOrW());
        npsp->generateSegmentation();
        mascaraAux = MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(),fi.getOrH(), fi.getOrW());
        delete npsp;   
        
        fi.setSegmentationMask(mascaraAux, fi.getOrH(), fi.getOrW()); 
        fi.alterOrientations();         
        fi.computeCore_r92a();
    }


    int squareSize = (numBandsJain * ((int)(bandWidthJain * ImgUtil::fAdjust)) * 2) + ((int)(bandWidthJain * ImgUtil::fAdjust)) + 1;   
       
    int auxGSX = gaborSupportXJain * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYJain * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;
    
    GaborFilter *gb = new GaborFilter(numGaborFiltersJain, gaborFrecuencyJain / ImgUtil::fAdjust, 1, gaborDXJain * ImgUtil::fAdjust, gaborDYJain * ImgUtil::fAdjust, auxGSX, auxGSY);
    
    initializeFC(squareSize, fi.getSegmentedImageFloat(), fi.getH(), fi.getW(), fi.getCoreX(), fi.getCoreY(),((int) centerDisplacementJain * ImgUtil::fAdjust), auxGSX, auxGSY);

    setSegmentationMask(mascaraAux, fi.getOrH());
    MemUtil::freeMatrix(mascaraAux);
    
    if (setLabelsFingerCode(defNumSectorsJain, numBandsJain, ((int)bandWidthJain * ImgUtil::fAdjust)) != -1) { // Fingerprint not rejected
        numSectors = getNumSectors();   
        normalize(V0Jain, M0Jain);  
        filterImage(gb);
        computeFingerCode(0);        // fingercode =  
    }
    else{      
        numSectors = getNumSectors();
        fc = MemUtil::create2DMatrix<float>(numGaborFiltersJain, numBandsJain * numSectors,-1.0);     
    }

    dimFV = numGaborFiltersJain*numSectors;
    
    float *featureVector = MemUtil::createVector<float>(dimFV);
    int cont = 0;
    
    for (int i = 0; i < numGaborFiltersJain; i++) {
        for (int j = 0; j < numSectors; j++) {
            featureVector[cont++] = fc[i][j];
        }
    }
    cout << endl;
 
    delete gb;    
    return featureVector;
}

float *FingerCode::run(float **img, int h, int w){

    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(img, h, w);
    npsp->createOrientationImage((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));

    int nSmooth = npsp->computeCoresAndDeltas(curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare); //(curveWHongFE, smoothWPoincareHongFE, poincareModeHongFE, rulesToBeAppliedHongFE, minNBlocksHongFE); // (curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);
    //(curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);
    int smoothSize = smoothOrientatioWPoincare;
    while( ((npsp->getnCores() > 2) ) ) { 
        nSmooth = npsp->computeCoresAndDeltas(curveWPoincare, smoothSize, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);

        if (nSmooth >= maxSmoothPoincare)
            smoothSize+=2;
        //  cout << "entro jojoo" << endl;
    }


    int centerI, centerJ;

    int *covMargins= MemUtil::createVector<int>(4);
    covMargins[0]=floor((covMarginsCBCDShah * ImgUtil::fAdjust - covMarginsAdjustCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//north
    covMargins[1]=ceil((covMarginsCBCDShah * ImgUtil::fAdjust + covMarginsAdjustCBCDShah * ImgUtil::fAdjust)/(int)(defaultBlockSizePoincare * ImgUtil::fAdjust));//south
    covMargins[2]=floor((covMarginsCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//east
    covMargins[3]=ceil((covMarginsCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//west
    
    if (ImgUtil::isSfinge) {
        covMargins[0] = 0;
        covMargins[1] = 0;
        covMargins[2] = 0;
        covMargins[3] = 0;
    }

    if (npsp->getnCores()==0){
        CovarianceBasedCoreDetector *cbcd = new CovarianceBasedCoreDetector(npsp->getXCos(),npsp->getYSin(),npsp->getSegmentation(), npsp->getOrH(), npsp->getOrW());

        cbcd->computeCore(covMargins);
        float *auxVector = cbcd->getCore();
        centerI = (auxVector[0] + 0.5)* ((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        centerJ = (auxVector[1] + 0.5)* ((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        MemUtil::freeMatrix(&auxVector);

        delete cbcd;
    }
    else if(npsp->getnCores()==1){
        vector<float> auxVector=npsp->getCoreCentroids();
        int *centerCand=  MemUtil::createVector<int>(2);
        centerCand[0]=auxVector[0];
        centerCand[1]=auxVector[1];
        int *finalCenter = npsp->selectCenterInBlock(centerCand,covMargins, h, w,((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        centerI=finalCenter[0];
        centerJ=finalCenter[1];

        MemUtil::freeMatrix(&centerCand);
        MemUtil::freeMatrix(&finalCenter);
    }
    else{
        vector<float> auxVector=npsp->getCoreCentroids();
        float candI1=round((auxVector[0]+0.5)*((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        float candI2=(auxVector[2]+0.5)*((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        int *centerCand=  MemUtil::createVector<int>(2);
        if (candI1<candI2){
            centerCand[0]=auxVector[2];
            centerCand[1]=auxVector[3];
        }else{
            centerCand[0]=auxVector[0];
            centerCand[1]=auxVector[1];    
        }
        int *finalCenter = npsp->selectCenterInBlock(centerCand,covMargins,h,w,((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        centerI=finalCenter[0];
        centerJ=finalCenter[1];

        MemUtil::freeMatrix(&centerCand);
        MemUtil::freeMatrix(&finalCenter);
    }

    int orH = npsp->getOrH();
    int orW = npsp->getOrW();
   
    delete npsp;
    MemUtil::freeMatrix(&covMargins);     

    int squareSize = (numBandsJain * ((int)(bandWidthJain * ImgUtil::fAdjust)) * 2) + ((int)(bandWidthJain * ImgUtil::fAdjust)) + 1; 

    int auxGSX = gaborSupportXJain * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYJain * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;
    
    GaborFilter *gb = new GaborFilter(numGaborFiltersJain, gaborFrecuencyJain / ImgUtil::fAdjust, 1, gaborDXJain * ImgUtil::fAdjust, gaborDYJain * ImgUtil::fAdjust, auxGSX, auxGSY);
    
    initializeFC(squareSize, img, h, w, centerJ, centerI,((int) centerDisplacementJain * ImgUtil::fAdjust), auxGSX, auxGSY);


    //Necesario para que no de error la normalización de Jain
    //(utiliza fingerMasc que se inicializa en setSegmentationMask())
    bool **mascaraAux2 = MemUtil::create2DMatrix<bool>(orH,orW,true);
    setSegmentationMask(mascaraAux2, orH);
    MemUtil::freeMatrix(mascaraAux2);
    
    if (setLabelsFingerCode(defNumSectorsJain, numBandsJain, ((int)bandWidthJain * ImgUtil::fAdjust)) != -1) { // Fingerprint not rejected
        numSectors = getNumSectors();   
        normalize(V0Jain, M0Jain);  
        filterImage(gb);
        computeFingerCode(0);        // fingercode =  
    }
    else{      
        numSectors = getNumSectors();
        fc = MemUtil::create2DMatrix<float>(numGaborFiltersJain, numBandsJain * numSectors,-1.0);     
    }

    dimFV = numGaborFiltersJain*numSectors;
    
    float *featureVector = MemUtil::createVector<float>(dimFV);
    int cont = 0;
    
    for (int i = 0; i < numGaborFiltersJain; i++) {
        for (int j = 0; j < numSectors; j++) {
            featureVector[cont++] = fc[i][j];
        }
    }
    cout << endl;
 
    delete gb;    
    return featureVector;
}

float *FingerCode::runReject(string file, PCASYSManager *pcasys, bool segmentar){

    FingerprintImg fi;
    
    if (ImgUtil::fAdjust != 1){  
        fi.setPCASYSManager(pcasys);    
        fi.readFingerprintImage(file);     
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
        
    if (segmentar){
        return runReject(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());   
    }
    else{
        fi.enhance();
        fi.computeOrientations();
        fi.computeCore_r92a();
    }


    int squareSize = (numBandsJain * ((int)(bandWidthJain * ImgUtil::fAdjust)) * 2) + ((int)(bandWidthJain * ImgUtil::fAdjust)) + 1;   
       
    int auxGSX = gaborSupportXJain * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYJain * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;
    
    GaborFilter *gb = new GaborFilter(numGaborFiltersJain, gaborFrecuencyJain / ImgUtil::fAdjust, 1, gaborDXJain * ImgUtil::fAdjust, gaborDYJain * ImgUtil::fAdjust, auxGSX, auxGSY);
    
    initializeFC(squareSize, fi.getSegmentedImageFloat(), fi.getH(), fi.getW(), fi.getCoreX(), fi.getCoreY(),((int) centerDisplacementJain * ImgUtil::fAdjust), auxGSX, auxGSY);

    bool **mascaraAux = MemUtil::create2DMatrix<bool>(fi.getOrH(),fi.getOrW(),true);
    setSegmentationMask(mascaraAux, fi.getOrH());
    MemUtil::freeMatrix(mascaraAux);
    
    dimFV = 1;
    float *featureVector = MemUtil::createVector<float>(dimFV);
    
    if (setLabelsFingerCode(defNumSectorsJain, numBandsJain, ((int)bandWidthJain * ImgUtil::fAdjust)) != -1) { // Fingerprint not rejected
        featureVector[0] = 1;
    }
    else{      
        featureVector[0] = 0;
    }

    delete gb;
    return featureVector;

}

float *FingerCode::runReject(float **img, int h, int w){
    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(img, h, w);
    npsp->createOrientationImage((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));

    int nSmooth = npsp->computeCoresAndDeltas(curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare); //(curveWHongFE, smoothWPoincareHongFE, poincareModeHongFE, rulesToBeAppliedHongFE, minNBlocksHongFE); // (curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);
    //(curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);
    int smoothSize = smoothOrientatioWPoincare;
    while( ((npsp->getnCores() > 2) ) ) { 
        nSmooth = npsp->computeCoresAndDeltas(curveWPoincare, smoothSize, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);

        if (nSmooth >= maxSmoothPoincare)
            smoothSize+=2;
        //  cout << "entro jojoo" << endl;
    }


    int centerI, centerJ;

    int *covMargins= MemUtil::createVector<int>(4);
    covMargins[0]=floor((covMarginsCBCDShah * ImgUtil::fAdjust - covMarginsAdjustCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//north
    covMargins[1]=ceil((covMarginsCBCDShah * ImgUtil::fAdjust + covMarginsAdjustCBCDShah * ImgUtil::fAdjust)/(int)(defaultBlockSizePoincare * ImgUtil::fAdjust));//south
    covMargins[2]=floor((covMarginsCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//east
    covMargins[3]=ceil((covMarginsCBCDShah * ImgUtil::fAdjust)/((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));//west
    
    if (ImgUtil::isSfinge) {
        covMargins[0] = 0;
        covMargins[1] = 0;
        covMargins[2] = 0;
        covMargins[3] = 0;
    }

    if (npsp->getnCores()==0){
        CovarianceBasedCoreDetector *cbcd = new CovarianceBasedCoreDetector(npsp->getXCos(),npsp->getYSin(),npsp->getSegmentation(), npsp->getOrH(), npsp->getOrW());

        cbcd->computeCore(covMargins);
        float *auxVector = cbcd->getCore();
        centerI = (auxVector[0] + 0.5)* ((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        centerJ = (auxVector[1] + 0.5)* ((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        MemUtil::freeMatrix(&auxVector);

        delete cbcd;
    }
    else if(npsp->getnCores()==1){
        vector<float> auxVector=npsp->getCoreCentroids();
        int *centerCand=  MemUtil::createVector<int>(2);
        centerCand[0]=auxVector[0];
        centerCand[1]=auxVector[1];
        int *finalCenter = npsp->selectCenterInBlock(centerCand,covMargins, h, w,((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        centerI=finalCenter[0];
        centerJ=finalCenter[1];

        MemUtil::freeMatrix(&centerCand);
        MemUtil::freeMatrix(&finalCenter);
    }
    else{
        vector<float> auxVector=npsp->getCoreCentroids();
        float candI1=round((auxVector[0]+0.5)*((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        float candI2=(auxVector[2]+0.5)*((int)(defaultBlockSizePoincare * ImgUtil::fAdjust));
        int *centerCand=  MemUtil::createVector<int>(2);
        if (candI1<candI2){
            centerCand[0]=auxVector[2];
            centerCand[1]=auxVector[3];
        }else{
            centerCand[0]=auxVector[0];
            centerCand[1]=auxVector[1];    
        }
        int *finalCenter = npsp->selectCenterInBlock(centerCand,covMargins,h,w,((int)(defaultBlockSizePoincare * ImgUtil::fAdjust)));
        centerI=finalCenter[0];
        centerJ=finalCenter[1];

        MemUtil::freeMatrix(&centerCand);
        MemUtil::freeMatrix(&finalCenter);
    }

    int orH = npsp->getOrH();
    int orW = npsp->getOrW();
   
    delete npsp;
    MemUtil::freeMatrix(&covMargins);     

    int squareSize = (numBandsJain * ((int)(bandWidthJain * ImgUtil::fAdjust)) * 2) + ((int)(bandWidthJain * ImgUtil::fAdjust)) + 1; 

    int auxGSX = gaborSupportXJain * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYJain * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;
    
    GaborFilter *gb = new GaborFilter(numGaborFiltersJain, gaborFrecuencyJain / ImgUtil::fAdjust, 1, gaborDXJain * ImgUtil::fAdjust, gaborDYJain * ImgUtil::fAdjust, auxGSX, auxGSY);
    
    initializeFC(squareSize, img, h, w, centerJ, centerI,((int) centerDisplacementJain * ImgUtil::fAdjust), auxGSX, auxGSY);


    
    bool **mascaraAux = MemUtil::create2DMatrix<bool>(orH,orW,true);
    setSegmentationMask(mascaraAux, orH);
    MemUtil::freeMatrix(mascaraAux);
    
    dimFV = 1;
    float *featureVector = MemUtil::createVector<float>(dimFV);
    
    if (setLabelsFingerCode(defNumSectorsJain, numBandsJain, ((int)bandWidthJain * ImgUtil::fAdjust)) != -1) { // Fingerprint not rejected
        featureVector[0] = 1;
    }
    else{      
        featureVector[0] = 0;
    }

    delete gb;
    return featureVector;
}

int FingerCode::setLabelsFingerCode(int nSectors, int nBands, int bandWidth) {
    

    int posRow, posCol;
    int dToCenter;
    float fcMinD = -1.0; // Radius of the inner circle of the fingercode
    float fcMaxD = -1.0; // Radius of the external perimeter
    
    float offsetCols=0;
    float offsetRows=0;
    float offsetAngle=0.0;

    int label;
    int sector;
    int bandNumber=0;
    
    fcMinD = (bandWidth / 2); //min distance to the center
    fcMaxD = fcMinD + (nBands) * bandWidth;
    
    numSectors = nBands * nSectors;
    
    if (ImgUtil::DEBUG) {
        cout << "initialRowOffset = " << iniRowOffset << endl;
        cout << "initialColumnOffset = " << iniColOffset << endl;
        cout << "endRowOffset = " << endRowOffset << endl;
        cout << "endColumnOffset = " << endColOffset << endl;
        cout << "hFi = " << hFi << " wFi = " << wFi << endl;
    }
    
    if ((iniRowOffset < 0) || (iniColOffset < 0) || 
            (endRowOffset > hFi) || (endColOffset > wFi)) {
        // NOT VALID FINGERPRINT; REJECT
        cout << "Fingerprint rejected!" << endl;
        return -1;
    }
    
    
    nSectorsInBand = nSectors;
           
    pixelsPerLabel = MemUtil::createVector<int>(numSectors);
    
    iniRowOffset = 0;
    iniColOffset = 0;
    endRowOffset = h;
    endColOffset = w;
    
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
        for (posCol = iniColOffset; posCol < endColOffset; posCol++){

            dToCenter = sqrt((posRow - displacedCenter[0]) * (posRow - displacedCenter[0]) + 
                    (posCol - displacedCenter[1]) * (posCol - displacedCenter[1]));

            if ((dToCenter > fcMinD) && (dToCenter <= fcMaxD)) {

                bandNumber = ceil((dToCenter - fcMinD) / bandWidth);

                offsetRows = posCol - displacedCenter[1];
                offsetCols = posRow - displacedCenter[0];

                offsetAngle = atan(offsetRows / offsetCols); // Incidence angle of the oriented segment Point->Center
                if(offsetCols < 0)
                    offsetAngle = offsetAngle + PI;
                if (offsetAngle < 0)
                    offsetAngle = offsetAngle + (2 * PI);
                if (offsetAngle == 2 * PI)
                    offsetAngle = 0;
                
                sector = floor(offsetAngle * (nSectors / (2 * PI))); // Number of band of the fingercode it belongs to
                label = ((bandNumber - 1) * nSectors + sector); // Number of tile of the fingercode it belongs to
                
                 
                pixelsPerLabel[label]++;
                fcLabels[posRow][posCol] = label;
            }
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName3  = "FCLabelsCircular";
        ImgUtil::writeGrayScaleImage(fcLabels, h, w, fileName3, 255./numSectors);
    }
    
    return 0;
}

int FingerCode::setLabelsSquareFingerCode() {
    
    int i, j;
    int posRow, posCol;
    int dToCenter;
    int iro, ero, ico, eco; 
    
    int blocksPerRow = (numBlocksLeung * 2 + 1);
    int sectorLabels[blocksPerRow * blocksPerRow];
  
    numSectors = 0; // The center is going to be counted
    for (i = 0; i < blocksPerRow ; i++)
        for (j = 0; j < blocksPerRow; j++) {
            dToCenter = sqrt((i - numBlocksLeung) * (i - numBlocksLeung) +
                    (j - numBlocksLeung) * (j - numBlocksLeung));
            if (dToCenter < (numBlocksLeung + 1) && (dToCenter > 0)) {
                sectorLabels[j * blocksPerRow + i] = numSectors;
                numSectors++;
            }
            else
                sectorLabels[j * blocksPerRow + i] = -1;
        }
    if (ImgUtil::DEBUG)
        cout << "numSectors = " << numSectors << endl;
    
    iro = iniRowOffset;
    ero = endRowOffset;
    ico = iniColOffset;
    eco = endColOffset;
    for (i = 0; iniRowOffset < 0; iniRowOffset += ((int)(pixelPerBLockLeung * ImgUtil::fAdjust)), i++)
        for (j = 0; j < blocksPerRow; j++)
            sectorLabels[j * blocksPerRow + i] = -1;
    for (i = 0; iniColOffset < 0; iniColOffset += ((int)(pixelPerBLockLeung * ImgUtil::fAdjust)), i++)
        for (j = 0; j < blocksPerRow; j++)
            sectorLabels[i * blocksPerRow + j] = -1;
    for (i = numBlocksLeung * 2; endRowOffset > hFi; endRowOffset -= ((int)(pixelPerBLockLeung * ImgUtil::fAdjust)), i--)
        for (j = 0; j < blocksPerRow; j++)
            sectorLabels[j * blocksPerRow + i  ] = -1;
    for (i = numBlocksLeung * 2; endColOffset > wFi; endColOffset -= ((int)(pixelPerBLockLeung * ImgUtil::fAdjust)), i--)
        for (j = 0; j < blocksPerRow ; j++)
            sectorLabels[i * blocksPerRow + j] = -1;
    
    iniRowOffset -= iro;
    endRowOffset = h + (endRowOffset - ero);
    iniColOffset -= ico;
    endColOffset = w + (endColOffset - eco);
    if (ImgUtil::DEBUG) {
        cout << "initialRowOffset = " << iniRowOffset << " antes = " << iro << endl;
        cout << "initialColumnOffset = " << iniColOffset << " antes = " << ero  << endl;
        cout << "endRowOffset = " << endRowOffset << " antes = " << ico  << endl;
        cout << "endColumnOffset = " << endColOffset << " antes = " << eco  << endl;
        for (i = 0; i < blocksPerRow; i++) {
            for (j = 0; j < blocksPerRow; j++) {
                    cout << sectorLabels[i * blocksPerRow + j] << " ";
            }
            cout << endl;
        }
    }
        
    
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
        for (posCol = iniColOffset; posCol < endColOffset; posCol++){
            i = (posRow) / ((int)(pixelPerBLockLeung * ImgUtil::fAdjust));
            j = (posCol) / ((int)(pixelPerBLockLeung * ImgUtil::fAdjust));
            fcLabels[posRow][posCol] = sectorLabels[j * blocksPerRow + i];
        }
    }
   
    pixelsPerLabel = MemUtil::createVector<int>(numSectors,((int)(pixelPerBLockLeung * ImgUtil::fAdjust)) * ((int)(pixelPerBLockLeung * ImgUtil::fAdjust)));
    
    if (ImgUtil::DEBUG) {
        string fileName3  = "FCLabelsSquare";
        ImgUtil::writeGrayScaleImage(fcLabels, h, w, fileName3, 1);
    }
    
    return 0;
}

int FingerCode::normalize(int Vo,int Mo) {
    
    int label;
    int posRow, posCol;
    float aux;

    float *sectorsMean = MemUtil::createVector<float>(numSectors + 1);

    float *sectorsVariance = MemUtil::createVector<float>(numSectors + 1);
    int outPixels = 0;
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++) {
        for (posCol = iniColOffset; posCol < endColOffset; posCol++) {
            label = fcLabels[posRow][posCol];
            
            if ((label != -1) /*&& (label != 0)*/) {
                sectorsMean[label] += fingerImg[posRow][posCol] / (float)pixelsPerLabel[label];
            }
            else {
                sectorsMean[numSectors] += fingerImg[posRow][posCol];
                outPixels++;
            }
        }
    }
    sectorsMean[numSectors] /= (float) outPixels;
   
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
        for (posCol = iniColOffset; posCol < endColOffset; posCol++){
            label = fcLabels[posRow][posCol];
            if ((label != -1) /*&& (label != 0)*/){
                sectorsVariance[label] += 
                        (sectorsMean[label] - fingerImg[posRow][posCol]) * 
                        (sectorsMean[label] - fingerImg[posRow][posCol]) / (float)pixelsPerLabel[label];
            }
            else {
                sectorsVariance[numSectors] += 
                        (sectorsMean[numSectors] - fingerImg[posRow][posCol]) * 
                        (sectorsMean[numSectors] - fingerImg[posRow][posCol]) / (float)outPixels;
            }
        }
    }

    float min = 255, max = 0;
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
        for (posCol = iniColOffset; posCol < endColOffset; posCol++){
            label = fcLabels[posRow][posCol];
            if ((label != -1)/* && (label != 0)*/) {             
                if ((sectorsVariance[label] < 0.0001) || (fingerMasc[posRow][posCol] == false)) {
                    // aux = -sectorsMean[numSectors / 2];
                    aux = 0;
                }
                else {
                    aux = sqrt((Vo * ((fingerImg[posRow][posCol] - sectorsMean[label]) * 
                    (fingerImg[posRow][posCol] - sectorsMean[label]))) / sectorsVariance[label]);
                }
                
                aux *= fingerImg[posRow][posCol] > sectorsMean[label] ? 1.0 : -1.0;

                fingerImg[posRow][posCol] = (Mo + aux);

                // For further normalization
                if (fingerImg[posRow][posCol] < min)
                    min = fingerImg[posRow][posCol];
                if (fingerImg[posRow][posCol] > max)
                    max = fingerImg[posRow][posCol]; 

            }
            else {
                if ((sectorsVariance[numSectors] < 0.0001) || (fingerMasc[posRow][posCol] == false)) {
                    // aux = -sectorsMean[numSectors / 2];
                    aux = 0;
                }
                else {
                    aux = sqrt((Vo * ((fingerImg[posRow][posCol] - sectorsMean[numSectors]) * 
                        (fingerImg[posRow][posCol] - sectorsMean[numSectors]))) / sectorsVariance[numSectors]);
                }
                
                
                aux *= fingerImg[posRow][posCol] > sectorsMean[numSectors] ? 1.0 : -1.0;
                
                fingerImg[posRow][posCol] = (Mo + aux);
                
                // For further normalization
                if (fingerImg[posRow][posCol] < min)
                    min = fingerImg[posRow][posCol];
                if (fingerImg[posRow][posCol] > max)
                    max = fingerImg[posRow][posCol];               
                
            }
        }
    }
   
    MemUtil::freeMatrix(&sectorsVariance);
    MemUtil::freeMatrix(&sectorsMean);
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "FingerprintSegmentadaNorm1";
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName2, 1);
    }
    
    for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
        for (posCol = iniColOffset; posCol < endColOffset; posCol++){
            //label = fcLabels[posRow][posCol];
            //if (label != -1) {
                fingerImg[posRow][posCol] = (fingerImg[posRow][posCol] - min) / (max - min) * 255.0;
            //}
        }
    }
    
        if (ImgUtil::DEBUG) {
        string fileName2 = "FingerprintSegmentadaNorm3";
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName2, 1);
    }
    
    //Para rellenar del color de fondo gris aquellas zonas de las huellas realzadas IEH
    //que pertenecen al fondo (Son blancas y entorpecen el cálculo del fingercode porque
    //se detecta como borde)
    ImgUtil::fillBorderColor(fingerImg, h , w, iniRowOffset, endRowOffset, iniColOffset, endColOffset, (Mo - min)/(max - min) * 255.0);
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "FingerprintSegmentadaNorm2";
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName2, 1);
    }
    return 0;
}

int FingerCode::filterImage(GaborFilter *gb) {
    int gSupX = (gaborSupportX - 1);
    int gSupY = (gaborSupportY - 1);
    int gSup2X = gSupX / 2.0;
    int gSup2Y = gSupY / 2.0;
    nFilters = gb->getnFilters();
    
    for (int i = gSup2X; i < h + gSup2X; i++) {
        for (int j = gSup2Y; j < w + gSup2Y; j++) {
                fingerImgPadded[i][j] = fingerImg[i - gSup2X][j - gSup2Y];     
        }
    }
    
    ImgUtil::fillBorderExtending(fingerImgPadded, fingerImg, gSup2X, gSup2Y, h, w);
    if (ImgUtil::DEBUG) {
        string fileName3  = "FingerImgPaddedNorm";
        ImgUtil::writeGrayScaleImage(fingerImgPadded, h + gSupX, w + gSupY, fileName3, 1);
    }
    filteredImages = gb->filterImage(fingerImgPadded,  h + gSupX, w + gSupY);
    
    return 0;
}


float **FingerCode::computeFingerCode(int type) {  // el h y w se omiten ya que se supone que es la misma imagen
    
    if (ImgUtil::DEBUG)
        cout << "FINGERCODE: computeFingerCode()" << endl;
    
    int posRow, posCol;
    int label;
    int idxFilter;
    
    //
    // Creating the filgercodes: Calculating the labels
    //
    /* Fingercode */

    fc = MemUtil::create2DMatrix<float>(nFilters,numSectors);
    
    float **fcMean = MemUtil::create2DMatrix<float>(nFilters,numSectors);  

    for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
        for (posRow = iniRowOffset; posRow < endRowOffset; posRow++) {
            for (posCol = iniColOffset; posCol < endColOffset; posCol++) {
                label = fcLabels[posRow][posCol];
                
                if (label != -1){
                // accumulating for the mean
                    fcMean[idxFilter][label] += filteredImages[idxFilter][posRow][posCol] 
                            / (float)pixelsPerLabel[label];
                }
            }
        }       
    }

    if (type == 0){
        // Creating the fingercodes: Calculating the std per label
        for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
            for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
                for (posCol = iniColOffset; posCol < endColOffset; posCol++){
                    label = fcLabels[posRow][posCol];
                    if(label != -1){
                        //Standard deviation at each tile
                        fc[idxFilter][label] +=
                                (filteredImages[idxFilter][posRow][posCol] - fcMean[idxFilter][label]) * 
                                (filteredImages[idxFilter][posRow][posCol] - fcMean[idxFilter][label]) / (float)pixelsPerLabel[label];
                    }
                }
            }
        }

        for (idxFilter = 0; idxFilter < nFilters; idxFilter++)
            for (label = 0; label < numSectors; label++){
                //cout << fc[idxFilter][label] << " ";
                fc[idxFilter][label] = sqrt(fc[idxFilter][label]);
            }
    }
    else {
        // Creating the fingercodes: Calculating the std per label
        for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
            for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
                for (posCol = iniColOffset; posCol < endColOffset; posCol++){
                    label = fcLabels[posRow][posCol];
                    if(label != -1){
                        //Standard deviation at each tile
                        fc[idxFilter][label] +=
                                abs(filteredImages[idxFilter][posRow][posCol] - fcMean[idxFilter][label]) 
                                                                                / (float)pixelsPerLabel[label];
                    }
                }
            }
        }
        
    }
    if (ImgUtil::DEBUG) {
        cout << "FingerprintImg: Paso 5c.1 terminado" << endl;

        for (idxFilter = 0; idxFilter < nFilters; idxFilter++) {
            for (label = 0; label < numSectors; label++)
                    cout << fc[idxFilter][label] << " ";
            cout << endl;
        }
    }

    if (blockMasc != 0) {
        for (int i = 0; i < numSectors; i++) {
            if (blockMasc[i] == false) {
                for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
                    fc[idxFilter][i] = -1;
                }
            }
        }
    }

    if (ImgUtil::DEBUG) {
        // Creating the fingercodes: fingerImages
        float ***fcImages = MemUtil::create3DMatrix<float>(nFilters,w,h);

        for (idxFilter = 0; idxFilter < nFilters; idxFilter++){
            for (posRow = iniRowOffset; posRow < endRowOffset; posRow++){
                for (posCol = iniColOffset; posCol < endColOffset; posCol++){
                    label = fcLabels[posRow][posCol];
                    if (label >= 0){
                            fcImages[idxFilter][posRow][posCol] = fc[idxFilter][label];
                    }
                }
            }
        }
        
        for (int k = 0; k < nFilters; k++) {
            string fileName = "FingerCode";
            char str[20];
            sprintf(str, "%d", k);
            fileName.append(str);
            ImgUtil::writeGrayScaleImage(fcImages[k], h, w, fileName, 2);
            
            fileName = "GaborFilteredImage";
            fileName.append(str);
            ImgUtil::writeGrayScaleImage(filteredImages[k], h, w, fileName, 1);
        }

        MemUtil::freeMatrix(&fcImages);
    }
    
    MemUtil::freeMatrix(&fcMean);
        
    if (ImgUtil::DEBUG)
        cout << "FINGERCODE: END computeFingerCode()" << endl;
    
    
    return fc;
}

int FingerCode::circleAverage(int filterRadio) {
    
    int x, y;
    int card = 0;   //Filter element's number (equal to 1)
    int posRow,posCol;
    int filterRow,filterCol;
    float sum;
    
    // Define the size of the filter; Initially the filter is squared.
    float **circleFilter = MemUtil::create2DMatrix<float>(filterRadio * 2 + 1,filterRadio * 2 + 1);            
    /* Initialize the filter to 0s & 1s; 0s are used to discard
       the pixels that don't belong to the circle filter */
    for (int i = 0; i < filterRadio * 2 + 1; i++){
        x = filterRadio - i;             // Distance pixel(coordinate X)-center
        for (int j = 0; j < filterRadio * 2 + 1; j++){
            y = filterRadio - j;         // Distance pixel(coordinate Y)-center
            if (sqrt((x * x) + (y * y)) <= filterRadio){
                circleFilter[i][j] = 1.0;
                card++;
            }
        }
    }
    
    //Create a new image bigger than 'image' in 2radio pixels height and width
    float **imgExtend = MemUtil::create2DMatrix<float>(h+2*filterRadio,w+2*filterRadio);
    // imgExtend initialization
    for (int i = 0; i < h + 2*filterRadio; i++) {
        for (int j = 0; j < w + 2*filterRadio; j++) {
            if (!(j < filterRadio || (j > w + filterRadio - 1) || i < filterRadio || (i > h + filterRadio - 1)))
                imgExtend[i][j] = fingerImg[i-filterRadio][j-filterRadio];
        }
    }
  
    ImgUtil::fillBorderExtending(imgExtend, fingerImg, filterRadio, filterRadio, h, w);
    
    // Define the final filtered image size.
    float **imgFiltered = MemUtil::create2DMatrix<float>(h,w);
    // The image is filtered with the circle filter.
    for (posRow = filterRadio; posRow < h + filterRadio; posRow++){
        for (posCol = filterRadio; posCol < w + filterRadio; posCol++){
            sum = 0;
            for (filterRow = 0; filterRow <= (filterRadio * 2); filterRow++){
                for (filterCol = 0; filterCol <= (filterRadio * 2); filterCol++){
                    sum += (imgExtend[posRow - filterRadio + filterRow]
                                     [posCol - filterRadio + filterCol]
                            * circleFilter[filterRow][filterCol]);
                }
            }
            imgFiltered[posRow - filterRadio][posCol - filterRadio] = (sum / (float) card);
        }
    }
    
    if (ImgUtil::DEBUG){
        string fileName2 = "FingerImgExtended";
        string fileName  = "FingerImgJuan";
        string fileName1  = "FingerImgAverageJuan";
        
        ImgUtil::writeGrayScaleImage(imgExtend, h+2*filterRadio, w+2*filterRadio, fileName2, 1);
        ImgUtil::writeGrayScaleImage(fingerImg, h, w, fileName, 1);
        ImgUtil::writeGrayScaleImage(imgFiltered, h, w, fileName1, 1);
    }
    
       
// Free the circle filter
    MemUtil::freeMatrix(&circleFilter);
    MemUtil::freeMatrix(&imgExtend);
    MemUtil::freeMatrix(&fingerImg);
    
    fingerImg = imgFiltered;
    
    return 0;
}


void FingerCode::computeMascPerBlock() {
    
    if (ImgUtil::DEBUG)
        cout << "FINGERCODE: computeMascPerBlock()" << endl;
    
    blockMasc = MemUtil::createVector(numSectors,true);
    // blockMasc initialization
    for (int i = iniRowOffset; i < endRowOffset; i++)
        for (int j = iniColOffset; j < endColOffset; j++){
            int label = fcLabels[i][j];
            if(label != -1)
                if (fingerMasc[i][j] == false)
                    blockMasc[label] = false;            
        }
    
    if (ImgUtil::DEBUG)
        cout << "FINGERCODE: END computeMascPerBlock()" << endl;
}

float **FingerCode::rotateFingerCode(int rotate) {
    
        /* Correct the rotation of the fingerprint */
        // en principio tenemos en fingercode[i] la direcciÃ³n 0, 1, 2, 3... 15
        // si R es 270 = 12 -> 3 / 4 de defNumSectors, no hay rotaciÃ³n! (hacia abajo)
        int R = 3 * nSectorsInBand / 4 - rotate; // positive -> clockwise, negative -> counter-clockwise

        float **fcR = MemUtil::create2DMatrix<float>(nFilters,numSectors);
        // fcR initialization
        for (int i = 0; i < numSectors; i++)
            for (int theta = 0; theta < nFilters; theta++) {
                
                int iprima = (i + nSectorsInBand + R) % nSectorsInBand + (i / nSectorsInBand) * nSectorsInBand; 
                int thetaprima = (theta + nFilters + R) % nFilters;
                
                fcR[theta][i] = fc[thetaprima][iprima];
            }
        
        MemUtil::freeMatrix(&fc);
        
        fc = fcR;
        
        if (ImgUtil::DEBUG) {
                // Creating the fingercodes: fingerImages
            float ***fcImages = MemUtil::create3DMatrix<float>(nFilters, w, h);

            for (int idxFilter = 0; idxFilter < nFilters; idxFilter++){
                for (int posRow = iniRowOffset; posRow < endRowOffset; posRow++){
                    for (int posCol = iniColOffset; posCol < endColOffset; posCol++){
                        int label = fcLabels[posRow][posCol];
                        if (label >= 0){
                                fcImages[idxFilter][posRow][posCol] = fc[idxFilter][label];
                        }
                    }
                }
            }
            for (int k = 0; k < nFilters; k++) {
                string fileName = "FingerCodeRotated";
                char str[20];
                sprintf(str, "%d", k);
                fileName.append(str);
                ImgUtil::writeGrayScaleImage(fcImages[k], h, w, fileName, 2);

            }
            
            MemUtil::freeMatrix(&fcImages);
        }
    
        return fc;
}



void FingerCode::setSegmentationMask(bool **masc, int hMasc) {
    
    int hBlockMasc = hFi / hMasc;
    int wMasc = wFi / hBlockMasc;
    
    
    fingerMasc = MemUtil::create2DMatrix<bool>(h,w);
    // fingerMasc initialization
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++) {
            int iMasc = (i + iniRowOffset) / hBlockMasc;
            int jMasc = (j + iniColOffset) / hBlockMasc;
            if ( (i + iniRowOffset >= 0) && (i + iniRowOffset < hFi) && 
                    (j + iniColOffset >= 0) && (j + iniColOffset < wFi) && (iMasc < hMasc) && (jMasc < wMasc)) {
                fingerMasc[i][j] = masc[iMasc][jMasc];            
            }
        }
    
    if (ImgUtil::DEBUG) {
        string fileName3  = "FingerMasc";
        ImgUtil::writeGrayScaleImage(fingerMasc, h, w, fileName3);
    }

}
