/* 
 * \file        FingerCode.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for FingerCode  class
 *
 * 
 */
#ifndef FINGERCODE_H
#define	FINGERCODE_H

#include "GaborFilter.h"
#include "PCASYSManager.h"


using namespace std;

/**
 * \class FingerCode
 *
 * The FingerCode class implements FingerCode algorithm for feature extraction
 * based on Gabor Filters
 */
class FingerCode {
public:
    /** Default constructor */
    FingerCode();
    
    /**
     * FingerCode constructor for Jain
     * \param squareSize Size of the square segment to be extracted 
     *  from the original fingerprint image
     * \param fi Fingerprint image from which the square segment is obtained
     */
    /**
     * 
     * \param squareSize Size of the square segment to be extracted 
     *  from the original fingerprint image
     * \param image image to extract the fingercode
     * \param hImg height of the image
     * \param wImg width of the image
     * \param coreX core location in rows (height)
     * \param coreY core location in cols (width)
     * \param centerDisplacement displacement of the center downwards
     * \param gSupX size of the gabor support X
     * \param gSupY size of the gabor support Y
     */
    FingerCode(int squareSize, float **image, int hImg, int wImg, int coreX, int coreY, 
                                                int centerDisplacement, int gSupX, int gSupY);
    
    
    /** Default destructor */
    ~FingerCode();
     
    /**
     * 
     * \param squareSize Size of the square segment to be extracted 
     *  from the original fingerprint image
     * \param image image to extract the fingercode
     * \param hImg height of the image
     * \param wImg width of the image
     * \param coreX core location in rows (height)
     * \param coreY core location in cols (width)
     * \param centerDisplacement displacement of the center downwards
     * \param gSupX size of the gabor support X
     * \param gSupY size of the gabor support Y
     */
    void initializeFC(int squareSize, float **image, int hImg, int wImg, int coreX, int coreY, 
                                                int centerDisplacement, int gSupX, int gSupY);
    
    /**
     * Executes the Finger Code method from the fingerprint in file
     * \param file fingerprint
     * \param segmentar indicates if Poincare's segmentation must be done
     * \param pcasys pcasys manager
     * \return fingerprint fingerprint feature vector
     */
    float *run(string file, PCASYSManager *pcasys, bool segmentar = false);
    
    /**
     * Executes the Finger Code method from the fingerprint image
     * \param img fingerprint image
     * \param h height of the image
     * \param w width of the image
     * \return fingerprint fingerprint feature vector
     */
    float *run(float **img, int h, int w);
    
    /**
     * Executes the Finger Code method from the fingerprint in file and extract whether
     * it has been rejected or not
     * \param file fingerprint
     * \param segmentar indicates if Poincare's segmentation must be done
     * \param pcasys pcasys manager
     * \return fingerprint fingerprint feature vector
     */
    float *runReject(string file, PCASYSManager *pcasys, bool segmentar = false);
    
    /**
     * Executes the Finger Code method from the fingerprint in file and extract whether
     * it has been rejected or not
     * \param img fingerprint image
     * \param h height of the image
     * \param w width of the image
     * \return fingerprint fingerprint feature vector
     */
    float *runReject(float **img, int h, int w);
    
    /**
     * It computes the FingerCode Labels map (whether a pixel belongs to a sector
     * or not, and which sector is it) for the original circular FingerCode.
     * \param nSectors
     * \param nBands
     * \param bandWidth
     * \return  -1 if an error has ocurred (the fingerprint is rejected). 0 otherwise
     */
    int setLabelsFingerCode(int nSectors, int nBands, int bandWidth);
    
    /**
     * It computes the FingerCode Labels map (whether a pixel belongs to a sector
     * or not, and which sector is it) for the modified square FingerCode.
     * \return  -1 if an error has ocurred (the fingerprint is rejected). 0 otherwise
     */
    int setLabelsSquareFingerCode();
    
    /**
     * It normalizes the fingerprint image in each sector of the FingerCode to the
     * desired mean and variance
     * \return   -1 if an error has ocurred. 0 otherwise
     */
    int normalize(int Vo, int Mo);
    
    /**
     * It filters the fingerprint image (previously normalized) using the given
     * Gabor Filters, these images are stored in 
     * \param gb Gabor Filters to filter the image float ***filteredImages
     * \return -1 if an error has ocurred. 0 otherwise
     */
    int filterImage(GaborFilter *gb);
    
    /**
     * It computes the FingerCode froom the filtered images obtained with filterImage(GaborFilter *gb)
     * \return the FingerCode (one dimension is used for each filter (numFilters x numSectors)
     */
    float **computeFingerCode(int type); // 0 = STD, 1 = ABSDIFF
    
    /**
     * It filters the fingerprint image averaging with a circle filter.
     * \param filterRadio the radio of the circle filter
     * \return -1 if an error has ocured. 0 otherwise.
     */
    int circleAverage(int filterRadio);
    
    /**
     * It initializes every block of the masc to fingerprint(true) or background(false)
     */
    void computeMascPerBlock();
    
    /**
     * It returns the number of sectors used in the finger code
     * \return numSectors
     */
    int getNumSectors() const;
    
    /**
     * It rotates the extracted fingercode by the given rotation
     * \param rotate the rotation length (it depends on the number of sectors per bands
     * \return the rotated fingercode
     */
    float **rotateFingerCode(int rotate);
    
    /**
     * It sets the mask of the segmentation for the fingerprint
     * \param masc mask of the fingerprint image (0 background, 1 foreground)
     * \param hMasc height of the masc (it is supposed that is a block mask constructed with square blocks
     */
    void setSegmentationMask(bool **masc, int hMasc);
    
    /**
     * It returns the feature vector's dimension
     * \return feature vector's dimension
     */
    int getDimFV() const;
    
private:
    
    float displacedCenter[2];   //!< Displaced center h en 0 y w en 1
    
    int **fcLabels;           //!< Label for each pixel of the image; each sector is labeled with a different label
    int *pixelsPerLabel;        //!< Number of pixels in each sector / label
    
    int iniRowOffset;           //!< Initial row offset of the square from which the FingerCode is extracted
    int iniColOffset;           //!< Initial column offset
    int endRowOffset;           //!< End row offset
    int endColOffset;           //!< End column offset
    
    int squareSize;             //!< Size of the square for the FingerCode extraction
    int halfSquare;             //!< Half size of the square
    int numSectors;             //!< Number of sectors in the FingerCode
    int h;                      //!< Height of the image segment extracted
    int w;                      //!< Width of the image segment extracted
    int hFi;                    //!< Height of the initial image
    int wFi;                    //!< Width of the initial image
    
    float **fingerImg;          //!< Image segment extracted
    float **fingerImgPadded;    //!< Image segment extracted padded to compute the Gabor Filter
    float ***filteredImages;    //!< Gabor Filtered images
    int nFilters;               //!< Number of Gabor Filters to be used

    /* Features for LEUNG2010 */
    bool **fingerMasc;          //!< Segmentation of the fingerprint image in the square of the FingerCode
    bool *blockMasc;            //!< Whether each block of the fingerCode in Leung2010 is foreground or background
    
    int gaborSupportX;           //!< Support of the Gabor Filter
    int gaborSupportY;           //!< Support of the Gabor Filter
    
    float **fc;                 //!< The fingercode extracted for the fingerprint
    int nSectorsInBand;         //!< Number of sectors in each bands, needed to perform the rotation
    
    int dimFV;                          //!< Dimension of feature vector
};

inline int FingerCode::getDimFV() const { return dimFV;}

inline int FingerCode::getNumSectors() const { return numSectors; }

#endif	/* FINGERCODE_H */

