/* 
 * \file        HongFeatureExtractor.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of HongFeatureExtractor (Hong08) class
 *
 * 
 */

#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "FingerprintImg.h"

#include "FingerCode.h"
#include "HongFeatureExtractor.h"
#include "CovarianceBasedCoreDetector.h"
#include "NewPoincareSingularPoints.h"
#include "PseudoridgeTracer.h"


#include "ParametersHongFE.h"
#include "ParametersPoincare.h"


HongFeatureExtractor::HongFeatureExtractor() {

    this->img = 0;
    this->w = -1;
    this->h = -1;

    this->coreI = -1;
    this->coreJ = -1;
    this->delta1I = -1;
    this->delta1J = -1;
    this->delta2I = -1;
    this->delta2J = -1;

    this->oriImage = 0;
    this->discreteOriImage = 0;
    this->orW = -1;
    this->orH = -1;

    this->endOneI = -1;
    this->endOneJ = -1;
    this->turnedOne = false;

    this->endTwoI = -1;
    this->endTwoJ = -1;
    this->turnedTwo = false;

    this->NC = -1;
    this->ND = -1;

    this->DoneL = -1;
    this->DtwoL = -1;
    this->DoneD = -1;
    this->DtwoD = -1;
    this->RoneL = -1;
    this->RtwoL = -1;
    this->RoneD = -1;
    this->RtwoD = -1;

    this->xCos = 0;
    this->ySin = 0;
    
    this->bwiseSegmentation=0;
}

HongFeatureExtractor::~HongFeatureExtractor() {

    MemUtil::freeMatrix(&img);
    MemUtil::freeMatrix(&oriImage);
    MemUtil::freeMatrix(&discreteOriImage);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&bwiseSegmentation);
}

HongFeatureExtractor::HongFeatureExtractor(float **image, int width, int height) {

    //Copying dimensions
    this->w = width;
    this->h = height;

    //Declaring the image
    this->img = MemUtil::create2DMatrix<float>(h, w);

    //Copying the data
    for (int i = 0; i < this->h; i++) {
        for (int j = 0; j < this->w; j++) {
            this->img[i][j] = (float) image[i][j];
        }
    }
    return;

    //Default values
    this->coreI = -1;
    this->coreJ = -1;
    this->delta1I = -1;
    this->delta1J = -1;
    this->delta2I = -1;
    this->delta2J = -1;

    this->oriImage = 0;
    this->discreteOriImage = 0;
    this->orW = -1;
    this->orH = -1;
    
    this->bwiseSegmentation=0;

    this->endOneI = -1;
    this->endOneJ = -1;
    this->turnedOne = false;

    this->endTwoI = -1;
    this->endTwoJ = -1;
    this->turnedTwo = false;

    this->NC = -1;
    this->ND = -1;

    this->DoneL = -1;
    this->DtwoL = -1;
    this->DoneD = -1;
    this->DtwoD = -1;
    this->RoneL = -1;
    this->RtwoL = -1;
    this->RoneD = -1;
    this->RtwoD = -1;

    this->xCos = 0;
    this->ySin = 0;
}

float *HongFeatureExtractor::run(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){

        fi.setPCASYSManager(pcasys);    
        fi.readFingerprintImage(file);
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    /*
        * PART I: HONG
        */

    //Copying dimensions
    w = fi.getW();
    h = fi.getH();

    //Declaring the image
    img = MemUtil::copy2DMatrix<float>(fi.getSegmentedImageFloat(), h, w);
    
    computeFeatures();
    float *feats = retrieveFeatures();


    /*
        * PART II: FINGERCODE
        */

    FingerCode *fc = new FingerCode();
    float *fetVec = fc->run(img, h , w);
    int dim = fc->getDimFV();
        dimFV = 10 + dim;
    
    
    
    dimFV = 10 + dim;
    float *fv = MemUtil::createVector<float>(dimFV);

    for (int i = 0; i < 10; i++) { 
        fv[i] = feats[i];
    }
    
    for (int i = 0; i < dim; i++){ 
        fv[10 + i] = fetVec[i];
    }
    cout << endl;
        
    delete fc;
    MemUtil::freeMatrix(&fetVec);
    MemUtil::freeMatrix(&feats);

    return fv;

}

float *HongFeatureExtractor::runReject(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){

        fi.setPCASYSManager(pcasys);   
        fi.readFingerprintImage(file);    
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    /*
        * PART I: HONG
        */

    //Copying dimensions
    w = fi.getW();
    h = fi.getH();

    //Declaring the image
    img = MemUtil::copy2DMatrix<float>(fi.getSegmentedImageFloat(), h, w);

    /*
        * PART II: FINGERCODE
        */

    FingerCode *fc = new FingerCode();
    float *fetVec = fc->runReject(img, h , w);
    dimFV = fc->getDimFV();
    
    delete fc;
    
    return fetVec;
}

void HongFeatureExtractor::computeFeatures() {

    //Calculating the orientation field/map
    computeOrientations(((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust)));

    //Discretize the orientation map
    discretizeOrientations(numOfDiscreteDirectionsHongFE);

    //Compute the core & the deltas
    computeCoresAndDeltas();

    
    // Tracing
    tracePseudoridges();


    calculateFeatureValues();


    return;
}

float* HongFeatureExtractor::retrieveFeatures() {

    float* featureVector = MemUtil::createVector<float>(10);

    featureVector[0] = this->NC;
    featureVector[1] = this->ND;
    featureVector[2] = this->DoneL;
    featureVector[3] = this->DtwoL;
    featureVector[4] = this->RoneL;
    featureVector[5] = this->RtwoL;
    featureVector[6] = this->DoneD;
    featureVector[7] = this->DtwoD;
    featureVector[8] = this->RoneD;
    featureVector[9] = this->RtwoD;

    return featureVector;
}

void HongFeatureExtractor::tracePseudoridges() {


    PseudoridgeTracer prt(this->oriImage, this->orH, this->orW, (int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));

    prt.setBlockwiseSegmentation(this->bwiseSegmentation);
    prt.setCenter(coreI, coreJ);
    

    prt.tracePseudoridges(numOfDiscreteDirectionsHongFE, ((int)(pseudoRidgeMaxlengthHongFE * ImgUtil::fAdjust)), TURNINGRULES_HONG);

    int *endPosition = prt.getFirstEnd();
    this->endOneI = endPosition[0];
    this->endOneJ = endPosition[1];
    this->turnedOne = prt.firstPRidgeTurned();

    endPosition = prt.getSecondEnd();
    this->endTwoI = endPosition[0];
    this->endTwoJ = endPosition[1];
    this->turnedTwo = prt.secondPRidgeTurned();

    if (ImgUtil::DEBUG) {
        cout << "HongFeatureExtractor: The end of the first pRidge is (" << endOneI << "," << endOneJ << ") turned=" << turnedOne << endl;
        cout << "HongFeatureExtractor: The end of the second pRidge is (" << endTwoI << "," << endTwoJ << ") turned=" << turnedTwo << endl;
        cout << "HongFeatureExtractor: The number of turns in the pRidge is " << prt.getNumberOfTurns() << endl;
    }



    if (ImgUtil::DEBUG) {
cout << "HF:\n\th = " << h << endl << "w = " << w << endl;  
        int **pseudoRImage = prt.getPseudoridge();

        float** displayableImage = MemUtil::create2DMatrix<float>(this->h, this->w);
        //Copying the data
        for (int i = 0; i < this->orH * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust)); i++) {
            for (int j = 0; j < this->orW * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust)); j++) {

                displayableImage[i][j] = (this->img[i][j])*0.5 + ((float) pseudoRImage[i][j])*125;
            }
        }
        string filename = "PRidge";
        ImgUtil::writeGrayScaleImage(displayableImage, this->h, this->w, filename, 1);

        MemUtil::freeMatrix(&displayableImage);

    }

    return;
}

void HongFeatureExtractor::computeOrientations(int blockW) {

    this->orH = h / blockW;
    this->orW = w / blockW;

    this->xCos = MemUtil::create2DMatrix<float>(orH, orW);
    this->ySin = MemUtil::create2DMatrix<float>(orH, orW);
    this->oriImage = MemUtil::create2DMatrix<float>(orH, orW);

    float **Vx = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH, orW);

    float **gradX = MemUtil::create2DMatrix<float>(h, w);
    float **gradY = MemUtil::create2DMatrix<float>(h, w);


    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] -
                    img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]);

            gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                    img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]);

        }
    }


    float Gsx, Gsy;
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                    Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;
                }
            }

            oriImage[i][j] = 0.5 * atan2(Vx[i][j], Vy[i][j]);
            oriImage[i][j] += (oriImage[i][j] <= 0) ? 0.5 * PI : -0.5 * PI;

            float mag = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]);
            xCos[i][j] = cos(2.0 * oriImage[i][j]) * mag;
            ySin[i][j] = sin(2.0 * oriImage[i][j]) * mag;
        }
    }




    if (ImgUtil::DEBUG) {

        string fileName2 = "OrientationVyVx";
        ImgUtil::writeOrImg(Vx, Vy, orW, orH, fileName2);
        fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);

    }

    ImgUtil::smoothOrientations(oriImage, &xCos, &ySin, orH, orW, smoothOrientatioWHongFE);

    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    
    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&Vy);
    MemUtil::freeMatrix(&gradY);

}

void HongFeatureExtractor::computeCoresAndDeltas() {

    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(oriImage, xCos, ySin, orH, orW);
    
    npsp->generateSegmentation(img, h, w);

    //computeCoresAndDeltas(int smoothW, int curveW, int specialRules, bool normalize)
    //By default this call uses increasing smooth, as Karu did
    // specialRules==2 means Hong rules
    npsp->computeCoresAndDeltas(curveWHongFE, smoothWPoincareHongFE, poincareModeHongFE, rulesToBeAppliedHongFE, minNBlocksHongFE); // AQUI TIENE QUE SER 7 Y PARAMETRIZADO


    //Retrieving & setting up the segmentation    
    this->bwiseSegmentation=MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(), this->orH,this->orW);
    
    // Working with the deltas
    this->ND = npsp->getnDeltas();
    vector<float> deltaCentroids = npsp->getDeltaCentroids();
    if (this->ND == 1) {
        this->delta1I = deltaCentroids[0] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
        this->delta1J = deltaCentroids[1] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
    }
    if (this->ND == 2) {
        this->delta2I = deltaCentroids[2] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
        this->delta2J = deltaCentroids[3] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
    }


    //Working with the cores
    this->NC = npsp->getnCores();
    vector<float> coreCentroids;
    if (this->NC == 0) {
        if (ImgUtil::DEBUG)
            cout << "HongFeatureExtractor: No cores were found by the PI method" << endl;

        /*
         * New version: center of the image
         */
        this->coreI = (this->h/2);
        this->coreJ = (this->w/2);
        
        if (ImgUtil::DEBUG)
            cout << "\t HongFeatureExtractor:  new core is (" << this->coreI << ", " << this->coreJ << ")" << endl;

        
    } else if (this->NC == 1) {
        coreCentroids = npsp->getCoreCentroids();
        this->coreI = coreCentroids[0] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
        this->coreJ = coreCentroids[1] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
    } else if (this->NC == 2) {
        //If so get the closest one to the center
        coreCentroids = npsp->getCoreCentroids();
        float imageCenterI = this->orH / 2;
        float imageCenterJ = this->orW / 2;
        float dist1 = (imageCenterI - coreCentroids[0])*(imageCenterI - coreCentroids[0]) +
                (imageCenterJ - coreCentroids[1])*(imageCenterJ - coreCentroids[1]);
        float dist2 = (imageCenterI - coreCentroids[2])*(imageCenterI - coreCentroids[2]) +
                (imageCenterJ - coreCentroids[3])*(imageCenterJ - coreCentroids[3]);

        if (dist1 <= dist2) {
            //Then we select the first one
            this->coreI = coreCentroids[0] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
            this->coreJ = coreCentroids[1] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
        } else {
            //Then we select the first one
            this->coreI = coreCentroids[2] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
            this->coreJ = coreCentroids[3] * ((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust));
        }

    }

    if (ImgUtil::DEBUG) {
        cout << "HongFeatureExtractor:  The P-I method found " << npsp->getnDeltas() << " and " << npsp->getnCores() << "cores" << endl;
    }
    
    if (TEST_1){
        string fileName=ImgUtil::makeTestFileName("SingularPoints");
        
        int *auxDeltaCentroids= 0;
        if (ND > 0)
            auxDeltaCentroids = MemUtil::createVector<int>(ND*2);
        
        int *auxCoreCentroids = 0;
        if (NC>0)
            auxCoreCentroids=MemUtil::createVector<int>(NC*2);
            
        
        cout <<"HongFeatureExtractor::computeCoresAndDeltas::TEST_1:: Deltas-> ";
        for (int i=0; i<ND*2; i++){auxDeltaCentroids[i]=(deltaCentroids[i]+0.5)*((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust)); cout <<auxDeltaCentroids[i]<<"\t";}
        cout <<endl;
        cout <<"HongFeatureExtractor::computeCoresAndDeltas::TEST_1:: Cores-> ";
        for (int i=0; i<NC*2; i++){auxCoreCentroids[i]=(coreCentroids[i]+0.5)*((int)(orientationsBlockSizeHongFE * ImgUtil::fAdjust)); cout <<auxCoreCentroids[i]<<"\t";}
        cout <<endl;
        
        
        
        ImgUtil::writeSPImage(this->img,this->h,this->w,auxDeltaCentroids,ND,auxCoreCentroids,NC,fileName);
        
        MemUtil::freeMatrix(&auxDeltaCentroids);
        MemUtil::freeMatrix(&auxCoreCentroids);
    }

    delete npsp;
}

void HongFeatureExtractor::calculateFeatureValues() {

    // FASE 5: Calculo de las caracteristicas

    /*
     * Numero de puntos singulares 
     *       NC-Number of Cores
     *       ND-Number of Deltas
     */

    /*
     * Location of the delta points wrt the core
     *       D1L-Location of the first delta
     *       D2L-Location of the first delta
     * Distance of the delta points to the core
     *       D1D-Distance to the first delta
     *       D2D-Distance to the first delta
     */

    if (ND == 0) {
        this->DoneL = location_absentHongFE;
        this->DtwoL = location_absentHongFE;
        this->DoneD = distance_absentHongFE;
        this->DtwoD = distance_absentHongFE;
    } else if (ND == 1) {
        this->DoneL = computeLocation(coreI, coreJ, delta1I, delta1J);
        this->DoneD = computeDistance(coreI, coreJ, delta1I, delta1J);
        this->DtwoL = location_absentHongFE;
        this->DtwoD = distance_absentHongFE;
    } else {
        this->DoneL = computeLocation(coreI, coreJ, delta1I, delta1J);
        this->DoneD = computeDistance(coreI, coreJ, delta1I, delta1J);
        this->DtwoL = computeLocation(coreI, coreJ, delta2I, delta2J);
        this->DtwoD = computeDistance(coreI, coreJ, delta2I, delta2J);
    }


    if (turnedOne) {
        this->RoneL = location_turnedHongFE;
        this->RoneD = distance_absentHongFE;
    } else {
        this->RoneL = computeLocation(coreI, coreJ, endOneI, endOneJ);
        this->RoneD = computeDistance(coreI, coreJ, endOneI, endOneJ);
    }


    if (turnedTwo) {
        this->RtwoL = location_turnedHongFE;
        this->RtwoD = distance_absentHongFE;
    } else {
        this->RtwoL = computeLocation(coreI, coreJ, endTwoI, endTwoJ);
        this->RtwoD = computeDistance(coreI, coreJ, endTwoI, endTwoJ);
    }


}

void HongFeatureExtractor::discretizeOrientations(int numOfOrientations) {

    this->discreteOriImage = MemUtil::create2DMatrix<int>(orH, orW);

    //Orientations are assumed to be [-pi,pi]
    //numOfOrientations should (MUST!) be divisible by 4

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            //Dirnum is the one initially representing -pi
            int dirNum = 0.75 * numOfOrientations;
            //(PI/numOfOrientation) = half of one of the (numOfOrientations) sections amplitude
            float angleLimit = -(PI / 2)+(PI / (float) numOfOrientations);
            while (angleLimit < oriImage[i][j]) {
                dirNum = (dirNum + 1) % 16;
                angleLimit += ((2 * PI) / (float) numOfOrientations);
            }

            this->discreteOriImage[i][j] = dirNum;
        }
    }


    return;

}

int HongFeatureExtractor::computeLocation(int coreI, int coreJ, int pointI, int pointJ) {

    int loc;

    if (coreI > pointI) {
        if (coreJ > pointJ)
            loc = 0;
        else
            loc = 4;
    } else if ((coreJ > pointJ) && (abs(coreI - pointI) < 4 * abs(coreJ - pointJ))) {
        loc = 1;
    } else if (abs(coreI - pointI) >= 4 * abs(coreJ - pointJ)) {
        loc = 2;
    } else {
        loc = 3;
    }

    return loc;
}

int HongFeatureExtractor::computeDistance(int coreI, int coreJ, int pointI, int pointJ) {

    float squaredEuc = ((coreJ - pointJ) * (coreJ - pointJ)) + ((coreI - pointI) * (coreI - pointI));
    float blockSize = this->w / this->orW;

    int dist = -1;

    if (squaredEuc <= 10 * blockSize) {
        dist = 1;
    } else if (squaredEuc <= 20 * blockSize) {
        dist = 2;
    } else {
        dist = 3;
    }

    return dist;
}

int HongFeatureExtractor::getPRidgeOriginI() {
    return this->coreI;
}

int HongFeatureExtractor::getPRidgeOriginJ() {
    return this->coreJ;
}
