/* 
 * \file        HongFeatureExtractor.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     Feature extraction method based on pseudoridges by Hong et al.
 * 
 * Header file for HongFeatureExtractor class
 *
 * 
 */


#ifndef HONGFEATUREEXTRACTOR_H
#define	HONGFEATUREEXTRACTOR_H

#include "PCASYSManager.h"

using namespace std;

/**
 * \class HongFeatureExtractor
 *
 * The HongFeatureExtractor class implements the feature extraction technique
 *      used in
 *  
 * Jin-Hyuk Hong, Jun-Ki Min, Ung-Keun Cho, Sung-Bae Cho
 * "Fingerprint classification using one-vs-all support vector machines dynamically ordered with naïve Bayes classifiers" 
 * Pattern Recognition, 
 * pp. 662-971, February 2008
 * 
 * Note that this class only computes the features related to the pseudoridge. The authors
 *      use this information in combination with the fingercodes, whose extraction is implemented
 *      separately.
 */
class HongFeatureExtractor {
public:
    HongFeatureExtractor();
    HongFeatureExtractor(float **image, int width, int height);
    virtual ~HongFeatureExtractor();
    
    //Main functions
    
    /**
     * Executes the Feature Extractor Hong method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return fingerprint class
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * Executes the Feature Extractor Hong method from the fingerprint in file
     * to know whether it has been rejected or not
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return fingerprint class
     */
    float *runReject(string file, PCASYSManager *pcasys);

    /**
     * Processes the image to compute the features
     */
    void computeFeatures();
    
    /**
     * Returns a vector of 10 integers with the feature values
     */
    float *retrieveFeatures();
    
    //Auxiliars
    
    /**
     * Computes the orientations map using blockW x blockW tiles
     * \param blockW    Size of the blocks used for avg orientation calculation.
     */
    void computeOrientations(int blockW);
    
    /**
     * Computes the cores and deltas using the Poincare Index.
     */
    void computeCoresAndDeltas();
    
    /**
     * Fills the discrete orientation map. Must be called after computeOrientations();
     * \param numOfOrientations Number of discrete orientations the circumference is divided in.
     */
    void discretizeOrientations(int numOfOrientations);
    
    /**
     * Traces the pseudoridges from the core
     */
    void tracePseudoridges();
    
    /**
     * Calculates the feature values using the core/delta info and the pseudoridges.
     */
    void calculateFeatureValues();
    
    /**
     * Updates the J (horizontal) component of the pseudoridge based on the dir of the orientation
     * \param oldJ  Current colum
     * \param dir   Orientation dir
     * \return      Updated colums
     */
    int evolveJ(int oldJ, int dir);
    
    /**
     * Updates the I (vertical) component of the pseudoridge based on the dir of the orientation
     * \param oldI  Current row
     * \param dir   Orientation dir
     * \return      Updated row
     */
    int evolveI(int oldI, int dir);
    
    /**
     * Computes the value of a given "location" feature wrt to the position of the core.
     * \param coreI Vertical position of the core
     * \param coreJ Horizontal position of the core
     * \param pointI    Vertical position of the point
     * \param pointJ    Horizontal position of the point
     * \return  Location value (which can only be in {0,4})
     */
    int computeLocation(int coreI,int coreJ, int pointI, int pointJ);
    
    /**
     * Computes the value of a given "distance" feature wrt to the position of the core.
     * \param coreI Vertical position of the core
     * \param coreJ Horizontal position of the core
     * \param pointI    Vertical position of the point
     * \param pointJ    Horizontal position of the point
     * \return  Location value (which can only be in {1,3})
     */
    int computeDistance(int coreI,int coreJ, int pointI, int pointJ);
    
    
    /**
     * Gets the position from which the PRidge started
     * \return the I (vertical) component of the center of the PRidge
     */
    int getPRidgeOriginI();
    
    /**
     * Gets the position from which the PRidge started
     * \return the J (vertical) component of the center of the PRidge
     */
    int getPRidgeOriginJ();
    
    float **getXCos() const;
    float **getYSin() const;
    
    float **getOriImage() const;
    
    int getOrH() const;
    int getOrW() const;
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:

    float **img;                //!< Fingerprint image
    int w;                      //!< Width of the image
    int h;                      //!< Height of the image
    
    float coreI;                  //!< Vertical position of the core
    float coreJ;                  //!< Horizontal position of the core
    float delta1I;                //!< Vertical position of the first delta (if any, otherwise -1)
    float delta1J;                //!< Horizontal position of the first delta (if any, otherwise -1)
    float delta2I;                //!< Vertical position of the second delta (if any, otherwise -1)
    float delta2J;                //!< Horizontal position of the second delta (if any, otherwise -1)
    
    float **oriImage;           //!< Orientation map
    int **discreteOriImage;     //!< Discrete orientation map
    int orW;                    //!< Width of the orientation map
    int orH;                    //!< Height of the orientation map
    
    bool **bwiseSegmentation;   //!< Blockwise segmentation
    
    int endOneI;                //!< Vertical position of the first end of the pseudoridge
    int endOneJ;                //!< Horizontal position of the first end of the pseudoridge
    bool turnedOne;             //!< Wheter the second end of the pseudoridge is due to a turn
    
    int endTwoI;                //!< Vertical position of the first end of the pseudoridge
    int endTwoJ;                //!< Horizontal position of the first end of the pseudoridge
    bool turnedTwo;             //!< Wheter the second end of the pseudoridge is due to a turn
  
    int NC;                     //!< Number of cores in the image
    int ND;                     //!< Number of deltas in the image
    
    int DoneL;                  //!< "Location" of the first delta wrt the core
    int DtwoL;                  //!< "Location" of the second delta wrt the core
    int DoneD;                  //!< "Distance" of the first delta wrt the core
    int DtwoD;                  //!< "Distance" of the second delta wrt the core
    int RoneL;                  //!< "Location" of the first pseudoridge ending wrt the core
    int RtwoL;                  //!< "Location" of the second pseudoridge ending wrt the core
    int RoneD;                  //!< "Distance" of the first pseudoridge ending wrt the core
    int RtwoD;                  //!< "Distance" of the second pseudoridge ending wrt the core
    
    float **xCos;               //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;               //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    int dimFV;                          //!< Dimension of feature vector
};  

inline int HongFeatureExtractor::getDimFV() const { return dimFV;}

inline float **HongFeatureExtractor::getOriImage() const{ return this->oriImage;}
inline float **HongFeatureExtractor::getXCos() const{ return this->xCos;}
inline float **HongFeatureExtractor::getYSin() const{ return this->ySin;}

inline int HongFeatureExtractor::getOrH() const {return this->orH;}
inline int HongFeatureExtractor::getOrW() const {return this->orW;}

#endif	/* HONGFEATUREEXTRACTOR_H */

