/* 
 * \file        ImageEnhancementHong.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of ImageEnhancementHong (required for Leung2010) class
 *
 * 
 */

#include "ImageEnhancementHong.h"
#include "NewPoincareSingularPoints.h"
#include "FingerprintImg.h"
#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "ParametersLeung.h"


using namespace std;

ImageEnhancementHong::ImageEnhancementHong() {
    img = 0;
    imgCopy = 0;
    h = -1;
    w = -1;
    
    blockVariance = 0;
    ridgeFrequency = 0;
    
    averagePeaksValleysSeg = 0;
    blockVarianceSeg = 0;
    ridgeFrequencySeg = 0;
    
    ySin = 0;
    xCos = 0;
    O = 0;
    masc = 0;
    mascPrueba = 0;
    
    maxVar = -10000;
    maxAvg = -10000;
    maxFreq = -10000;
    minVar = 10000;
    minAvg = 10000;
    minFreq = 10000;
    
    initializeBlocks();
}


ImageEnhancementHong::ImageEnhancementHong(unsigned char **image, int h, int w) {
    
    img = MemUtil::create2DMatrix<float>(h,w);
    imgCopy = MemUtil::create2DMatrix<float>(h,w);
    // img initialization
    for (int i = 0; i < h; i++) {
        for (int j = 0; j < w; j++){
            img[i][j] = (float) image[i][j];
            imgCopy[i][j] = (float) image[i][j];
        }
    }
    
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "ImageHong";
        ImgUtil::writeGrayScaleImage(imgCopy, h, w, fileName2, 1);
    }
    
    this->h = h;
    this->w = w;
    blockVariance = 0;
    ridgeFrequency = 0;
    ySin = 0;
    xCos = 0;
    O = 0;
    masc = 0;
    
    maxVar = -10000;
    maxAvg = -10000;
    maxFreq = -10000;
    minVar = 10000;
    minAvg = 10000;
    minFreq = 10000;
    
    initializeBlocks();
}

ImageEnhancementHong::~ImageEnhancementHong() {
    
    MemUtil::freeMatrix(&img);
    
    MemUtil::freeMatrix(&imgCopy);
    
    MemUtil::freeMatrix(&O);
    
    MemUtil::freeMatrix(&xCos);
    
    MemUtil::freeMatrix(&ySin);
    
    MemUtil::freeMatrix(&ridgeFrequency);
    
    MemUtil::freeMatrix(&blockVariance);
    
    MemUtil::freeMatrix(&averagePeaksValleysSeg);
    
    MemUtil::freeMatrix(&ridgeFrequencySeg);
    
    MemUtil::freeMatrix(&blockVarianceSeg);

    MemUtil::freeMatrix(&masc);
    MemUtil::freeMatrix(&mascPrueba);
}

void ImageEnhancementHong::initialize(float **image, int h, int w){
    this->h = h;
    this->w = w;
    
    
    img = MemUtil::copy2DMatrix<float>(image, h, w);
    imgCopy = MemUtil::copy2DMatrix<float>(image, h, w);
}

void ImageEnhancementHong::initializeBlocks(){
    
    if (((int)(blockSizeLeung * ImgUtil::fAdjust)) % 2 == 1)
        block = blockSizeLeung * ImgUtil::fAdjust - 1;
    else
        block = blockSizeLeung * ImgUtil::fAdjust;

    if (((int)(wLeungHong * ImgUtil::fAdjust) % 2) == 1)
        Wblock = wLeungHong * ImgUtil::fAdjust - 1;
    else
        Wblock = wLeungHong * ImgUtil::fAdjust;

    if (((int)(lLeungHong * ImgUtil::fAdjust) % 2) == 1)
        Lblock = lLeungHong * ImgUtil::fAdjust - 1;
    else
        Lblock = lLeungHong * ImgUtil::fAdjust;
    
}

float *ImageEnhancementHong::run(string file, PCASYSManager *pcasys, bool case2, bool segmentar){
    FingerprintImg fi;
    float **fingercode;
    float **imgEnhanced;

    if (ImgUtil::fAdjust != 1){
        
		fi.setPCASYSManager(pcasys); 
        fi.readFingerprintImage(file);        
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{

		fi.setPCASYSManager(pcasys); 
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());

    enhance(M0Leung, V0Leung, case2); // Devuelve la imagen que deberá usar getModifiedFingerCode()

    if (!case2)
        varianceSegment(varianceThresholdLeung);

    if (segmentar){
        // wLeungHong
        NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(fi.getSegmentedImageFloat(),  fi.getH(), fi.getW());
        npsp->setOrH(gethOr());
        npsp->setOrW(getwOr());
        npsp->generateSegmentation(); //fi.getSegmentedImageFloat(),  fi.getH(), fi.getW());
        npsp->combineWithSegmentation(masc);
        MemUtil::freeMatrix(&masc);
        masc = MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(), gethOr(), getwOr());
        delete npsp;
    }
    
    int aux = gaborSupLeungHong * ImgUtil::fAdjust;
    if (aux % 2 == 0)
        aux--;
     imgEnhanced = filterImage(aux, deltaGaborLeungHong * ImgUtil::fAdjust, deltaGaborLeungHong * ImgUtil::fAdjust, case2);

     
    alterOrientations();
    
    //Calcula el core para centrar la imagen cuando se instancia el FingerCode
    fi.computeCore_r92a(getxCos(), getySin(), getwOr(), gethOr(), block);

    fingercode = fi.getFingerCodeLeung(imgEnhanced, centerDisplacementLeung, masc, gethOr());

    dimFV = fi.getNFilters()*fi.getNumSectors();
    
    float *featureVector = MemUtil::createVector<float>(dimFV);
    int cont = 0;
    
    for (int i = 0; i < fi.getNFilters(); i++) {
        for (int j = 0; j < fi.getNumSectors(); j++) {
            if (fingercode[i][j] == 0)
                fingercode[i][j] = -1;
            featureVector[cont++] = fingercode[i][j];
        }
    }

    MemUtil::freeMatrix(&imgEnhanced);
 
    cout << endl;
    
    return featureVector;
}

void ImageEnhancementHong::alterOrientations(){
    
    for (int i = 0; i < orH; i++)
        for (int j = 0; j < orW; j++)
            if (!masc[i][j]){
                xCos[i][j] = 0;
                ySin[i][j] = 0;
            }
    
    if (ImgUtil::DEBUG){
        
        string fileName2;
        fileName2 = "Orientations00";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
}

float **ImageEnhancementHong::enhance(int Mo, int Vo, bool binarizar) {
        
    //float **imagen;
    // AQUí va todo el proceso
    normalize(Mo, Vo); // M0 = 100 y V0 = 100
    
    createOrientationImage(Wblock); // block size = 16 x 16
    createRidgeFrecuencyImage(Wblock, Lblock);
    
    if (binarizar)
        segmentImg();
    // define GABOR_SUPPORT 11
    // #define GABOR_SIGMA_X 4 // 5 in leung
    //  #define GABOR_SIGMA_X 4 // in leung
//    imagen = filterImage(gaborSupLeungHong * ImgUtil::fAdjust, deltaGaborLeungHong * ImgUtil::fAdjust, deltaGaborLeungHong * ImgUtil::fAdjust, binarizar);

    return 0;
}

void ImageEnhancementHong::normalize(int desiredMean, int desiredVariance) {
    
    float mean = 0;
    float variance = 0;
    
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
            mean += img[i][j];
    mean /=  (float) (h * w);
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
            variance += (img[i][j] - mean) * (img[i][j] - mean);
    
    variance /= (float) (h * w);
    
    float min = 255, max = 0;
    float aux;
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++) {
            aux = sqrt((desiredVariance * 
                    ((img[i][j] - mean) * (img[i][j] - mean))) / variance);
            
            aux *= img[i][j] > mean ? 1.0 : -1.0;

            img[i][j] = (desiredMean + aux);
            
            // For further normalization
            if (img[i][j] < min)
                min = img[i][j];
            if (img[i][j] > max)
                max = img[i][j]; 
        }
        
    if (ImgUtil::DEBUG) {
        string fileName2 = "NormalizedHong";
        ImgUtil::writeGrayScaleImage(img, h, w, fileName2, 1);
    }
    
}


void ImageEnhancementHong::createOrientationImage(int blockW) {
    
    orH = h / blockW;
    orW = w / blockW;
    
    O = MemUtil::create2DMatrix<float>(orH, orW);
    xCos = MemUtil::create2DMatrix<float>(orH, orW);
    ySin = MemUtil::create2DMatrix<float>(orH, orW);

    float **Vx = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH, orW);

    float **xCosSmooth = MemUtil::create2DMatrix<float>(orH, orW);
    float **ySinSmooth = MemUtil::create2DMatrix<float>(orH, orW);   
    
    float **gradX = MemUtil::create2DMatrix<float>(h, w);
    float **gradY = MemUtil::create2DMatrix<float>(h, w);
    float **grad = MemUtil::create2DMatrix<float>(h, w);
    
    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] - 
                   img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]) ;
            
            gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                    img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]) ;
            
            grad[i][j] = sqrt(gradX[i][j] * gradX[i][j] + gradY[i][j] * gradY[i][j]);
        }
    }
    
    
    float Gsx, Gsy;
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                    Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;
                    // mag[i][j] += sqrt(Gsx * Gsx + Gsy * Gsy);
                }
            }
            // Coherence NOT NEEDED HERE!
            // mag[i][j] = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]) / mag[i][j];
            //float mag = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]);
            
            
            float ang = 0.5 * atan2(Vx[i][j], Vy[i][j]);
            if (ang <= 0)
                O[i][j] = ang + 0.5 * PI;
            else
                O[i][j] = ang - 0.5 * PI;
                
            xCos[i][j] = cos(2.0 * O[i][j]); // * mag;
            ySin[i][j] = sin(2.0 * O[i][j]); // * mag;
        }
    }
    
    /* Smooth orientations */
    
    int sWhalf = (smoothOrientatioWLeung - 1)/ 2;
    int nElements = 0;
    

    // nElements = smoothOrientatioW * smoothOrientatioW;
    for (int i = 0; i < orH ; i++) {
        for (int j = 0; j < orW ; j++) {
            nElements = 0;
            for (int u = max(i - sWhalf, 0); u < min(i + sWhalf + 1, orH); u++) {
                for (int v = max(j - sWhalf, 0); v < min(j + sWhalf + 1, orW); v++) {
                    xCosSmooth[i][j] += xCos[u][v];
                    ySinSmooth[i][j] += ySin[u][v];
                    nElements++;
                }
            }
            xCosSmooth[i][j] /= (float) nElements;
            ySinSmooth[i][j] /= (float) nElements;
            O[i][j] = 0.5 * atan2(ySinSmooth[i][j], xCosSmooth[i][j]);
            
            
            xCosSmooth[i][j] = cos(2.0 * O[i][j]);
            ySinSmooth[i][j] = sin(2.0 * O[i][j]);
        }
    }
    
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "GradX";
        ImgUtil::writeGrayScaleImage(gradX, h, w, fileName2, 3);
        fileName2 = "GradY";
        ImgUtil::writeGrayScaleImage(gradY, h, w, fileName2, 3);
        fileName2 = "Grad";
        ImgUtil::writeGrayScaleImage(grad, h, w, fileName2, 3);
        
        fileName2 = "OrientationVyVx";
        ImgUtil::writeOrImg(Vx, Vy, orW, orH, fileName2);
        fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
        
        fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCosSmooth, ySinSmooth, orW, orH, fileName2);
    }
    
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&Vy);
    
    xCos = xCosSmooth;
    ySin = ySinSmooth;
    
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&grad);
}


void ImageEnhancementHong::createRidgeFrecuencyImage(int blockW, int blockL) {
    
    this->blockW = blockW;
    
    blockVariance = MemUtil::create2DMatrix<float>(orH, orW);
    ridgeFrequency = MemUtil::create2DMatrix<float>(orH, orW);
    
    averagePeaksValleysSeg = MemUtil::create2DMatrix<float>(orH, orW);
    blockVarianceSeg = MemUtil::create2DMatrix<float>(orH, orW);
    ridgeFrequencySeg = MemUtil::create2DMatrix<float>(orH, orW);
    float **blockVarianceSmooth =MemUtil::create2DMatrix<float>(orH, orW);
    float **ridgeFrequencyInterpolated = MemUtil::create2DMatrix<float>(orH, orW);
    
    float *X = MemUtil::createVector<float>(blockL);
    
    float blockMean = 0;
    int nSig = 0;
    /* Compute signature X for each block l x w, centered in , j (the orientation gives the y-coordinate */
    int u, v;
    for (int iBlock = 0; iBlock < orH; iBlock++) {
        for (int jBlock = 0; jBlock < orW; jBlock++) {
            float ang = O[iBlock][jBlock];
            if (ang <= PI / 2.)
                ang = ang - 0.5 * PI;
            else
                ang = ang + 0.5 * PI;
            
            blockMean = 0;
            nSig = 0;
            
            for (int k = 0; k < blockL; k++) {
                int n = 0;
                X[k] = 0;
                for (int d = 0; d < blockW; d++) {
                    int i = (iBlock * blockW) + blockW / 2;
                    int j = (jBlock * blockW) + blockW / 2;
                    
                    u = (int) (i + (d - blockW / 2.) * cos(ang) + (k - blockL / 2.) * sin(ang));
                    v = (int) (j + (d - blockW / 2.) * sin(ang) + (blockL / 2. - k) * cos(ang));

                    if ((u >= 0) && (v >= 0) && (u < h) && (v < w)) {
                        X[k] += img[u][v];
                        n++;
                    }
                }
                if (n != 0) {
                        X[k] /= (float) n;
                        blockMean += X[k];
                        nSig++;
                }
            }
            blockMean /= (float) nSig;
            for (int k = 0; k < blockL; k++) 
                if (X[k] != 0)
                        blockVariance[iBlock][jBlock] += (X[k] - blockMean) * (X[k] - blockMean);
            blockVariance[iBlock][jBlock] /= (float) nSig;
                
            float ridgeF = computeRidgeFrequencyFromSignature(X, blockL, averagePeaksValleysSeg, iBlock, jBlock);   // nPeaks <= 1 no entra en el rango

            blockVarianceSeg[iBlock][jBlock] = blockVariance[iBlock][jBlock];
            
            if ((ridgeF <= ridgeFrequencyMinLeung) && (ridgeF >= ridgeFrequencyMaxLeung)) // for 500dpi images (nPeaks > 1) && 
                ridgeFrequency[iBlock][jBlock] = ridgeF;
            else
                ridgeFrequency[iBlock][jBlock] = -1;

            
            ridgeFrequencySeg[iBlock][jBlock] = ridgeF;
            
            if (isinf(ridgeFrequencySeg[iBlock][jBlock]))
                ridgeFrequencySeg[iBlock][jBlock] = 0;
            
            /* Guardamos los valores máximos y mínimos de cada variable (amplitud, frecuencia y varianza) */
            if (averagePeaksValleysSeg[iBlock][jBlock] > maxAvg)
                maxAvg = averagePeaksValleysSeg[iBlock][jBlock];
            if (averagePeaksValleysSeg[iBlock][jBlock] < minAvg)
                minAvg = averagePeaksValleysSeg[iBlock][jBlock];
            
            if (blockVarianceSeg[iBlock][jBlock] > maxVar)
                maxVar = blockVarianceSeg[iBlock][jBlock];
            if (blockVarianceSeg[iBlock][jBlock] < minVar)
                minVar = blockVarianceSeg[iBlock][jBlock];
            
            if (ridgeFrequencySeg[iBlock][jBlock] > maxFreq)
                maxFreq = ridgeFrequencySeg[iBlock][jBlock];
            if (ridgeFrequencySeg[iBlock][jBlock] < minFreq)
                minFreq = ridgeFrequencySeg[iBlock][jBlock];
            
        }
    }
    
    if (TEST_1){
        
        ofstream f2;
        
        string fileName = ImgUtil::makeTest2FileName("BlockVariance");
        fileName.append(".txt");
        char *nombre = MemUtil::createVector<char>(fileName.size());
        strcpy(nombre, fileName.c_str());
        f2.open(nombre, ofstream::out | ofstream::trunc);
        for (int i = 0; i < orH; i++){
            for (int j = 0; j < orW; j++)
                f2 << blockVarianceSeg[i][j] << " ";
            
            f2 << endl;
        }
        f2.close();
        
        MemUtil::freeMatrix(&nombre);

        fileName = ImgUtil::makeTest2FileName("AveragePeaksValleys");
        fileName.append(".txt");
        nombre = MemUtil::createVector<char>(fileName.size());
        strcpy(nombre, fileName.c_str());
        f2.open(nombre, ofstream::out | ofstream::trunc);
        for (int i = 0; i < orH; i++){
            for (int j = 0; j < orW; j++)
                f2 << averagePeaksValleysSeg[i][j] << " ";
            
            f2 << endl;
        }
        f2.close();
        MemUtil::freeMatrix(&nombre);

        fileName = ImgUtil::makeTest2FileName("RidgeFrequency");
        fileName.append(".txt");
        nombre = MemUtil::createVector<char>(fileName.size());
        strcpy(nombre, fileName.c_str());
        f2.open(nombre, ofstream::out | ofstream::trunc);
        for (int i = 0; i < orH; i++){
            for (int j = 0; j < orW; j++)
                f2 << ridgeFrequencySeg[i][j] << " ";
            
            f2 << endl;
        }
        f2.close();
        MemUtil::freeMatrix(&nombre);
    }
    
    MemUtil::freeMatrix(&X);
    
    /* Interpolate ridge frequency */
    // Create the gaussian kernel W 7x7 mean 0 and variance 9
    // ridgeFrequencyInterpolateW 7
    // ridgeFrequencyInterpolateMean 0
    // ridgeFrequencyInterpolateVariance 9
    float mean = ridgeFrequencyInterpolateMeanLeung;
    float var = ridgeFrequencyInterpolateVarianceLeung;
    int WgHalf = (ridgeFrequencyInterpolateWLeung - 1)/ 2;
    float x, y;
    float sumWg = 0;
    
    float **Wg = MemUtil::create2DMatrix<float>(ridgeFrequencyInterpolateWLeung,ridgeFrequencyInterpolateWLeung);
    for (int i = 0; i < ridgeFrequencyInterpolateWLeung; i++) {
        for (int j = 0; j < ridgeFrequencyInterpolateWLeung; j++) {
            x = i - WgHalf;
            y = j - WgHalf;
            Wg[i][j] = 1.0 / (2. * PI * var) * 
                    exp( -((x - mean) * (x - mean) + (y - mean) * (y - mean)) / (2. * var) );//
            sumWg += Wg[i][j];
        }
    }
    for (int i = 0; i < ridgeFrequencyInterpolateWLeung; i++) 
        for (int j = 0; j < ridgeFrequencyInterpolateWLeung; j++) 
            Wg[i][j] /= sumWg;
    
    int nElements = 0;
    float muSum, deltaSum;
    float mu, delta;
    int iW, jW;
    bool allFrequenciesOK = true;
    
    
    do {
        allFrequenciesOK = true;
        for (int i = 0; i < orH ; i++) {
            for (int j = 0; j < orW ; j++) {
                ridgeFrequencyInterpolated[i][j] = ridgeFrequency[i][j];
                float sumGauss = 0;
                if ( ridgeFrequencyInterpolated[i][j] == -1) {
                    nElements = 0;
                    muSum = 0; deltaSum = 0;
                    iW = i - WgHalf < 0 ? abs(i - WgHalf) : 0;
                    for (int u = max(i - WgHalf, 0); u < min(i + WgHalf + 1, orH); u++, iW++) {
                        jW = j - WgHalf < 0 ? abs(j - WgHalf) : 0;
                        for (int v = max(j - WgHalf, 0); v < min(j + WgHalf + 1, orW); v++, jW++) {
                            mu = ridgeFrequency[u][v] <= 0 ? 0 : ridgeFrequency[u][v];
                            delta = ridgeFrequency[u][v] + 1 <= 0 ? 0 : 1;

                            muSum += Wg[iW][jW] * mu;
                            deltaSum += Wg[iW][jW] * delta;
                            nElements++;
                            sumGauss += Wg[iW][jW];
                        }
                    }
                    
                   if ((deltaSum == 0) || (muSum == 0))
                        ridgeFrequencyInterpolated[i][j] = -1;
                    else {
                       muSum /= sumGauss;
                       deltaSum /= sumGauss;
                        ridgeFrequencyInterpolated[i][j] = muSum / deltaSum;
                    }
                    if (!((ridgeFrequencyInterpolated[i][j] <= ridgeFrequencyMinLeung) && (ridgeFrequencyInterpolated[i][j] >= ridgeFrequencyMaxLeung))) // for 500dpi images (nPeaks > 1) && 
                    {
                        ridgeFrequencyInterpolated[i][j] = -1;
                        allFrequenciesOK = false;
                    }
                }

            }
        }
        float **aux;
        aux = ridgeFrequencyInterpolated;
        ridgeFrequencyInterpolated = ridgeFrequency;
        ridgeFrequency = aux;
    } while (!allFrequenciesOK);
    
    MemUtil::freeMatrix(&Wg);
    
    /* Smooth blockVariance 5 x 5*/
    int WvarHalf = (blockVarianceWLeung - 1) / 2;
    float sumWvar = 0;
    float **Wvar = MemUtil::create2DMatrix<float>(blockVarianceWLeung,blockVarianceWLeung);
    // Wvar initialization
    for (int i = 0; i < blockVarianceWLeung; i++) {
        for (int j = 0; j < blockVarianceWLeung; j++) {
            x = i - WvarHalf;
            y = j - WvarHalf;
            Wvar[i][j] = 1.0 / (2. * PI * var) * 
                    exp( -((x - mean) * (x - mean) + (y - mean) * (y - mean)) / (2. * var) );//
            sumWvar += Wvar[i][j];
        }
    }
    for (int i = 0; i < blockVarianceWLeung; i++) 
        for (int j = 0; j < blockVarianceWLeung; j++) 
            Wvar[i][j] /= sumWvar;
    
    for (int i = 0; i < orH ; i++) {
        for (int j = 0; j < orW ; j++) {
            float sumGauss = 0;
            iW = i - WvarHalf < 0 ? abs(i - WvarHalf) : 0;
            for (int u = max(i - WvarHalf, 0); u < min(i + WvarHalf + 1, orH); u++, iW++) {
                jW = j - WgHalf < 0 ? abs(j - WvarHalf) : 0;
                for (int v = max(j - WvarHalf, 0); v < min(j + WvarHalf + 1, orW); v++, jW++) {
                    blockVarianceSmooth[i][j] += Wvar[iW][jW] * blockVariance[u][v];
                    sumGauss += Wvar[iW][jW];
                }
            }
            blockVarianceSmooth[i][j] /= sumGauss;
        }
    }

    MemUtil::freeMatrix(&Wvar);
    
    /* Smooth ridgeFrequency 7x7 and  */
    //ridgeFrequencySmoothW 7
    int WlHalf = (ridgeFrequencySmoothWLeung - 1)/ 2;
    nElements = 0;
    
    for (int i = 0; i < orH ; i++) {
        for (int j = 0; j < orW ; j++) {
            nElements = 0;
            float avg = 0;
            for (int u = max(i - WlHalf, 0); u < min(i + WlHalf + 1, orH); u++) {
                for (int v = max(j - WlHalf, 0); v < min(j + WlHalf + 1, orW); v++) {
                    avg += ridgeFrequency[u][v];
                    nElements++;
                }
            }
            avg /= (float) nElements;
            ridgeFrequencyInterpolated[i][j] = avg;
        }
    }
    
    MemUtil::freeMatrix(&ridgeFrequency);
    MemUtil::freeMatrix(&blockVariance);
    
    ridgeFrequency = ridgeFrequencyInterpolated;
    blockVariance = blockVarianceSmooth;
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "ridgeFrequency";
        ImgUtil::writeGrayScaleImage(ridgeFrequency, orH, orW, fileName2, 1*250 *3);
        fileName2 = "blockVarianceHong";
        ImgUtil::writeGrayScaleImage(blockVariance, orH, orW, fileName2, 1);
    }
}

float ImageEnhancementHong::computeRidgeFrequencyFromSignature(float *X, int l, float **tabla, int i, int j) {
    float min = 1000; 
    float max = -1000;
    int minPos = -1; 
    int maxPos = -1;
    bool lookformax = 1;
    float delta = 7.0;
    int firstPeak = -1;
    int lastPeak = -1;
    int nPeaksMax =0, nPeaksMin = 0;
    float totalMin = 0;
    float totalMax = 0;
    
    for (int k = 0; k < l; k++) {        
        float val = X[k];
        
        if (val > max)
            max = val;
        if (val < min)
            min = val;
    }
    for (int k = 0; k < l; k++) {        
        if (X[k] == 0)
            X[k] = (max - min) / 2;
    }
    
    delta = (max - min) * 0.6;
    max = -1000;
    min = 1000;
    
    for (int k = 0; k < l; k++) { 
        float val = X[k];

        if (val > max) {
            max = val; 
            maxPos = k;
        }
        if (val < min) {
            min = val; 
            minPos = k;
        }

        if (lookformax) {
//cout << "val = " << val << " max = " << max << " delta = " << delta << endl;
            if (val < max - delta) {
//cout << "Pico en " << maxPos << " valor " << max <<  endl; 
                min = val;
                totalMax += max;
//                totalMin += min;
                minPos = k;
                lookformax = 0;
                if (firstPeak == -1)
                    firstPeak = maxPos;
                else
                    lastPeak = maxPos;
                nPeaksMax++;
            }  
        }
        else { 
            if (val > min - delta) {
//cout << "Pico bajo en " << minPos << " valor " << min << endl; 
                max = val;
//                totalMax += max;
                totalMin += min;
                maxPos = k;
                lookformax = 1;
                nPeaksMin++;
            }
        }

    }

    if ((nPeaksMax == 0) || (nPeaksMin == 0))
        tabla[i][j] = 0;
    else
        tabla[i][j] = ((totalMax/(float)nPeaksMax) - (totalMin/(float)nPeaksMin));
//    
//    if (tabla[i][j] > 10) {
//        cout << "tabla[i][j] = " << tabla[i][j] << endl;
//        cout << "totalMax = " << totalMax << " totalMin = " << totalMin << " nPMax = " << nPeaksMax <<
//                " nPMin = " << nPeaksMin << endl;
//        for (int k = 0; k < l; k++)
//            cout << X[k] << " ";
//        cout << endl;
//    }
    
    float ridgeF = 1.0 / ( (lastPeak - firstPeak) / (float)(nPeaksMax - 1));
    return ridgeF;
}




float **ImageEnhancementHong::filterImage(int Wg, float sigmaX, float sigmaY, bool binarizar) {
    
    float **filteredImg = MemUtil::create2DMatrix<float>(h, w);
    float **filteredImg2 = MemUtil::create2DMatrix<float>(h, w);
    
    
    
    float **maxBlock = MemUtil::create2DMatrix<float>(orH, orW, -1000000);
    float **minBlock = MemUtil::create2DMatrix<float>(orH, orW, 1000000);
    float **meanBlock = MemUtil::create2DMatrix<float>(orH, orW);
    float **meanValueBlock = MemUtil::create2DMatrix<float>(orH, orW);
    float **tamBlock = MemUtil::create2DMatrix<float>(orH, orW);
    
    float **gaborFactors = MemUtil::create2DMatrix<float>(Wg,Wg);
    
    int WgHalf = (Wg - 1) / 2;
    int p, q;
    float sumG = 0;
//    float max = -1000000;
//    float min = 1000000;
    if (masc == 0)
        masc = MemUtil::create2DMatrix<bool>(orH, orW, true);

    bool **aux = ImgUtil::imFill(masc, orH, orW, 3);
    MemUtil::freeMatrix(&masc);
    masc = aux;

    for (int i = WgHalf; i < h - WgHalf ; i++) {
        for (int j = WgHalf; j < w - WgHalf ; j++) {
            p = i / blockW;
            q = j / blockW;

            if  ((p < orH) && (q < orW)){
                if ((masc[p][q]) || ((binarizar))){
                    if ( (i % blockW == 0) || (j % blockW == 0)) { // all pixels in a block shares the same gabor filter
                        for (int u = - WgHalf; u < WgHalf + 1; u++) {
                            for (int v = - WgHalf; v < WgHalf + 1; v++) 
                                gaborFactors[u + WgHalf][v + WgHalf] = 
                                        computeGaborFactor(u, v, O[p][q], ridgeFrequency[p][q], sigmaX, sigmaY);
                        }
                    }

                    sumG = 0;
                    float sumAux = 0;
                    for (int u = i - WgHalf; u < i + WgHalf + 1; u++) {
                        for (int v = j - WgHalf; v < j + WgHalf + 1; v++) {
                            filteredImg[i][j] += gaborFactors[u - i + WgHalf][v - j + WgHalf] * img[u][v];
                            sumAux +=gaborFactors[u - i + WgHalf][v - j + WgHalf];
                        }
                    }

                    if (filteredImg[i][j] > maxBlock[p][q])
                        maxBlock[p][q] = filteredImg[i][j];
                    else if (filteredImg[i][j] < minBlock[p][q])
                        minBlock[p][q] = filteredImg[i][j];

                    meanBlock[p][q] += filteredImg[i][j];
                    meanValueBlock[p][q] += imgCopy[i][j];
                    tamBlock[p][q]++;
    //
    //                if (filteredImg[i][j] > max)
    //                    max = filteredImg[i][j];
    //                else if (filteredImg[i][j] < min)
    //                    min = filteredImg[i][j];
                }else
                    tamBlock[p][q] = 1; //Evito división por cero.
            }
        }
    }     
    
    // Normalize each block
    for (int i = 0; i < h ; i++) {
        for (int j = 0; j < w; j++) {
            p = i / blockW;
            q = j / blockW;

            if  ((p < orH) && (q < orW)){
                
                if (masc[p][q] && binarizar){
                    if (maxBlock[p][q] -  minBlock[p][q] < 0.1) 
                        filteredImg[i][j] = 255;
                    else if (filteredImg[i][j] > (meanBlock[p][q]) / tamBlock[p][q] ){
                        filteredImg2[i][j] = 255;
                        filteredImg[i][j] = 255;
//                        if (mascPrueba[p][q] == 2)
//                            filteredImg2[i][j] = 170;
//                        else if (mascPrueba[p][q] == 3)
//                            filteredImg2[i][j] = 70;
//                        else if (mascPrueba[p][q] == 4)
//                            filteredImg2[i][j] = 128;
//                        else if (mascPrueba[p][q] == 5)
//                            filteredImg2[i][j] = 255;
                    }
                    else {//if (filteredImg[i][j] <= meanBlock[p][q] / tamBlock[p][q]){
//                        filteredImg2[i][j] = 0;
                        filteredImg[i][j] = 0; //img[i][j];
                    }
//                    if (mascPrueba[p][q] == 3)
//                        filteredImg[i][j] = (imgCopy[i][j] >= (meanValueBlock[p][q] / tamBlock[p][q])) ? 255 : 0;
                }
                else if (!(masc[p][q]) && binarizar){
                        filteredImg[i][j] = 255;
//                    if (mascPrueba[p][q] == 1){
//                        filteredImg2[i][j] = 255;
//                        filteredImg[i][j] = 255;
//                    }
//                    else if (mascPrueba[p][q] == 6){
////                        filteredImg[i][j] = 0;
//                        filteredImg2[i][j] = 20;
//                    }
//                    else                        
//                        filteredImg[i][j] = 255;
                }
                else{
                    if (maxBlock[p][q] -  minBlock[p][q] < 0.01) {
                        filteredImg[i][j] = 255;
                        filteredImg2[i][j] = 255;
                    }
                    else {
                        filteredImg[i][j] = (filteredImg[i][j] - minBlock[p][q]) / (maxBlock[p][q] - minBlock[p][q]) * 255.;
                        filteredImg2[i][j] = (filteredImg[i][j] - minBlock[p][q]) / (maxBlock[p][q] - minBlock[p][q]) * 255.;
                    }
                }
            }
            else
                filteredImg[i][j] = 255;
        }
    }
 
    bool **aux2 = ImgUtil::imErode(masc, orH, orW);
    MemUtil::freeMatrix(&masc);
    masc = aux2;
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "FilteredHong";
        string fileName3 = "SegmentedFilteredHong";
        ImgUtil::writeGrayScaleImage(filteredImg2, h, w, fileName2, 1);
        ImgUtil::writeGrayScaleImage(filteredImg, h, w, fileName3, 1);
    }
    
    if (TEST_2){
        string fileName = ImgUtil::makeTest3FileName("Segmentation");
        ImgUtil::writeGrayScaleImage(filteredImg, h, w, fileName,1);
    }
    
    MemUtil::freeMatrix(&maxBlock);
    MemUtil::freeMatrix(&minBlock);
    MemUtil::freeMatrix(&meanBlock);
    MemUtil::freeMatrix(&tamBlock);
    MemUtil::freeMatrix(&gaborFactors);

    MemUtil::freeMatrix(&meanValueBlock);
    MemUtil::freeMatrix(&filteredImg2);
    
    return filteredImg;
}

float ImageEnhancementHong::computeGaborFactor(int x, int y, float theta, float f, float sigmaX, float sigmaY) {
    float h = 0;
    
    float xprima = x * cos(theta) + y * sin(theta);
    float yprima = -x * sin(theta) + y * cos(theta);
    
    h = exp(-0.5 * (xprima * xprima / (sigmaX * sigmaX) + yprima * yprima / (sigmaY * sigmaY)  )) 
            * cos(2. * PI * f * xprima);
    
    return h;
}

void ImageEnhancementHong::varianceSegment(float threshold){
    
    if (masc == 0)
        masc = MemUtil::create2DMatrix<bool>(orH,orW);
    

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW ; j++) {
            if (blockVariance[i][j] >= threshold)
                masc[i][j] = true;
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "SegmentedHongV1";
        ImgUtil::writeGrayScaleImage(masc, orH, orW, fileName2);
    }
}

void ImageEnhancementHong::varianceSegment(float threshold, float **filteredImage){
    
    if (masc == 0 )
        masc = MemUtil::create2DMatrix<bool>(orH,orW);

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW ; j++) {
            if (blockVariance[i][j] >= threshold)
                masc[i][j] = true;
            else {
                for (int u = i * blockW; u < (i + 1) * blockW; u++)
                    for (int v = j * blockW; v < (j + 1) * blockW; v++)
                        filteredImage[u][v] = 255;
            }
        }
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "SegmentedHongV2";
        ImgUtil::writeGrayScaleImage(masc, orH, orW, fileName2);
    }
}

bool **ImageEnhancementHong::segmentImg(){
    
    float distance = 100000;
    float distanceAux;
    int clase = -1;
    int clases[6]={0,0,0,0,0,0};
    
    masc = MemUtil::create2DMatrix<bool>(orH, orW);
    mascPrueba = MemUtil::create2DMatrix<int>(orH, orW);

    for (int i = 0; i < orH; i++) {

        for (int j = 0; j < orW ; j++) {
            
            distance = 100000;
            
            //Guardar la clase con la mínima distancia a un cluster.
            for (int k = 1; k < 7; k++){
                distanceAux = calculateDistance(i, j, k);

                if (distanceAux < distance){
                    clase = k;
                    distance = distanceAux;
                }
            }
            clases[clase-1]++;            
            if (clusterLeungRealce[clase-1]){
                masc[i][j] = true;          
            }
            mascPrueba[i][j] = clase;
                
        }
    }
    

    if (TEST_2){
        
        ofstream f2;
        
        string fileName = ImgUtil::makeTest2FileName("ClasesUsadas");
        fileName.append(".txt");
        char *nombre = MemUtil::createVector<char>(fileName.size());
        strcpy(nombre, fileName.c_str());
        f2.open(nombre, ofstream::out | ofstream::trunc);

        for (int i = 0 ; i < 6; i++)
                f2 << clases[i] << " bloques de la clase " << i+1 << endl;
        
        f2.close();
        MemUtil::freeMatrix(&nombre);
    }
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "SegmentedHong";
        ImgUtil::writeGrayScaleImage(masc, orH, orW, fileName2);
    }

    return  masc;
}

float ImageEnhancementHong::calculateDistance(int i, int j, int clase){
    
    float varAux = (blockVarianceSeg[i][j] - minVar) / (maxVar - minVar);
    float avgAux = (averagePeaksValleysSeg[i][j] - minAvg) / (maxAvg - minAvg);
    float freqAux = (ridgeFrequencySeg[i][j] - minFreq) / (maxFreq - minFreq);    
//
//cout << "MinVar = " << minVar << " MaxVar = " << maxVar << endl;
//cout << "MinAvg = " << minAvg << " MaxAvg = " << maxAvg << endl;
//cout << "MinFreq = " << minFreq << " MaxFreq = " << maxFreq << endl;
//cout << "blockVarianceSeg["<<i<<"]["<<j<<"] = " << blockVarianceSeg[i][j] << endl; 
//cout << "averagePeaksValleysSeg["<<i<<"]["<<j<<"] = " << averagePeaksValleysSeg[i][j] << endl; 
//cout << "ridgeFrequencySeg["<<i<<"]["<<j<<"] = " << ridgeFrequencySeg[i][j] << endl; 
    
    switch(clase){
        case 1:{
            return sqrt((varAux - varC1LeungRealce)*(varAux - varC1LeungRealce)+
                         (avgAux - alphaC1LeungRealce)*(avgAux - alphaC1LeungRealce)+ 
                         (freqAux -betaC1LeungRealce)*(freqAux -betaC1LeungRealce));
        }
        case 2:{            
            return sqrt((varAux - varC2LeungRealce)*(varAux - varC2LeungRealce)+
                         (avgAux - alphaC2LeungRealce)*(avgAux - alphaC2LeungRealce)+ 
                         (freqAux -betaC2LeungRealce)*(freqAux -betaC2LeungRealce));
            
        }
        case 3:{            
            return sqrt((varAux - varC3LeungRealce)*(varAux - varC3LeungRealce)+
                         (avgAux - alphaC3LeungRealce)*(avgAux - alphaC3LeungRealce)+ 
                         (freqAux -betaC3LeungRealce)*(freqAux -betaC3LeungRealce));
            
        }
        case 4:{            
            return sqrt((varAux - varC4LeungRealce)*(varAux - varC4LeungRealce)+
                         (avgAux - alphaC4LeungRealce)*(avgAux - alphaC4LeungRealce)+ 
                         (freqAux -betaC4LeungRealce)*(freqAux -betaC4LeungRealce));
        }
        case 5:{            
            return sqrt((varAux - varC5LeungRealce)*(varAux - varC5LeungRealce)+
                         (avgAux - alphaC5LeungRealce)*(avgAux - alphaC5LeungRealce)+ 
                         (freqAux -betaC5LeungRealce)*(freqAux -betaC5LeungRealce));
        }
        default:{            
            return sqrt((varAux - varC6LeungRealce)*(varAux - varC6LeungRealce)+
                         (avgAux - alphaC6LeungRealce)*(avgAux - alphaC6LeungRealce)+ 
                         (freqAux -betaC6LeungRealce)*(freqAux -betaC6LeungRealce));
        }
    }  
}