/* 
 * \file        LinWangFE.cpp
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of LinWangFE class
 *
 * 
 */

#include <algorithm>

#include "LinWangFE.h"
#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "ComputeLabels.h"
#include "NewPoincareSingularPoints.h"
#include "ParametersPoincare.h"
#include "GHMFiltering.h"
#include "LinWangSegmentation.h"
#include "WangCoreDeltaDetector.h"
#include "FingerprintImg.h"

#include "ParametersLinWangFE.h"

LinWangFE::LinWangFE(){
    img = 0;

    h = 0;
    w = 0;

    orH = 0;
    orW = 0;
    blockW = 0;

    O = 0;
    xCos = 0;
    ySin = 0;

    X = 0;
    Xp = 0;

    cores.clear();
    deltas.clear();

    Poincare = 0;

    imgFiltered = 0;
    
    C=0;

    segmentation = 0;

    featureVector = 0;
    
    coreDeltasFound=false;
    
    initializeBlocks();
}

LinWangFE::LinWangFE(float **image, int h, int w) {

    img = MemUtil::copy2DMatrix<float>(image, h, w);

    this->h = h;
    this->w = w;

    orH = 0;
    orW = 0;
    blockW = 0;

    O = 0;
    xCos = 0;
    ySin = 0;

    X = 0;
    Xp = 0;

    cores.clear();
    deltas.clear();

    Poincare = 0;

    imgFiltered = 0;
    
    C=0;

    segmentation = 0;

    featureVector = 0;
    
    coreDeltasFound=false;
    
    initializeBlocks();
}

LinWangFE::LinWangFE(const LinWangFE& orig) {
}

LinWangFE::~LinWangFE() {

    MemUtil::freeMatrix(&img);
    MemUtil::freeMatrix(&imgFiltered);
    MemUtil::freeMatrix(&O);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&X);
    MemUtil::freeMatrix(&Xp);
    MemUtil::freeMatrix(&Op);
    MemUtil::freeMatrix(&xCosOp);
    MemUtil::freeMatrix(&ySinOp);
    MemUtil::freeMatrix(&segmentation);
    MemUtil::freeMatrix(&OrI);
    
    MemUtil::freeMatrix(&C);
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&energy);
}

void LinWangFE::initialize(float **imageOrig, int h, int w){
    this->h = h;
    this->w = w;

    img = MemUtil::copy2DMatrix<float>(imageOrig, h, w);
//    imgOrig = 0;

}

void LinWangFE::initializeBlocks(){
   
    if (((int)(blockSizeLinWangFE * ImgUtil::fAdjust)) % 2 == 1)
        block = blockSizeLinWangFE * ImgUtil::fAdjust - 1;
    else
        block = blockSizeLinWangFE * ImgUtil::fAdjust;
}

float *LinWangFE::run(string file, PCASYSManager *pcasys){
	
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){

        fi.setPCASYSManager(pcasys);    
        fi.readFingerprintImage(file);
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
        
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
       
    initializeBlocks();
    
    dimFV = 1;    
    float *featurevector = MemUtil::createVector<float>(1);
    featurevector[0] = computeFeatureVector();
    
    return featurevector;
}

int LinWangFE::GaussianFiltering(int gausFilterH, int gausFilterW) {

    int x, y;
    double total = 0;
    float **gaussianFilter = MemUtil::create2DMatrix<float>(gausFilterH, gausFilterW);
    //gaussianFilter initialization
    for (int i = 0; i < gausFilterH; i++) {
        for (int j = 0; j < gausFilterW; j++) {
            x = i - (gausFilterH / 2);
            y = j - (gausFilterW / 2);
            gaussianFilter[i][j] = exp(-(x * x + y * y) / (2 * sigmaLinWangFE * sigmaLinWangFE));
            total += gaussianFilter[i][j];
        }
    }
    
    //Normalization of gaussian filter
    for (int i = 0; i < gausFilterH; i++) {
        for (int j = 0; j < gausFilterW; j++) {
            x = i - (gausFilterH / 2);
            y = j - (gausFilterW / 2);
            gaussianFilter[i][j] = gaussianFilter[i][j] / total;
        }
    }

    imgFiltered = ImgUtil::getFilteredImg<float>(img, h, w, gaussianFilter, gausFilterH, gausFilterW);

    
    MemUtil::freeMatrix(&gaussianFilter);

    if (ImgUtil::DEBUG) {
        string fileName = "LinWangFEFilteredImage";
        ImgUtil::writeGrayScaleImage(imgFiltered, h, w, fileName, 1);
    }

    return 0;
}


int LinWangFE::createOrientationImageOverlapping(int blockW) {

    this->blockW = blockW / 2;
    orH = (float)h / (float)blockW * 2. - 1;
    orW = (float)w / (float)blockW * 2. - 1;
    float ang; //Angulo para theta y las orientaciones

    O = MemUtil::create2DMatrix<float>(orH, orW);
    xCos = MemUtil::create2DMatrix<float>(orH, orW);
    ySin = MemUtil::create2DMatrix<float>(orH, orW);

    float **Vx = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH, orW);
    float **Mag = MemUtil::create2DMatrix<float>(orH, orW);
    this->gradX = MemUtil::create2DMatrix<float>(h, w);
    this->gradY = MemUtil::create2DMatrix<float>(h, w);
    float **grad = MemUtil::create2DMatrix<float>(h, w);
    
    X = MemUtil::create2DMatrix<float>(orH, orW); //Coherence
    Xp = MemUtil::create2DMatrix<float>(orH, orW); //Final coherence  

    //GHMs
    this->energy = MemUtil::create2DMatrix<float>(h, w);
    float *maxmin=GHMFiltering::computeGHMImage(imgFiltered, h, w, gradX, gradY,energy,grad,ghmSupportHalfSizeLinWangFE,ghmSigmaLinWangFE,lambdaLinWangFE);
    
    //ImgUtil::writeGrayScaleImage(energy,h,w,"firstEnergy",255);  
    
    //Sobel
//    this->energy=MemUtil::create2DMatrix<float>(h, w);
    
    float **gradX2 = MemUtil::create2DMatrix<float>(h, w);
    float **gradY2 = MemUtil::create2DMatrix<float>(h, w);
    maxmin = ImgUtil::computeSobel(imgFiltered, h, w, gradX2, gradY2, grad); //Max and Min amplitude of the gradient
//    float maxEnergy=0;
//    for (int i=0;i<h; i++){
//        for (int j=0;j<w; j++){
//            this->energy[i][j]=(gradX[i][j]*gradX[i][j])+(gradY[i][j]*gradY[i][j]);
//            if (this->energy[i][j]>maxEnergy){
//                maxEnergy=this->energy[i][j];
//            }
//        }
//    }
//    for (int i=0;i<h; i++){
//        for (int j=0;j<w; j++){
//            this->energy[i][j]/=maxEnergy;
//        }
//    }
    

    if (ImgUtil::DEBUG) {
        cout << "Maximo: " << maxmin[0] << endl;
        cout << "Minimo: " << maxmin[1] << endl;
        cout << "Media: " << maxmin[2] << endl;
        string fileName1 = "LinWangFEGradientX";
        string fileName2 = "LinWangFEGradientY";
        string fileName3 = "LinWangFEGradient";
        ImgUtil::writeGrayScaleImage(gradX, h, w, fileName1, 1);
        ImgUtil::writeGrayScaleImage(gradY, h, w, fileName2, 1);
        ImgUtil::writeGrayScaleImage(grad, h, w, fileName3, 1);
    }

    float Gsx, Gsy;
    float Gth = thresholdFactorLinWangFE * (maxmin[2] * 2. - maxmin[1]) + maxmin[1]; // Calculate the gradient threshold we'll use to the coherence.
    float cst = (double) 1 / (blockW * blockW); // Constante usada en la coherencia.    

    MemUtil::freeMatrix(&maxmin);
    
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            for (int u = i * (blockW / 2); u < i * (blockW / 2) + blockW; u++) {
                for (int v = j * (blockW / 2); v < j * (blockW / 2) + blockW; v++) {
                    Gsx = 2.0 * gradX2[u][v] * gradY2[u][v];
                    Gsy = (gradX2[u][v] * gradX2[u][v] - gradY2[u][v] * gradY2[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;
                    Mag[i][j] += (gradX2[u][v] * gradX2[u][v] + gradY2[u][v] * gradY2[u][v]);
                }
            }

            //Calculates the block coherence
            if ((cst * Mag[i][j]) < (Gth * Gth))
                X[i][j] = -1;
            else {
                X[i][j] = sqrt((Vx[i][j] * Vx[i][j]) + (Vy[i][j] * Vy[i][j])) / Mag[i][j];
            }

            //Calculate theta and orientations
            ang = 0.5 * atan2(Vx[i][j], Vy[i][j]);
            if (ang <= 0)
                O[i][j] = ang + 0.5 * PI;
            else
                O[i][j] = ang - 0.5 * PI;

            xCos[i][j] = cos(2.0 * O[i][j]);
            ySin[i][j] = sin(2.0 * O[i][j]);
        }
    }

    if (ImgUtil::DEBUG) {
        string fileName1 = "LinWangFECoherence";
        ImgUtil::writeGrayScaleImage(X, orH, orW, fileName1, 255);
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, "LinWangFEOrientation");
    }

    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&Vy);
    MemUtil::freeMatrix(&Mag);
    MemUtil::freeMatrix(&gradX2);
    MemUtil::freeMatrix(&gradY2);
    MemUtil::freeMatrix(&grad);
    //MemUtil::freeMatrix(&energy);

    return 0;
}

int LinWangFE::computeCoherenceImage() {

    ComputeLabels compLabels;
    int **labelImg;
    int labelHuella, numComp;
    int **imgBin = ImgUtil::binarizeImg(X, orH, orW, 0); //Binarizo la coherencia
    segmentation = MemUtil::create2DMatrix<bool>(orH, orW);

    if (ImgUtil::DEBUG) {
        string fileName1 = "LinWangFECoherenceBin";
        ImgUtil::writeGrayScaleImage(imgBin, orH, orW, fileName1);
    }

    numComp = compLabels.calculateLabels(imgBin, orH, orW);
    MemUtil::freeMatrix(&imgBin);
    labelImg = compLabels.getLabeledImage();
    labelHuella = compLabels.getBiggestLabel();

    if (ImgUtil::DEBUG) {
        cout << "Numero de componentes conectadas: " << numComp << endl;
        cout << "Label que representa la huella: " << labelHuella << endl;
        string fileName1 = "LinWangFELabel";
        ImgUtil::writeGrayScaleImage(labelImg, orH, orW, fileName1);
    }

    //Creación de la coherencia alternativa X'
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (labelImg[i][j] == 0) {
                Xp[i][j] = -1;
                segmentation[i][j] = false;
            } else if (labelImg[i][j] == labelHuella) {
                Xp[i][j] = X[i][j];
                segmentation[i][j] = true;
            } else {
                Xp[i][j] = 0;
                segmentation[i][j] = false;
            }
        }
    }

    //    segmentation = ImgUtil::imDilate(segmentation, orH, orW);
    //    segmentation = ImgUtil::imErode(segmentation, orH, orW);

    if (ImgUtil::DEBUG) {
        string fileName1 = "LinWangFECoherence2";
        ImgUtil::writeGrayScaleImage(Xp, orH, orW, fileName1, 255);
    }
    return 0;
}

int LinWangFE::smoothOrientationImage() {

    /* Variables para calcular alpha, beta e imagen suavizada */
    float muSinT, muSin_, muCosT, muCos_;
    int muXp2T, muXpT, muXp2_, muXp_;

    /* Cambian para cada bloque */
    float alpha;
    float beta;

    Op = MemUtil::create2DMatrix<float>(orH, orW);
    xCosOp = MemUtil::create2DMatrix<float>(orH, orW);
    ySinOp = MemUtil::create2DMatrix<float>(orH, orW);

    int u, v, up, vp;
    float ang, ang2;
    for (int iBlock = 0; iBlock < orH; iBlock++) {
        for (int jBlock = 0; jBlock < orW; jBlock++) {
            alpha = beta = muSinT = muSin_ = muCosT = muCos_ = muXp2T = muXpT = muXp_ = muXp2_ = 0;
            ang2 = O[iBlock][jBlock];
            if (ang2 <= PI / 2.)
                ang = ang2 - 0.5 * PI;
            else
                ang = ang2 + 0.5 * PI;

            for (int k = 0; k < blockWLinWangFE; k++) {
                for (int d = 0; d < blockLLinWangFE; d++) {
                    //Calcular u y v (angulo normal)
                    u = iBlock + floor((d - blockLLinWangFE / 2) * cos(ang) + (k - blockWLinWangFE / 2) * sin(ang) + 0.5); // con floor(x + 0.5) conseguimos que si es un nº con decimal
                    v = jBlock + floor((d - blockLLinWangFE / 2) * sin(ang) + (blockWLinWangFE / 2 - k) * cos(ang) + 0.5); // <= .5 elija el mínimo y si es > .5 elija el máximo

                    //Calcular up y vp (angulo perpendicular)
                    up = iBlock + floor((d - blockLLinWangFE / 2) * cos(ang2) + (k - blockWLinWangFE / 2) * sin(ang2) + 0.5);
                    vp = jBlock + floor((d - blockLLinWangFE / 2) * sin(ang2) + (blockWLinWangFE / 2 - k) * cos(ang2) + 0.5);

                    if ((u >= 0) && (v >= 0) && (u < orH) && (v < orW)) {
                        muXp2T += ((Xp[u][v] * Xp[u][v]) > thresholdCLinWangFE) ? 1 : 0;
                        muXpT += (Xp[u][v] > thresholdCLinWangFE) ? 1 : 0;
                        muSinT += ((Xp[u][v] > thresholdCLinWangFE) ? 1 : 0) * sin(2.0 * O[u][v]);
                        muCosT += ((Xp[u][v] > thresholdCLinWangFE) ? 1 : 0) * cos(2.0 * O[u][v]);
                    }
                    if ((up >= 0) && (vp >= 0) && (up < orH) && (vp < orW)) {
                        muXp2_ += ((Xp[up][vp] * Xp[up][vp]) > thresholdCLinWangFE) ? 1 : 0;
                        muXp_ += (Xp[up][vp] > thresholdCLinWangFE) ? 1 : 0;
                        muSin_ += ((Xp[up][vp] > thresholdCLinWangFE) ? 1 : 0) * sin(2.0 * O[up][vp]);
                        muCos_ += ((Xp[up][vp] > thresholdCLinWangFE) ? 1 : 0) * cos(2.0 * O[up][vp]);
                    }
                }
            }

            //Calculamos alpha y beta teniendo en cuanta las divisones por cero.
            if ((muXpT != 0) && ((muXp2T / muXpT) != 0)) {
                if ((muXp_ != 0) && ((muXp2_ / muXp_) != 0)) {
                    alpha = (float) (muXp2T / muXpT) / (float) ((muXp2T / muXpT) + ((muXp2_ / muXp_)));
                    beta = (float) (muXp2_ / muXp_) / (float) ((muXp2T / muXpT) + ((muXp2_ / muXp_)));
                } else {
                    alpha = 1;
                    beta = 0;
                }
            } else {
                if ((muXp_ != 0) && ((muXp2_ / muXp_) != 0)) {
                    alpha = 0;
                    beta = 1;
                } else {
                    alpha = 0;
                    beta = 0;
                }
            }

            if ((alpha + beta) == 0) {
                Op[iBlock][jBlock] = O[iBlock][jBlock];
                xCosOp[iBlock][jBlock] = xCos[iBlock][jBlock];
                ySinOp[iBlock][jBlock] = ySin[iBlock][jBlock];
            } else {
                Op[iBlock][jBlock] = 0.5 * atan2((alpha * muSinT) + (beta * muSin_), (alpha * muCosT) + (beta * muCos_));
                xCosOp[iBlock][jBlock] = cos(2.0 * Op[iBlock][jBlock]);
                ySinOp[iBlock][jBlock] = sin(2.0 * Op[iBlock][jBlock]);
            }

        }
    }

    if (ImgUtil::DEBUG) {
        string fileName1 = "LinWangFESmoothOrientations";
        ImgUtil::writeOrImg(xCosOp, ySinOp, orW, orH, fileName1);
        ImgUtil::writeGrayScaleImage(segmentation, orH, orW, "LinWangFESegmentation");
    }
    
    bool **aux = ImgUtil::imErode(segmentation, orH, orW);
    MemUtil::freeMatrix(&segmentation);
    segmentation = aux;
}

int LinWangFE::computeCoresAndDeltas() {

    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(Op, xCosOp, ySinOp, orH, orW);

    npsp->setWidth(w); // Necesaria porque no hemos utilizado en el constructor la imagen real
    npsp->setHeight(h); // y para las reglas de karu nos hace falta.

    
    // use npsp segmentation (exclusively)
    //npsp->generateSegmentation(img, h, w);
    
    //
    //use zhang segmentation (exclusively)
    //
    //npsp->setBlockSegmentation(segmentation);
    
    //
    //use NPSP and zhang segmentation (exclusively)
    //
    //  npsp->generateSegmentation(img, h, w);
    //  npsp->combineWithSegmentation(segmentation);


    //
    //Use lin wang segmentation in combination with NPSP
    
    this->C=MemUtil::create2DMatrix<float>(h,w);
    GHMFiltering::computeCoherence(C,h,w,this->gradX,this->gradY,coherenceHalfWindowLinWangFE);
    LinWangSegmentation *lws= new LinWangSegmentation(this->h, this->w);
    lws->setEnergy(this->energy);
    lws->setCoherence(C);
    lws->createSegmentation();
    lws->createSmoothSegmentation(segmentationSmoothingWindow_LinWangFE);
    npsp->generateSegmentation(img, h, w);
    npsp->combineWithSegmentation(lws->getBlockwiseSmothSegmentation(this->blockW,true));
    
    
    //
    //Use lin wang segmentation (exclusively)
    //
//    this->C=MemUtil::create2DMatrix<float>(h,w);
//    GHMFiltering::computeCoherence(C,h,w,this->gradX,this->gradY,coherenceHalfWindowLinWangFE);
//    LinWangSegmentation *lws= new LinWangSegmentation(this->h, this->w);
//    lws->setEnergy(this->energy);
//    lws->setCoherence(C);
//    lws->createSegmentation();
//    lws->createSmoothSegmentation(segmentationSmoothingWindow_LinWangFE);
//    npsp->setBlockSegmentation(lws->getBlockwiseSmothSegmentation(this->blockW,true));
    
    if (TEST_1){
        string fileName = ImgUtil::makeTestFileName("_Segmentation_");
        ImgUtil::writeGrayScaleImage(npsp->getSegmentation(),orH,orW,fileName);
        string fileName2 = ImgUtil::makeTestFileName("__");
        string fileName3 = ImgUtil::makeTestFileName("_Segmentation_");
    }
    
    int nSmooth = npsp->computeCoresAndDeltas(curveWLinWangFE, smoothWPoincareLinWangFE, poincareModeLinWangFE, rulesToBeAppliedLinWangFE, minNumBlockLinWangFE, normalizeLinWangFE);
    int smoothSize = smoothWPoincareLinWangFE;
    while ((npsp->getnCores() > 2) || (npsp->getnDeltas() > 2)) {
        nSmooth = npsp->computeCoresAndDeltas(curveWLinWangFE, smoothSize, poincareModeLinWangFE, rulesToBeAppliedLinWangFE, minNumBlockLinWangFE, normalizeLinWangFE);

        if (nSmooth >= maxSmoothPoincare)
            smoothSize += 2;
        //cout << "SMOOTH increased " << smoothSize << endl;
    }

    cores = npsp->getCoreCentroids();
    deltas = npsp->getDeltaCentroids();
    if (ImgUtil::DEBUG) {
        if (npsp->getnCores() > 0 && cores.size() > 0) {
            cout << "Hay " << npsp->getnCores() << " cores" << endl;
            cout << cores[0] << ", " << cores[1] << endl;
            for (int i = 0; i < npsp->getnCores(); i++)
                if ((cores[2 * i] > 0) && (cores[2 * i + 1] > 0)) {
                    for (int j = 0; j < 5; j++)
                        for (int k = 0; k < 5; k++)
                            img [(int) ((block / 2) * cores[2 * i]) + j][(int) ((block / 2) * cores[2 * i + 1]) + k] = 255;
                }
        }
        if (npsp->getnDeltas() > 0 && deltas.size() > 0) {
            cout << "Hay " << npsp->getnDeltas() << " deltas" << endl;
            cout << deltas[0] << ", " << deltas[1] << endl;
            for (int i = 0; i < npsp->getnDeltas(); i++)
                if ((deltas[2 * i] > 0) && (deltas[2 * i + 1] > 0))
                    for (int j = 0; j < 5; j++)
                        for (int k = 0; k < 5; k++)
                            img [(int) ((block / 2) * deltas[2 * i]) + j][(int) ((block / 2) * deltas[2 * i + 1]) + k] = 0;

        }
        if ((npsp->getnCores() == 0) && (npsp->getnDeltas() == 0))
            cout << "No hay ningun core ni delta" << endl;

        ImgUtil::writeGrayScaleImage(img, h, w, "LinWangFEImgCoresDeltas", 1);
    }
    if (TEST_1){
        string fileName = ImgUtil::makeTestFileName("SingularPoints");
        ImgUtil::writeSPImage(img, h, w, deltas, cores, blockW, fileName); 
        fileName = ImgUtil::makeTestFileName("OrientationMap");
        ImgUtil::writeOrImg(xCosOp, ySinOp, orW, orH, fileName);
    }
    
    if ((npsp->getnCores() == 0) && (npsp->getnDeltas() == 0)){
        
        //Only in case C is not used before!
        if (this->C==0){
            this->C=MemUtil::create2DMatrix<float>(h,w);
            GHMFiltering::computeCoherence(C,h,w,this->gradX,this->gradY,coherenceHalfWindowLinWangFE);
        }
        WangCoreDeltaDetector *wcdd= new WangCoreDeltaDetector(C,Op, h,w, this->blockW,true);
        wcdd->setBlockwiseSegmentation(npsp->getSegmentation());
        wcdd->setKaruRulesForSPCleaning(true);
        
        wcdd->findCoreDeltas();
        
        this->coreDeltasFound=wcdd->foundCoreDelta();
    }
    
    delete lws;
    delete npsp;
}

int LinWangFE::analyzeCoreDelta() {

    //Leemos la posición del core.
    float x1 = cores[0] + 2;
    float y1 = cores[1]; //+0.5; //+1;

    //Leemos la posición del delta.
    float x2 = deltas[0] + 2;
    float y2 = deltas[1]; //+0.5;

    //    if (y1>y2) y1++;
    //    else if (y1<y2) y1--;

    float thetaK = 0;

    int row, col;

    float min = 1000000, sum;
    int Oc = 0;

    /* Creamos 16 bloques alrededor del core en 16 direcciones diferentes 
       para elegir la dirección con menor error respecto de la orientación
       real en la huella */

    float thetaKAux;
    for (int i = 0; i < 16; i++) {
        //   i = 8;
        thetaK = i * (PI / 16.) - PI / 2.;
        thetaKAux = i * (PI / 16.);
        if (thetaKAux > PI / 2.)
            thetaKAux -= PI;
        sum = 0;
        for (int u = 0; u < blockArroundCoreLLinWangFE; u++) {
            for (int v = 0; v < blockArroundCoreWLinWangFE; v++) {
                row = round(x1 + (((float) u) * cos(thetaK) + ((float) v - blockArroundCoreWLinWangFE / 2.) * sin(thetaK)));
                col = round(y1 + (((float) u) * sin(thetaK) + ((float) blockArroundCoreWLinWangFE / 2. - (float) v) * cos(thetaK)));
                if ((row >= 0) && (col >= 0) && (row < orH) && (col < orW))
                    sum += abs(sin(this->Op[row][col] - thetaKAux));
                //cout << Op[row][col] * 180. / PI << " ";
                // xCosOp[row][col] = 0;
                // ySinOp[row][col] = 0;
            }
        }
        if (ImgUtil::DEBUG) {
            cout << "i = " << i << " sum = " << sum << endl;
        }
        // ImgUtil::writeOrImg(xCosOp,ySinOp,orW,orH,"LinWangFEPrueba");
        //  return 0;
        if (sum < min) {
            min = sum;
            Oc = i;
        }
    }

    float ang = atan2(x1 - x2, y2 - y1);
    //cout << "angulo: " << ang*180./PI << endl;
    //ang = atan2(y2-y1,x2-x1);
    //cout << "angulo v2: " << ang*180./PI << endl;
    float Ocd = ang; //ang /(PI / 16.) + 8;
    float Ocf = Oc * (PI / 16.) - PI;


    float deltaO = Ocf - Ocd; // Indice que indica el ángulo que corresponde (deltaO * (PI / 16.))

    if (ImgUtil::DEBUG) {
        cout << "Direccion del core: " << Ocf * 180. / PI << endl;
        cout << "Direccion core-delta: " << Ocd * 180. / PI << endl;
        cout << "Angulo formado: " << deltaO * 180. / PI << endl;
    }


    if ((abs(deltaO)) < (PI / 12.)) {
        //cout << "T" << endl;
        return 2; // Tented-arch
    } else if (deltaO < 0) {
        //cout << "L" << endl;
        return 4; // Left loop
    } else {
        //cout << "R" << endl;
        return 3; // Right loop
    }

}

int LinWangFE::findTurningPoint(vector<Point> trace) {
    
    int d, turns, i;
    turns = 0;
    int turning = -1, dif;
    d = 4;
    i = 0;
    turns = 0;
    do {
        do {
            for (d = 4; i + d + 4 < trace.size(); d++) {
                dif = fabs(angBtwn2Points(trace[i], trace[i + 4]) - angBtwn2Points(trace[i + d], trace[i + d + 4]));
                if (dif > PI) 
                    dif = 2 * PI - dif;
                if (dif >= 5. * PI / 6.) {
                    turns++;
                    //turning = i + d; 
                    turning = i + d;//(int) floor(i + (d + 4) / 2);
                    i = i + d;
                    d = 4;
                    break;
                }
            }
        } while ((i + d + 4 < trace.size()) && (turns == 0)); // && (turns == 0));        
        i++;
    } while ( (i + 8 < trace.size())  && (turns == 0));
    
    return turning;
}

float LinWangFE::computePseudoRidge(vector<float> cores, vector<float> deltas) {

    Point delta, core;
    vector<Point> trace;
    int t, s;

 //  cout << "go to computePseudo" << endl;
//    cout << "deltas size = " << deltas.size() << endl;
//    cout << "cores size = " << cores.size() << endl;
    if ((cores.size() == 0)) { // || (cores.size() > 0)) { //There isn't a core  (cores.size() == 0) {

//    cout << "deltas size = " << deltas.size() << endl;
        int pseudoRidgeSize = pseudoRidgeSizeLinWangFE * ImgUtil::fAdjust * 2;
        vector<Point> trace1, trace2, trace3, trace4;
        int t1, t2, t3, t4;
        int tMax = 0;
        delta.x = deltas[1] + 2;
        delta.y = deltas[0] - 2;
        trace1 = Tracepseudoridge(delta, pseudoRidgeSize);
        t1 = traceturns(trace1);
//        cout << "pseudoRidge 1 turns = " << t1 << endl;
//        cout << "pseudoRidge 1 trace.size = " << trace1.size() << endl;
//        
//        cout << "(t1 > 0) = " << (t1 > 0) << endl;
//        cout << "(trace1.size() > tMax) = " << (trace1.size() > tMax) << endl;
//        cout << "trace1.size() = " << trace1.size() << " tMax = " << tMax << endl;
        if ((t1 > 0) && (trace1.size() > tMax)) {
//            cout << "entro t1" << endl;
            tMax = trace1.size();
            trace = trace1;
            t = t1;
        }
        
        delta.x = deltas[1] - 2;
        delta.y = deltas[0] - 2;
        trace2 = Tracepseudoridge(delta, pseudoRidgeSize);
        t2 = traceturns(trace2);
//        cout << "pseudoRidge 2 turns = " << t2 << endl;
//        cout << "pseudoRidge 2 trace.size = " << trace2.size() << endl;
        if ((t2 > 0) && (trace2.size() > tMax)) {
//            cout << "entro t2" << endl;
            tMax = trace2.size();
            trace = trace2;
            t = t2;
        }
        
        delta.x = deltas[1] - 2;
        delta.y = deltas[0] + 2;
        trace3 = Tracepseudoridge(delta, pseudoRidgeSize);
        t3 = traceturns(trace3);
//        cout << "pseudoRidge 3 turns = " << t3 << endl;
//        cout << "pseudoRidge 3 trace.size = " << trace3.size() << endl;
        if ((t3 > 0) && (trace3.size() > tMax)) {
//            cout << "entro t3" << endl;
            tMax = trace3.size();
            trace = trace3;
            t = t3;
        }
        
        
        delta.x = deltas[1] + 2;
        delta.y = deltas[0] + 2;
        trace4 = Tracepseudoridge(delta, pseudoRidgeSize);
        t4 = traceturns(trace4);
//        cout << "pseudoRidge 4 turns = " << t4 << endl;
//        cout << "pseudoRidge 4 trace.size = " << trace4.size() << endl;
        if ((t4 > 0) && (trace4.size() > tMax)) {
//            cout << "entro t4" << endl;
            tMax = trace4.size();
            trace = trace4;
            t = t4;
        }
                

        
        if (ImgUtil::DEBUG) {
            int **M;
            vector<Point> traceN;
            traceN.insert(traceN.end(), trace1.begin(), trace1.end());
            traceN.insert(traceN.end(), trace2.begin(), trace2.end());
            traceN.insert(traceN.end(), trace3.begin(), trace3.end());
            traceN.insert(traceN.end(), trace4.begin(), trace4.end());
            
            M = LocationMatrix2(traceN);
            writeImageAndTrace2(img, M, "TraceFromDeltaPoints");
        }
        
        int turning = findTurningPoint(trace);
        
        reverse(trace.begin(), trace.end());
        
        int turning2 = findTurningPoint(trace);
        
        int turnPoint = (turning + turning2) / 2;
        if (turning2 == -1)
            turnPoint = turning;
        
        if ((trace.size() == 0) || (t == 0) || (turning == -1) || (turning2 == -1))  {
            //cout << "-1" << endl;
            return -1;
        }
        
//        cout << "turning point: " << trace[turning].x << ", " <<  trace[turning].y << endl;
//        cout << "turning point block: " << (trace[turning].x + blockW / 2) / blockW << ", " <<  (trace[turning].y + blockW / 2) / blockW << endl;
//        
//        cout << "turning = " << turning << " turning2 = " << turning2 << endl;
        core.x = (trace[turnPoint].x + blockW / 2) / blockW; // Used in computeS when there is not core detected
        core.y = (trace[turnPoint].y + blockW / 2) / blockW;

        if (ImgUtil::DEBUG) {
            for (int j = 0; j < 5; j++)
                for (int k = 0; k < 5; k++)
                    img [(int) ((block / 2) * core.y) + j][(int) ((block / 2) * core.x) + k] = 0;

            ImgUtil::writeGrayScaleImage(img, h, w, "LinWangFEImgCoresDeltasPR", 1);
        }
            
    } else { //There is one core

//        cout << " tengo un core " << endl;
        core.x = cores[1];
        core.y = cores[0];
    }

    
    trace = Tracepseudoridge(core, pseudoRidgeSizeLinWangFE * ImgUtil::fAdjust * 2);
    t = traceturns(trace);
    if (t == 0) {
        if (ImgUtil::DEBUG)
            cout << "No hay turns" << endl;
        //cout << "T" << endl;
        return 2;
    } else if (t == 1) {
        s = computeS(trace, core);
        // cout << "s = " << s << endl;
        if (s > thresholdTentedArchLinWangFE * ImgUtil::fAdjust) {
            //cout << "L" << endl;
            return 4;
        } else if (s < -thresholdTentedArchLinWangFE * ImgUtil::fAdjust) {
            //cout << "R" << endl;
            return 3;
        } else {
            ///cout << "T" << endl;
            return 2;
        }

    } else {
        //cout << "W" << endl;
        return 0;
    }


    return -1;
}

void LinWangFE::Orientation2Index() {

    OrI = MemUtil::create2DMatrix<int>(orH, orW);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (Op[i][j] >= 0)
                OrI[i][j] = (int) round(4 * Op[i][j] / (PI / 2));
            else
                OrI[i][j] = 8 + (int) round(4 * Op[i][j] / (PI / 2));
        }
    }

    // Crear la matriz de orientaciones a nivel de pixel
    int Ws = blockW;
    OrIp = MemUtil::create2DMatrix<int>(h, w);
    for (int y = 0; y < orH; y++) {
        for (int x = 0; x < orW; x++) {
            for (int i = y * Ws; i < (y + 1) * Ws; i++) {
                for (int j = x * Ws; j < (x + 1) * Ws; j++) {
                    OrIp[i][j] = OrI[y][x];
                }
            }
        }
    }

    smaskp = MemUtil::create2DMatrix<bool>(h, w);
    for (int y = 0; y < orH; y++) {
        for (int x = 0; x < orW; x++) {
            for (int i = y * Ws; i < (y + 1) * Ws; i++) {
                for (int j = x * Ws; j < (x + 1) * Ws; j++) {
                    smaskp[i][j] = segmentation[y][x];
                }
            }
        }
    }
}

vector<Point> LinWangFE::Tracepseudoridge(Point start, int pseudoRidgeSize) {

    vector<Point> trace;
    int i, avi, avj, dif, dirb, dir;
    Point p, aux;
    int **M;
    bool border, crossing;

    // Push the start position in the vector

    start.x = start.x * blockW + blockW / 2;
    start.y = start.y * blockW + blockW / 2;

    trace.push_back(start);
    i = 0;
    // First trace in the positive direction (0-8 index)
    dir = OrIp[trace[i].y][trace[i].x];
    avi = getimove(dir);
    avj = getjmove(dir);

    p.x = trace[i].x + (int) floor((float) avj / 2);
    p.y = trace[i].y + (int) floor((float) avi / 2);
    trace.push_back(p);
    i++;
    p.x = start.x + avj;
    p.y = start.y + avi;
    trace.push_back(p);
    i++;
    border = false;
    do {
        // cout << "trace " << trace[i-1].y << ", " << trace[i-1].x << " dir = " << dir << " OrIP = " << OrIp[trace[i-1].y][trace[i-1].x] << endl;
        dirb = dir;
        dir = OrIp[trace[i].y][trace[i].x];
        // trace in the positive direction (0-8 index)
        // Tracing step is 2
        dif = abs(dir - dirb);
        if (dif > 4)
            dir = (dir + 8) % 16;

        avi = getimove(dir);
        avj = getjmove(dir);

        if ((trace[i].x + avj > 0) && (trace[i].x + avj < w) && (trace[i].y + avi < h) && (trace[i].y + avi > 0)) {
            if ( ((trace[i].y + avi * 2 * blockW) > 0) &&
                    ((trace[i].y + avi * 2 * blockW) < h) &&
                    ((trace[i].x + avj * 2 * blockW) > 0) &&
                    ((trace[i].x + avj * 2 * blockW) < w) &&
                    (smaskp[(trace[i].y + avi)][(trace[i].x + avj)]) && 
                    (smaskp[(trace[i].y + avi * 2 * blockW)][(trace[i].x + avj * 2 * blockW)])) {
                crossing = false;
                int possibleCross = 0;
                for (int z = 0; z < trace.size() - 2; z++) {
                    if ( ((abs(trace[z].x - (trace[i].x + avj)) == 0) && (abs(trace[z].y - (trace[i].y + avi))) == 0) ) {
                        crossing = true;
                    }
                    else if ( ((abs(trace[z].x - (trace[i].x + avj)) == 1) && (abs(trace[z].y - (trace[i].y + avi))) == 1) && 
                            ((abs(trace[z].x - (trace[i].x)) == 1) && (abs(trace[z].y - (trace[i].y))) == 1)) {
                        possibleCross++;
                    }

                }
                if (possibleCross > 2)
                    crossing = true;
               // if (crossing == false) {
                    p.x = trace[i].x + (int) floor((float) avj / 2);
                    p.y = trace[i].y + (int) floor((float) avi / 2);
                    trace.push_back(p);
                    i++;
                    p.x = trace[i - 1].x + avj;
                    p.y = trace[i - 1].y + avi;
                    trace.push_back(p);
                    i++;
             //   }
            } else {
                border = true;
            }
        } else {
            border = true;
        }



    } while ((trace.size() < pseudoRidgeSize * 2) && (border == false) && (crossing == false));
    if (ImgUtil::DEBUG)
        cout << "trace.size() = " << trace.size() << " border = " << border << " crossing = " << crossing << endl;
    
    // Swap the trace
    for (int j = 0; j < trace.size() / 2; j++) {
        aux = trace[j];
        trace[j] = trace[trace.size() - j - 1];
        trace[trace.size() - j - 1] = aux;
    }

    dir = OrIp[trace[i].y][trace[i].x] + 8;
    avi = getimove(dir);
    avj = getjmove(dir);

    p.x = trace[i].x + (int) floor((float) avj / 2);
    p.y = trace[i].y + (int) floor((float) avi / 2);
    trace.push_back(p);
    i++;

    p.x = trace[i - 1].x + avj;
    p.y = trace[i - 1].y + avi;
    trace.push_back(p);
    i++;
    border = false;
    do {
        // trace in the positive direction (0-8 index)
        // Tracing step is 2
        dirb = dir;
        dir = OrIp[trace[i].y][trace[i].x];
        // trace in the positive direction (0-8 index)
        // Tracing step is 2
        dif = abs(dir - dirb);
        if (dif > 4)
            dir = (dir + 8) % 16;


        avi = getimove(dir);
        avj = getjmove(dir);
        // Insert into the trace the mid point
        if ((trace[i].x + avj > 0) && (trace[i].x + avj < w) && (trace[i].y + avi < h) && (trace[i].y + avi > 0)) {
            if ( ((trace[i].y + avi * 2 * blockW) > 0) &&
                    ((trace[i].y + avi * 2 * blockW) < h) &&
                    ((trace[i].x + avj * 2 * blockW) > 0) &&
                    ((trace[i].x + avj * 2 * blockW) < w) &&
                    (smaskp[(trace[i].y + avi)][(trace[i].x + avj)]) 
                    && (smaskp[(trace[i].y + avi * 2 * blockW)][(trace[i].x + avj * 2 * blockW)])) {
                int possibleCross = 0;
                for (int z = 0; z < trace.size() - 2; z++) {
                    if ( ((abs(trace[z].x - (trace[i].x + avj)) == 0) && (abs(trace[z].y - (trace[i].y + avi))) == 0) ) {
                        crossing = true;
                    }
                    else if ( ((abs(trace[z].x - (trace[i].x + avj)) == 1) && (abs(trace[z].y - (trace[i].y + avi))) == 1) && 
                            ((abs(trace[z].x - (trace[i].x)) == 1) && (abs(trace[z].y - (trace[i].y))) == 1)) {
                        possibleCross++;
                    }

                }
                if (possibleCross > 2)
                    crossing = true;
              //  if (crossing == false) {
                    p.x = trace[i].x + (int) floor((float) avj / 2);
                    p.y = trace[i].y + (int) floor((float) avi / 2);
                    trace.push_back(p);
                    i++;
                    p.x = trace[i - 1].x + avj;
                    p.y = trace[i - 1].y + avi;
                    trace.push_back(p);
                    i++;
             //   }
            } else {
                border = true;
            }
        } else {
            border = true;
        }



    } while ((trace.size() < pseudoRidgeSize * 4) && (border == false) && (crossing == false));
    if (ImgUtil::DEBUG)
        cout << "trace.size() = " << trace.size() << " border = " << border << " crossing = " << crossing << endl;

    if (ImgUtil::DEBUG) {
        string fileName = "ImageAndTrace";
        cout << "Tracesize: " << trace.size() << endl;
        M = LocationMatrix2(trace);
        writeImageAndTrace2(img, M, fileName);
    }

    return trace;
}

int LinWangFE::getimove(int index) {

    if ((index == 2) || (index == 3) || (index == 4) || (index == 5) || (index == 6))
        return -2;
    else if ((index == 10) || (index == 11) || (index == 12) || (index == 13) || (index == 14))
        return 2;
    else if ((index == 7) || (index == 1))
        return -1;
    else if ((index == 9) || (index == 15))
        return 1;
    else
        return 0;
}

int LinWangFE::getjmove(int index) {

    if ((index == 6) || (index == 7) || (index == 8) || (index == 9) || (index == 10))
        return -2;
    else if ((index == 2) || (index == 1) || (index == 0) || (index == 15) || (index == 14))
        return 2;
    else if ((index == 5) || (index == 11))
        return -1;
    else if ((index == 3) || (index == 13))
        return 1;
    else
        return 0;
}

int LinWangFE::traceturns(vector<Point> trace) {

    int i, d, turns;
    float dif;
    d = 4;
    i = 0;
    turns = 0;
   // cout << "trace size = " << trace.size() << endl;
    do {
        do {
            for (d = 4; i + d + 4 < trace.size(); d++) {
               // cout << "ang1 = " << angBtwn2Points(trace[i], trace[i + 4]) * 180. / PI << " ang2 = " << angBtwn2Points(trace[i + d], trace[i + d + 4]) * 180. / PI << endl;
                dif = fabs(angBtwn2Points(trace[i], trace[i + 4]) - angBtwn2Points(trace[i + d], trace[i + d + 4]));
                if (dif > PI) 
                    dif = 2 * PI - dif;

                //cout << " dif = " << dif << " i = " << i << " d = " << d << endl;
                if (dif >= 5. * PI / 6.) {
                    turns++;
                    // i += d;
                    i = i + d;// - 4;
                    d = 4;
                    break;
                }
            }
            //cout << " i = " << i << endl;
        } while ((i + d + 4 < trace.size())); // && (turns == 0));        
        i++;
           // cout << " i = " << i << endl;
    } while ( (i + 8 < trace.size())  && (turns == 0));

    if (ImgUtil::DEBUG)
        cout << turns << " turns have been found" << endl;
    return turns;
}

int LinWangFE::computeS(vector<Point> trace, Point turningpoint) {

    int i, j, d, turns;
    float dif;
    d = 4;
    int turning = -1;

//    cout << "turnig = " << turningpoint.x << ", " << turningpoint.y << endl;
//    if (turningpoint.x == -1) {
//        turns = 0;
//        
//    cout << "entro en -1" << endl;
//    d = 4;
//    i = 0;
//    turns = 0;
//        do {
//            do {
//                for (d = 4; i + d + 4 < trace.size(); d++) {
//                    dif = fabs(angBtwn2Points(trace[i], trace[i + 4]) - angBtwn2Points(trace[i + d], trace[i + d + 4]));
//                    if (dif > PI) 
//                        dif = 2 * PI - dif;
//                    if (dif >= 5. * PI / 6.) {
//                        turns++;
//                        //turning = i + d; 
//                        turning = (int) floor(i + (d + 4) / 2);
//                        i = i + d;
//                        d = 4;
//                        break;
//                    }
//                }
//            } while ((i + d + 4 < trace.size()) && (turns == 0)); // && (turns == 0));        
//            i++;
//        } while ( (i + 8 < trace.size())  && (turns == 0));
//        
//    } else {
        turningpoint.x = turningpoint.x * blockW + blockW / 2;
        turningpoint.y = turningpoint.y * blockW + blockW / 2;
        for (i = 0; i < trace.size(); i++) {
            if ((trace[i].x == turningpoint.x) && (turningpoint.y == trace[i].y)) {
                turning = i;
            }
        }
//    }

    if (turning == -1) {
        if (ImgUtil::DEBUG)
            cout << "NO hay turning point" << endl;
        return 0; // s = 0;
    }

    // Create midline
    vector<Point> midline;
    Point p;
    int step = 1;
    // First point
    midline.push_back(trace[turning]);
    do {
        p.x = (int) round((trace[turning - step].x + trace[turning + step].x) / 2.);
        p.y = (int) round((trace[turning - step].y + trace[turning + step].y) / 2.);
        if ((midline[step].x != p.x) || (midline[step].y != p.y)) {
                midline.push_back(p);
        }
        step++;
    } while (((turning - step) > 0) && ((turning + step) < trace.size()));


    vector<Point> juntas;
    // Create line between first and last point of midline
    vector<Point> line;
    Point last, first;
    float m;
    int y, x;


    last = midline[0];
    first = midline[midline.size() - 1];

    m = -(float) (last.y - first.y) / (last.x - first.x);
    float ang = atan2((first.y - last.y), (first.x - last.x));
    line.push_back(first);
    
    //cout << "first " << first.x << ", " << first.y << " end = " << last.x << ", " << last.y << " ang = " << ang * 180. / PI << endl;
    float xF = last.x, yF = last.y;
    do {
        
        xF += cos(ang);
        yF += sin(ang);
        
        p.x = round(xF);
        p.y = round(yF);
        
        if ((line[line.size() - 1].x != p.x) || (line[line.size() - 1].y != p.y))
                line.push_back(p);       
        
    } while ( (p.x != first.x) && (p.y != first.y) );

    for (i = 0; i < midline.size(); i++) {
        juntas.push_back(midline[i]);
    }
    for (i = 0; i < line.size(); i++) {
        juntas.push_back(line[i]);
    }

    if (ImgUtil::DEBUG) {
        int **M;
        M = LocationMatrix2(juntas);
        writeImageAndTrace2(img, M, "TraceandMidline");
    }

    // Calculate S
    int s;
    s = 0;
    int jAnt = 0;
    for (i = 1; i < line.size(); i++) {
        y = line[i].y;
        // cout << "line [" << i <<"] = "<< line[i].x << ", " << line[i].y << endl;
        if (y == line[i - 1].y) {// si es el mismo estoy contando de más 
         //   cout << line[i].y << " == ? " << line[i - 1].y<<  endl;
            continue;
        }
        for (j = jAnt; j < midline.size(); j++) {
       // cout << "midline [" << i <<"] = "<< midline[j].x << ", " << midline[j].y << endl;
            if (midline[j].y == y) {
                x = midline[j].x;
                break;
            }
        }
        jAnt = j;
        if (y < turningpoint.y)
            s -= (x - line[i].x);
        else
            s += x - line[i].x;
    }

    return s;

}

float LinWangFE::angBtwn2Points(Point p1, Point p2) {

    float ang;
    ang = atan2(-(p2.y - p1.y), (p2.x - p1.x));
    return ang;
}

int** LinWangFE::LocationMatrix(vector<Point> trace) {

    int **Mat;
    int i, j, x, c;

    Mat = MemUtil::create2DMatrix<int>(orH, orW);

    c = (int) floor(255 / trace.size());
    for (i = 0; i < orH; i++) {
        for (j = 0; j < orW; j++) {
            Mat[i][j] = 0;
            for (x = 0; x < trace.size(); x++) {
                if ((i == trace[x].y) && (j == trace[x].x))
                    Mat[i][j] = x * c;
            }
        }
    }
    return Mat;
}

int** LinWangFE::LocationMatrix2(vector<Point> trace) {

    int **Mat;
    int i, j, x;

    Mat = MemUtil::create2DMatrix<int>(h, w);

    for (i = 0; i < h; i++) {
        for (j = 0; j < w; j++) {
            Mat[i][j] = 0;
            for (x = 0; x < trace.size(); x++) {
                if ((i == trace[x].y) && (j == trace[x].x))
                    Mat[i][j] = 1;
            }
        }
    }
    return Mat;
}

void LinWangFE::writeImageAndTrace2(float **gImage, int **CI, string fileName) {

    ofstream f;


    f.open(fileName.append(".pgm").c_str(), ofstream::out | ofstream::trunc);
    if (!f) {
        cout << "FingerprintImg: Error opening file for Trace image!" << endl;
    }


    /* Writing the file in PGM format */
    f << "P2" << endl;
    f << "# This is the bitmap of the Cores and deltas image" << endl;
    f << w << " " << h << endl;
    f << (int) 255 << endl;

    for (int i = 0; i < h; i++) {
        for (int j = 0; j < w; j++) {
            if (CI[i][j] > 0)
                f << (int) CI[i][j] << " ";
            else
                f << (int) gImage[i][j] << " ";

        }
        f << endl;
    }
    /* End of the PGM file format */
    f.close();

}

float LinWangFE::computeFeatureVector() {

    float featurevector = -1;
    GaussianFiltering(gFilterHLinWangFE, gFilterWLinWangFE);
    createOrientationImageOverlapping(block); // block size = 16 x 16
    computeCoherenceImage();
    smoothOrientationImage();
    computeCoresAndDeltas();
    Orientation2Index();
    if (((cores.size() / 2) == 1) && ((deltas.size() / 2) == 1))
        //featurevector = computePseudoRidge(cores, deltas);
        featurevector = analyzeCoreDelta();
    else if (((cores.size() / 2) == 0) && ((deltas.size() / 2) == 0)) {
        if (this->coreDeltasFound){//cout << "T" << endl;
            featurevector = 3;
        }
        else{//cout << "A" << endl;
            featurevector = 1; // Arch
        }
    }
    else if ((cores.size() / 2) == 2) {
        featurevector = 0; // Whorl
        //cout << "W" << endl;
    }
    else if (((deltas.size() / 2) == 2) && ((cores.size() / 2) == 0) ) {
        //cout << "-1" << endl;
        return -1;
    }
    else{
        //pseudoridges
        featurevector = computePseudoRidge(cores, deltas);
        //cout << "PseudoRidge ";
    }

    return featurevector;
}
