/* 
 * \file        LinWangFE.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for LinWangFE class
 *
 * 
 */

#ifndef LINWANGFE_H
#define	LINWANGFE_H

#include <vector>

#include "Point.h"
#include "PCASYSManager.h"

class LinWangFE {
public:
    LinWangFE();
    LinWangFE(float **image, int h, int w);
    LinWangFE(const LinWangFE& orig);
    ~LinWangFE();
    
    
     /**
     * Main method of the class. It processes the fingerprint to obtain the 
     * feature vector for classification purposes
     * \return the feature vector of size .....
     */
    float computeFeatureVector();
    //TODO: parameterized the method with nScales, blockW, consThreshold
    
    /**
     * Computes the orientation map for the fingerprint image in blocks of size
     * blockW x blockW.It uses sobel's method for gradient computation
     * The orientation map is stored in O and xCos, ySin
     * \param blockW the size of each orientation block
     */
    int createOrientationImage(int blockW);
    int createOrientationImageOverlapping(int blockW);
    
    /**
     * Smooths the obtained orientaiton map. 
     * The smoothed orientation map is stored in O and xCos, ySin
     * \param ...
     * \param ...
     */
    int smoothOrientationImage();
    
    /**
     * It initializes the image, width and height.
     * \param image original image
     * \param h image's height 
     * \param w image's width
     */
    void initialize(float **image, int h, int w);
    
    /**
     * It initializes the block size of the orientation image by using ImgUtil::fAdjust
     * (A block size can't be odd so we create this method to check whether after the block adjust
     * it is pair or not).
     */
    void initializeBlocks();
	
	/**
     * Executes the Feature Extractor Hong method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return fingerprint class
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * Smooths the fingerprint image by a gaussian filter. The filtered image is
     * stored in filI
     * \param gausFilterH Gaussian filter's height
     * \param gausFilterW Gaussian filter's width
     */
    int GaussianFiltering(int gausFilterH, int gausFilterW);
    
     /**
     * Computes the Coherence matrix X and stores it in Xp 
     */
    int computeCoherenceImage();
    
    /**
     * Indicates the fingerprint's type depending on the core and delta location.
     * \return TA, L or R depending on fingerprint's type found
     */
    int analyzeCoreDelta();
    
    /**
     * Returns the fingerprint's type calculated with the trace of de ridges
     * around the cores or deltas
     * \param core vector with cores' coordinates
     * \param delta vector with deltas' coordinates
     * \return feature vector type (R, L, W, A, T)
     */
    float computePseudoRidge(vector<float> core, vector<float> delta);
    
    /**
     * Transforms the Orientation matrix into a indexes orientation matrix
     */
    void Orientation2Index();
    
    /**
     * Returns the vector of Points of the trace that starts from the point start
     * \param point start with the coordinates of the starting point
     * \param int pseudoridge with the maximum size of the pseudoridge tracing
     * \return a vector containing the points of the pseudoridge tracing
     */
    vector<Point> Tracepseudoridge(Point start, int pseudoRidgeSize);
    
    /**
     * Returns the position of the point that corresponds to the turning point of a trace curve
     * \param trace vector containing the points of the pseudoridge tracing
     * \return the position in the vector of the turning point
     */
    int findTurningPoint(vector<Point> trace);
    
    /**
     * Returns the number of rows to jump for the next tracing index
     * \param index is the next tracing index
     * \return the number of rows to jump for the next tracing index
     */
    int getimove(int index);
    
    /**
     * Returns the number of columns to jump for the next tracing index
     * \param index is the next tracing index
     * \return the number of columns to jump for the next tracing index
     */
    int getjmove(int index);
    
    /**
     * Returns the number of turn of the trace curve
     * \param trace is a vector with the points of the trace curve
     * \return the number of turns
     */
    int traceturns(vector<Point> trace);
            
    /**
     * Returns the area s placed between the medial axis of the trace curve and the straight  
     * line between the starting point and the end point
     * \param trace is a vector with the points of the trace curve
     * \param turning is the stating point of the medial axis
     * \return the area s
     */	
    int computeS(vector<Point> trace, Point turningpoint);
    
    /**
     * Returns the number angle of the line that connect two points
     * \param p1 is a point
     * \param p2 is a point
     * \return the angle between these two points
     */
    float angBtwn2Points(Point p1, Point p2);
    
    /**
     * Draws a matrix with the points of the traced curve
     * \param trace is a vector with the points of the trace curve
     * \return a matrix with the traced curve drawed
     */
    int** LocationMatrix(vector<Point> trace);
    
    /**
     * Obtains the cores and deltas of an orientation image.
     * \return vector with cores' and deltas' coordinates
     */
    int computeCoresAndDeltas();
    
    /**
     * Create an image file with the original fingerprint and the traced curve
     * \param gImage is the original fingerprint image
     * \param CI is the matrix with the traced curve drawed
     * \param fileName is the name of the file
     */
    void writeImageAndTrace2(float **gImage, int **CI, string fileName);
    
    int** LocationMatrix2(vector<Point> trace);
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:
    float **img;                //!< Reference to the fingerprint image data
//    float **imgOrig;            //!< Original image when using enhanced hong image
    int h;                      //!< Height of the image
    int w;                      //!< Width of the image
    
    float **imgFiltered;        //!< Reference to the image filtered with a gaussian filter 
     
    float **O;                  //!< Orientation of each block (in radians)
    float **xCos;               //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;               //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    int orH;                    //!< Height of the orientation map
    int orW;                    //!< Width of the orientation map
    int blockW;                 //!< Block size used to compute the orientation map
 
    float **X;                  //!< Coherence value of each orientation block
    float **Xp;                 //!< Final coherence (background, foreground, noise)
    
    float **Op;                 //!< Smoothed orientation image
    float **xCosOp;             //!< cos(2 Op) value of the cosine of the orientation (x-direction)
    float **ySinOp;             //!< sin(2 Op) value of the sine of the orientation (y-direction)
    
    float **gradX;
    float **gradY;
    
    vector<float> cores;        //!< Core coordinates' vector
    vector<float> deltas;       //!< Delta coordinates' vector
    bool coreDeltasFound;
    
    bool **segmentation;        //!< Binary image with the pixels that belongs to the fingerprint or background
    bool **smaskp;
    int numBlockBack;           //!< Segmentation number of blocks that belongs to background
    
    float **C;
    
    float **energy;
    
    int **OrI;                  //!< Orientation map translated to pixels
    int **OrIp;                 //!< Orientation map for each pixel discretized
    
    float **Poincare;           //!< Poincare value of each orientation block
    
    float **featureVector;      //!< Feature vector computed from the cores and deltas 
    bool **segmentationPoincare;        //!< Segmentation obtained from ImageEnhancedHong

    int dimFV;                          //!< Dimension of feature vector
    
    //Adjust parameters
    int block;             //!< Adjusted orientation image block size
};

inline int LinWangFE::getDimFV() const { return dimFV;}

#endif	/* LINWANGFE_H */

