/* 
 * \file        MaskRegion.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for MaskRegion (Capelli99) class
 *
 * 
 */


#ifndef MASKREGION_H
#define	MASKREGION_H

#include "MaskVertex.h"

/**
 * \class MaskRegion
 *
 * The MaskRegion class implements the structure of a region of a dynamic
 * mask used in Cappelli1999's algorithm
 */
class MaskRegion {
public:
    /* Default constructor */
    MaskRegion();
    
    /**
     * It constructs the minimum region of a dynamic mask (which requires 3 points)
     * THE VERTICES MUST BE INTRODUCED IN CLOCK-WISE ORDER
     * \param p1 First vertex of the region
     * \param p2 Second vertex of the region
     * \param p3 Third vertex of the region
     */
    MaskRegion(MaskVertex *p1, MaskVertex *p2, MaskVertex *p3);
    
    /* Default destructor */
    ~MaskRegion();
    
    /**
     * It adds a new vertex to the region
     * @param p The vertex to be added
     */
    void addVertex(MaskVertex *p);
    
    /**
     * It finds whether a point is inside the region or not
     * \param xt The x-coordinate of the point
     * \param yt The y-coordinate of the point
     * \return Whether a point is inside the region or not
     */
    bool isInRegion(int xt, int yt);
    
    /**
     * It prints the mask region
     */
    void print();
    
    /**
     * It sets the variance of the region
     * \param var The variance to be set
     */
    void setVariance(float var);
    
    /**
     * It sets the average orientation of the region
     * \param O The average orientation of the region
     */
    void setAverageO(float O);
    
    /**
     * It returns the variance of the region
     * \return The variance of the region
     */
    float getVariance() const;
    
    /**
     * It returns the average orientation of the region
     * \return The average orientation of the region
     */
    float getAverageO() const;
    
private:
    
    vector<MaskVertex *> vertices;      //!< The vertices which form the region
    
    float variance;                     //!< The variance of the region
    float averageO;                     //!< The average orientation of the region

};


inline void MaskRegion::setVariance(float var) { variance = var; }
inline void MaskRegion::setAverageO(float O) { averageO = O; }
    
inline float MaskRegion::getVariance() const { return variance; }
inline float MaskRegion::getAverageO() const { return averageO; }

#endif	/* MASKREGION_H */

