/* 
 * \file        MaskVertex.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for MaskVertex (Capelli99) class
 *
 * 
 */


#ifndef MASKVERTEX_H
#define	MASKVERTEX_H

#include <vector>


using namespace std;

enum vertexType {fix = 0, mobile = 1, dependentX = 2, dependentY = 3, dependentXY = 4};


/**
 * \class MaskVertex
 *
 * The MaskVertex class implements the structure of a vertex from a dynamic mask
 * which are used in Cappelli1999's algorithm
 */
class MaskVertex {
public:
    /* Defautl constructor */
    MaskVertex();
    
    /**
     * Constructor of a new vertex
     * \param x The x-position of the vertex
     * \param y The y-position of the vertex
     * \param type The type of the vertex (fix, mobile, dependentX, dependentY, dependentXY)
     * \param maxDeltaX The maximum displacement in X if it is a mobile vertex
     * \param maxDeltaY The maximum displacement in Y if it is a mobile vertex
     */
    MaskVertex(int x, int y, vertexType type, int maxDeltaX = 0, int maxDeltaY = 0);
    
    /* Default destructor */
    ~MaskVertex();
    
    /* It adds a new dependen vertex to myDependents vector
     * being these vertex those depend on my movements
     */
    int addDependent(MaskVertex *p);
    
    /**
     * It returns the actual x-position of the vertex
     * \return The actual x-position of the vertex
     */
    int getX()const;
    
    /**
     * It returns the actual y-position of the vertex
     * \return The actual y-position of the vertex
     */
    int getY()const;
    
    /**
     * It changes the global rotation of the vertex
     * \param ang The angle used to rotate the vertex
     */
    void setGlobalAng(float ang);
    
    /**
     * It sets the size of the orientation map, which is needed to decide
     * whether a vertex is a corner, a border or none of them
     * @param h The height of the orientation map
     * @param w The width of the orientation map
     */
    static void setOrientationMapSize(int h, int w);
    
    /**
     * It returns the height of the orientation map
     * \return The height of the orientation map
     */
    static int getOrH();
    
    /**
     * It returns the width of the orientation map
     * \return The width of the orientation map
     */
    static int getOrW();
    
    /**
     * It returns whether the vertex is in a border line or not
     * \return Whether the vertex is in a border line or not
     */
    bool border() const; 
    
    /**
     * It returns whether the vertex is a corner or not
     * \return Whether the vertex is is a corner or not
     */
    bool corner() const; 
    
    /**
     * It prints the vertex
     */
    void print();
    
    /**
     * It returns whether the vertex is mobile or not
     * \return Whether the vertex is mobile or not
     */
    bool isMobile() const;
    
    /**
     * It returns the maximum x-displacement allowed by the vertex
     * \return The maximum x-displacement allowed by the vertex
     */
    int getMaxDeltaX() const;
    
    /**
     * It returns the maximum y-displacement allowed by the vertex
     * \return The maximum y-displacement allowed by the vertex
     */
    int getMaxDeltaY() const;

        
    /**
     * It alters the global rotation and displacement of the vertex
     * \param deltaX The x-displacement
     * \param deltaY The y-displacement
     * \param ang The angle to be used for the rotation (0 is upward direction)
     */
    void alterGlobalRotationDisplacement(int deltaX, int deltaY, float ang);
    
    /**
     * It alters the local displacement of the vertex depending on its type
     * \param deltaX The x-displacement
     * \param deltaY The y-displacement
     */
    void alterLocalDisplacement(int deltaX, int deltaY);
    
    /**
     * It sets the reference point of the vertex, to be used to correct the 
     * vertex positioning after the rotation if the vertex is not mobile
     * @param referencePoint
     */
    void setReferencePoint(MaskVertex *referencePoint);
    
    /**
     * It corrects the vertex position depending on its reference point (if 
     * it is not a mobile vertex)
     * @param global Whether the correction is being carried out after the global rotation-positioning (true)
     * or after a local positioning of the vertex (false)
     */
    void correctVertexPosition(bool global = false);
    
private:
    int x;                      //!< The original x-position of the vertex
    int y;                      //!< The original y-position of the vertex
    
    vertexType type;            //!< The type of the vertex
    
    /* For mobile vertices (but also for the others) */
    int deltaX;                 //!< The x-displacement of the vertex
    int deltaY;                 //!< The y-displacement of the vertex
    
    int maxDeltaX;              //!< The maximum local x-displacement of the vertex
    int maxDeltaY;              //!< The maximum local y-displacement of the vertex 
    // For fixed vertices, deltas are always 0
    // For mobile vertices these are estimated
    // For dependent vertices, these are altered by their boss vertex
        
    
    vector<MaskVertex *> myDependents;  //!< The dependent vertices, which must be locally moved depending on the local displacement of the vertex
    
    MaskVertex *myReferencePoint;       //!< The reference point to correct the position of the given vertex after rotation/displacement
    
    int globalDeltaX;           //!< The global x-displacement applied to the vertex
    int globalDeltaY;           //!< The global y-displacement applied to the vertex 
    float globalAng;            //!< The global rotation applied to the vertex 
    
    static int orH;             //!< The height of the orientation map
    static int orW;             //!< The width of the orientation map
    
    bool isBorder;              //!< Whether the vertex is in a border line or not
    bool isCorner;              //!< Whether the vertex is a corner or no
    
    int actualX;                //!< The actual x-position of the vertex (after the rotation-displacement)
    int actualY;                //!< The actual y-position of the vertex (after the rotation-displacement)
    
    int globalActualX;          //!< The x-position of the vertex after the global rotation-displacement operation (after correction)
    int globalActualY;          //!< The y-position of the vertex after the global rotation-displacement operation (after correction) 
    
};

inline int MaskVertex::getX() const { return actualX; }

inline int MaskVertex::getY() const { return actualY; }

inline bool MaskVertex::border() const { return isBorder; }
inline bool MaskVertex::corner() const { return isCorner; }


inline bool MaskVertex::isMobile() const { return type == mobile; }

inline int MaskVertex::getMaxDeltaX() const { return maxDeltaX; }
inline int MaskVertex::getMaxDeltaY() const { return maxDeltaY; }


inline void MaskVertex::setReferencePoint(MaskVertex *referencePoint) { myReferencePoint = referencePoint; }

#endif	/* MASKVERTEX_H */

