/* 
 * \file        NewPoincareSingularPoints.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Implementation of the NewPoincareSingularPoints class
 *
 * 
 */



#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "FingerprintImg.h"

#include "NewPoincareSingularPoints.h"
#include "Cluster.h"

#include "ParametersKaru.h"
#include "ParametersPoincare.h"
#include "ParametersZhang.h"
#include "ParametersHongFE.h"


/*
 * Constructor / Destructor
 */
NewPoincareSingularPoints::NewPoincareSingularPoints() {


    w = -1;
    h = -1;

    img = 0;
    imgNormalized = 0;


    gradX = 0;
    gradY = 0;

    O = 0;

    gradMagnitude = 0;
    orW = -1;
    orH = -1;

    labelImg = 0;

    poincareIndex = 0;

    blockMean = 0;
    blockVariance = 0;
    blockSegmentation = 0;

    xCos = 0;
    ySin = 0;
}

NewPoincareSingularPoints::NewPoincareSingularPoints(float **image, int h, int w) {


    this->w = w;
    this->h = h;
    img = MemUtil::copy2DMatrix(image, h, w);
    imgNormalized = 0;


    gradX = 0;
    gradY = 0;

    O = 0;

    gradMagnitude = 0;
    orW = -1;
    orH = -1;

    labelImg = 0;

    poincareIndex = 0;

    blockMean = 0;
    blockVariance = 0;
    blockSegmentation = 0;


    xCos = 0;
    ySin = 0;
}

NewPoincareSingularPoints::NewPoincareSingularPoints(float **O, float **xCos, float **ySin, int orH, int orW) {

    w = -1;
    h = -1;

    img = 0;
    imgNormalized = 0;


    gradX = 0;
    gradY = 0;

    this->O = MemUtil::copy2DMatrix(O, orH, orW);
    this->xCos = MemUtil::copy2DMatrix(xCos, orH, orW);
    this->ySin = MemUtil::copy2DMatrix(ySin, orH, orW);

    gradMagnitude = 0;

    this->orW = orW;
    this->orH = orH;

    labelImg = 0;

    poincareIndex = 0;

    blockMean = 0;
    blockVariance = 0;
    blockSegmentation = 0;
 
    /*if (ImgUtil::DEBUG){
        string fileName = "InitiPoincareOrientations";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName);
    }*/

}

NewPoincareSingularPoints::~NewPoincareSingularPoints() {


    MemUtil::freeMatrix(&img);
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&O);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    // MemUtil::freeMatrix(&labelImg);
    // MemUtil::freeMatrix(&poincareIndex);
    MemUtil::freeMatrix(&gradMagnitude);
    MemUtil::freeMatrix(&blockSegmentation);
    MemUtil::freeMatrix(&blockMean);
    MemUtil::freeMatrix(&blockVariance);
    MemUtil::freeMatrix(&imgNormalized);

    blocksPerCore.clear();
    blocksPerDelta.clear();
    coreCentroids.clear();
    deltaCentroids.clear();

}

void NewPoincareSingularPoints::initialize(float **img, int h, int w){
    this->h = h;
    this->w = w;
    
    this->img = MemUtil::copy2DMatrix(img, h, w);
}

void NewPoincareSingularPoints::run(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    
    if (ImgUtil::fAdjust != 1){
        fi.setPCASYSManager(pcasys);   
        fi.readFingerprintImage(file);
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
    
    createOrientationImage(((int)defaultBlockSizePoincare * (ImgUtil::fAdjust)));
    // 1 at 3rd pos means KaruRules (0 would be no rules and 2 HongRules)
    // 3 at 4th pos means min block of 3 to validate poincare index regions
    computeCoresAndDeltas(curveWPoincare, smoothOrientatioWPoincare, poincareModePoincare, rulesPoincare, minNBlocksPoincare, normalizePoincare);
    //npsp->computeCoresAndDeltasFromImg(16, smoothOrientatioW_poincare, curveWPoincare, normalizePoincare, karuRulesPoincare);

}

int NewPoincareSingularPoints::runReject(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    
    if (ImgUtil::fAdjust != 1){
        fi.setPCASYSManager(pcasys); 
        fi.readFingerprintImage(file);
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
    
    int blockW = (int)(defaultBlockSizeAreaPoincare * (ImgUtil::fAdjust));
    orH = h / blockW;
    orW = w / blockW;
    
    float area =  this->generateSegmentation();
    
 //   if (ImgUtil::DEBUG)
        cout << "Area de huella = " << area << endl;

    if (area > (thresholdAreaRejectPoincare + 1. - ImgUtil::fAdjust))
        return 1;
    else
        return 0;
    
}

int *NewPoincareSingularPoints::selectCenterInBlock(int *auxVector,int *covMargins,int h, int w, int bSize){

    int northStripe=(covMargins[0]*bSize);
    int southStripe=(covMargins[1]*bSize);
    int eastStripe=(covMargins[2]*bSize);
    int westStripe=(covMargins[3]*bSize);

    int centerI=round((auxVector[0]+0.5)*bSize);
    int centerJ=round((auxVector[1]+0.5)*bSize);

    if (centerI<northStripe){
        if (centerI+(bSize/2)>northStripe){ centerI=centerI+(bSize/2); }
    }

    if (centerI>(h-southStripe)){
        if (centerI-(bSize/2)<(h-northStripe)){ centerI=centerI-(bSize/2); }
    }

    if (centerJ<westStripe){
        if (centerJ+(bSize/2)>westStripe){ centerJ=centerJ+(bSize/2); }
    }

    if (centerJ>(w-eastStripe)){
        if (centerJ-(bSize/2)<(w-eastStripe)){ centerJ=centerJ-(bSize/2); }
    }

    int *retVal= MemUtil::createVector<int>(2);
    retVal[0]=centerI;
    retVal[1]=centerJ;

    if (ImgUtil::DEBUG)
        cout << "Initial candidates are ("<<round((auxVector[0]+0.5)*bSize)<<","<<round((auxVector[1]+0.5)*bSize)<<"), finals are ("<<retVal[0]<<","<<retVal[1]<<")."<<endl;

    return retVal;

}

/*
 * Get / Set
 */
void NewPoincareSingularPoints::setBlockSegmentation(bool **blockSegmentation) {
    MemUtil::freeMatrix(&this->blockSegmentation);
    this->blockSegmentation = MemUtil::copy2DMatrix(blockSegmentation, orH, orW);
}

void NewPoincareSingularPoints::combineWithSegmentation(bool **segmentation) {
    for (int i = 0; i < orH; i++)
        for (int j = 0; j < orW; j++) {
            if (blockSegmentation[i][j] != segmentation[i][j])
                blockSegmentation[i][j] = false;
        }

    if (ImgUtil::DEBUG) {
        string fileName = "PoincareSegmentedMaskCombined";
        ImgUtil::writeGrayScaleImage(blockSegmentation, orH, orW, fileName);
    }
}

void NewPoincareSingularPoints::setOrientationMap(float **O, float **xCos, float **ySin, int orH, int orW) {
    
    MemUtil::freeMatrix(&this->O);
    MemUtil::freeMatrix(&this->xCos);
    MemUtil::freeMatrix(&this->ySin);
    
    this->O = MemUtil::copy2DMatrix(O, orH, orW);
    this->xCos = MemUtil::copy2DMatrix(xCos, orH, orW);
    this->ySin = MemUtil::copy2DMatrix(ySin, orH, orW);
    
    this->orH = orH;
    this->orW = orW;
}

/*
 * Main functions for feature computations
 */

int  NewPoincareSingularPoints::computeCoresAndDeltasFromImg(int curveW, int smoothW, operationModePoincare mode, int specialRules, int minNBlock, bool normalize){
    
    //If no orientations were previusly computed
    //  do so with the default block size
   // if (this->O == 0) {
        createOrientationImage(defaultBlockSizePoincare);
   // }


    //By Default, increasing smoothing computation    
    if (smoothW > 0)
        return computeCoresAndDeltas(curveW, smoothW, mode, specialRules, minNBlock, normalize);
    else {
        computeAllCoresAndDeltas(curveW);
        return 0;
    }

}


int NewPoincareSingularPoints::computeCoresAndDeltas(int curveW, int smoothW, operationModePoincare mode, int appliedRules, int minNBlock, bool normalize) {

   // cout << "curveW = " << curveW << " smoothW = " << smoothW << " mode = " << mode << " appliedRules = " << appliedRules <<
   //         " minBlock = " << minNBlock << " normalize " << normalize << endl;
    
    this->mode = mode;
    bool end = false;
    int nSmooth = 0;    
    int iniCurveW = curveW;
    blocksPerCore.clear();
    blocksPerDelta.clear();
    coreCentroids.clear();
    deltaCentroids.clear();

    
//    if (ImgUtil::isSfinge)
//        appliedRules = 0;
    processSpecialRules(appliedRules);

    while (!end) {
        computeAllCoresAndDeltas(curveW);
        
        //cout <<"pre min block rules"<<endl;
        
        if (minNBlock * ImgUtil::fAdjust > 0) 
            cleanSPsMinBlock(minNBlock * ImgUtil::fAdjust);
		
        //cout <<"applied min block rules"<<endl;
        if (clusterRules)
            cleanSPsWithClusterRules(minDistanceZhang * ImgUtil::fAdjust);
        
        if (karuRules) 
            cleanSPsWithKaruRules();

        if (hongRules)
            cleanSPsWithHongRules(minNBlock * ImgUtil::fAdjust);

        if (ImgUtil::DEBUG)
            cout << "Tras eliminar quedan " << blocksPerCore.size() << " cores y " << blocksPerDelta.size() << " deltas" << endl;

        	
        if (mode == increasingSmoothPoincare) {
            if (((blocksPerCore.size() <= 2) && (blocksPerDelta.size() <= 2)) || nSmooth > maxSmoothPoincare) {
                    end = true;
                    if (blocksPerCore.size() != blocksPerDelta.size())
                            ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, smoothW, normalize, &blockSegmentation);
                    break;
            }
            ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, smoothW, normalize, &blockSegmentation);
            nSmooth++;
            if (ImgUtil::DEBUG) {

                string fileName2 = "OrientationSinCosSmoothPoincare";
                char str[20];
                sprintf(str, "%d", smoothW);
                fileName2.append(str);
                sprintf(str, "%d", nSmooth);
                fileName2.append(str);
                ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
            }
        }
        else {
            if ((blocksPerCore.size() <= 2) && (blocksPerDelta.size() <= 2)) {
                    end = true;
                    break;
            }
            curveW--;
            if (curveW <= 1) {
                    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, smoothW, normalize, &blockSegmentation); // smooth a 3 y reiniciar
                    smoothW += 2;
                    curveW = iniCurveW;
                    nSmooth++;
            }
        }
        blocksPerCore.clear();
        blocksPerDelta.clear();
        coreCentroids.clear();
        deltaCentroids.clear();

        MemUtil::freeMatrix(&poincareIndex);
        poincareIndex = 0;
    }

    if (ImgUtil::DEBUG) {
        cout << "Al final hay " << blocksPerCore.size() << " cores y " << blocksPerDelta.size() << " deltas" << endl;
        cout << "CORES: " << endl;
        for (int i = 0; i < blocksPerCore.size(); i++)
            cout << "Number of blocks = " << blocksPerCore[i] << " centroid: " << coreCentroids[2 * i] << ", " << coreCentroids[2 * i + 1] << endl;
        cout << "DELTAS: " << endl;
        for (int i = 0; i < blocksPerDelta.size(); i++)
            cout << "Number of blocks = " << blocksPerDelta[i] << " centroid: " << deltaCentroids[2 * i] << ", " << deltaCentroids[2 * i + 1] << endl;
    }
    return nSmooth;
}


void NewPoincareSingularPoints::computeAllCoresAndDeltas(int curveW) {



    //3 Initialize Label image
    labelImg = MemUtil::create2DMatrix<int>(orH, orW);

    // cout<<"step 4\n";
    //4- Calculate poincare index
    computePoincareIndex(curveW);
    if (ImgUtil::DEBUG) {
        string fileNameGray = "poincareIndex_preclean";
        ImgUtil::writeGrayScaleImage(poincareIndex, orH, orW, fileNameGray, 255);
    }

    if (blockSegmentation != 0) {
        deleteLabelsOutsideSegmentedRegion();
        if (ImgUtil::DEBUG) {
            string fileNameGray = "poincareIndex";
            ImgUtil::writeGrayScaleImage(poincareIndex, orH, orW, fileNameGray, 255);
        }

        if (ImgUtil::DEBUG) {
            string fileNameGray = "segmented_mask";
            ImgUtil::writeGrayScaleImage(blockSegmentation, orH, orW, fileNameGray);
        }
    }

    int nCores = labelCores();
    int nDeltas = labelDeltas();
    nDeltas = abs(nDeltas);

    if (ImgUtil::DEBUG)
        cout << "Hay " << nCores << " cores y " << nDeltas << " deltas" << endl;

    MemUtil::freeMatrix(&labelImg);
    labelImg = 0;
}

/*
 * Main functions for feature computations
 */
void NewPoincareSingularPoints::generateSegmentation(float** image, int h, int w) {
    this->w = w;
    this->h = h;

   // int blockW = w / orW;
    MemUtil::freeMatrix(&img);
    img = MemUtil::copy2DMatrix(image, h, w);
    
    generateSegmentation();
}

float NewPoincareSingularPoints::generateSegmentation() {


//    this->w = w;
//    this->h = h;
//
    int blockW = w / orW;
    int blockH = h / orH;

    //MemUtil::freeMatrix(&img);
//    img = MemUtil::copy2DMatrix(image, h, w);
    imgNormalized = MemUtil::create2DMatrix<float>(h, w);
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "ImgBeforeNormalize";
        ImgUtil::writeGrayScaleImage(img, h, w, fileName2, 1);
    }
    
    float mean = 0;
    float variance = 0;
    
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
            mean += img[i][j];
    
    mean /=  (float) (h * w);
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++)
            variance += (img[i][j] - mean) * (img[i][j] - mean);
    
    variance /= (float) (h * w);
    
    float min = 255, max = 0;
    float aux;
    for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++) {
            aux = sqrt((desiredVariancePoincare * 
                    ((img[i][j] - mean) * (img[i][j] - mean))) / variance);
            
            aux *= img[i][j] > mean ? 1.0 : -1.0;

            imgNormalized[i][j] = (desiredMeanPoincare + aux);
        }
        
    if (ImgUtil::DEBUG) {
        string fileName2 = "NormalizedImg";
        ImgUtil::writeGrayScaleImage(imgNormalized, h, w, fileName2, 1);
    }

    //// RUN FAILED

    blockSegmentation = MemUtil::create2DMatrix<bool>(orH, orW);
    blockMean = MemUtil::create2DMatrix<float>(orH, orW);
    blockVariance = MemUtil::create2DMatrix<float>(orH, orW);


    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            for (int u = i * blockH; u < (i + 1) * blockH; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    blockMean[i][j] += imgNormalized[u][v];
                }
            }

            //Variance
            blockMean[i][j] /= (float) (blockH * blockW);

            for (int u = i * blockH; u < (i + 1) * blockH; u++)
                for (int v = j * blockW; v < (j + 1) * blockW; v++)
                    blockVariance[i][j] += (blockMean[i][j] - imgNormalized[u][v]) * (blockMean[i][j] - imgNormalized[u][v]);
            blockVariance[i][j] /= (float) (blockH * blockW);
        }
    }
    
    int countOK = 0;
    float vT = varThresholdPoincare;
    float mT = meanThresholdPoincare;
    float threshold = thresholdSegmentationPoincare * ImgUtil::fAdjust;
    if (ImgUtil::isSfinge) {
        threshold = 0;
        mT = 1000000;
    }
    do {
        countOK = 0;
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                if ((blockVariance[i][j] > vT) &&
                        (blockMean[i][j] < mT)) {
                    blockSegmentation[i][j] = true;
                    countOK++;
                }
            }
        }
//        vT -= 25;
//        mT += 10;
        vT -= vT * 0.025;
        mT += mT * 0.025;
        if (ImgUtil::DEBUG)
         cout << "countOK = " << countOK << " menor que " << orH * orW * thresholdSegmentationPoincare << endl;
    } while (countOK < orH * orW * threshold);

    if (ImgUtil::DEBUG) {
        string fileName2 = "blockMean";
        ImgUtil::writeGrayScaleImage(blockMean, orH, orW, fileName2, 1);
        fileName2 = "blockVariance";
        ImgUtil::writeGrayScaleImage(blockVariance, orH, orW, fileName2, 0.1);
    }

    if (ImgUtil::DEBUG) {
        string fileNameGray = "segmented_maskNoFill";
        ImgUtil::writeGrayScaleImage(blockSegmentation, orH, orW, fileNameGray);
    }

    bool **segmentedBlockImageFilled = ImgUtil::imFill(blockSegmentation, orH, orW, 3);
    if (ImgUtil::DEBUG) {
        string fileNameGray = "segmented_maskFilled";
        ImgUtil::writeGrayScaleImage(segmentedBlockImageFilled, orH, orW, fileNameGray);
    }
    MemUtil::freeMatrix(&blockSegmentation);
  // blockSegmentation = segmentedBlockImageFilled;
   /* bool **segmentedBlockImageDilated = ImgUtil::imDilate(blockSegmentation, orH, orW);
    MemUtil::freeMatrix(&blockSegmentation);
    if (ImgUtil::DEBUG) {
        string fileNameGray = "segmented_maskDilated";
        ImgUtil::writeGrayScaleImage(segmentedBlockImageDilated, orH, orW, fileNameGray);
    }*/

        blockSegmentation = ImgUtil::imErode(segmentedBlockImageFilled, orH, orW);
        if (ImgUtil::DEBUG) {
            string fileNameGray = "segmented_maskErored";
            ImgUtil::writeGrayScaleImage(blockSegmentation, orH, orW, fileNameGray);
        }
        MemUtil::freeMatrix(&segmentedBlockImageFilled);
    if (ImgUtil::isSfinge == false) {
        segmentedBlockImageFilled = ImgUtil::imErode(blockSegmentation, orH, orW);
        if (ImgUtil::DEBUG) {
            string fileNameGray = "segmented_maskErored2";
            ImgUtil::writeGrayScaleImage(segmentedBlockImageFilled, orH, orW, fileNameGray);
        }

        MemUtil::freeMatrix(&blockSegmentation);
        blockSegmentation = segmentedBlockImageFilled;
    }
        
    if (ImgUtil::DEBUG) {
        string fileNameGray = "segmentedImage";
        ImgUtil::writeGrayScaleImage(img, blockH * orH, blockW * orW, fileNameGray, 1, blockSegmentation, blockW);
    }

    float area = 0;

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (blockSegmentation[i][j]) 
                area++;
        }
    }

    area /= (float)(orH * orW);

    return area;
        
}

void NewPoincareSingularPoints::createOrientationImage(int blockW) {

    orH = h / blockW;
    orW = w / blockW;

    O = MemUtil::create2DMatrix<float>(orH, orW);
    xCos = MemUtil::create2DMatrix<float>(orH, orW);
    ySin = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vx = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH, orW);


    gradMagnitude = MemUtil::create2DMatrix<float>(h, w);
    gradX = MemUtil::create2DMatrix<float>(h, w);
    gradY = MemUtil::create2DMatrix<float>(h, w);


    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] -
                    img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]);

            gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                    img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]);

            gradMagnitude[i][j] = sqrt(gradX[i][j] * gradX[i][j] + gradY[i][j] * gradY[i][j]);
        }
    }


    float Gsx, Gsy;
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                    Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;

                }
            }

            O[i][j] = 0.5 * atan2(Vx[i][j], Vy[i][j]);
            O[i][j] += (O[i][j] <= 0) ? 0.5 * PI : -0.5 * PI;

            xCos[i][j] = cos(2.0 * O[i][j]);
            ySin[i][j] = sin(2.0 * O[i][j]);
        }
    }

    if (ImgUtil::DEBUG) {

        string fileName2 = "gxImage";
        ImgUtil::writeGrayScaleImage(gradX, h, w, fileName2, 3);
        fileName2 = "gyImage";
        ImgUtil::writeGrayScaleImage(gradY, h, w, fileName2, 3);
        fileName2 = "Grad";
        ImgUtil::writeGrayScaleImage(gradMagnitude, h, w, fileName2, 3);

        fileName2 = "OrientationVyVx";
        ImgUtil::writeOrImg(Vx, Vy, orW, orH, fileName2);
        fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);

    }

    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, smoothOrientatioWPoincare);

    if (ImgUtil::DEBUG) {
        string fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }


    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&Vy);

    this->generateSegmentation(); //img, h, w);
}

void NewPoincareSingularPoints::computePoincareIndex(int curveW) {

    float sumDelta;
    float delta, diff, absdiff;

    //cout << "computeing poincare with curveW = " << curveW << endl;

    //Image where the poincare index is stored
    poincareIndex = MemUtil::create2DMatrix<float>(orH, orW);


    int numPositions = (4 * curveW) - 4;
    float *orientationValues = MemUtil::createVector<float>(numPositions + 1); //Window values
    int wHalf = (curveW - 1) / 2;

    // For usual poincare
    int wHalf2 = wHalf;
    if (curveW % 2 == 0) // For poincare with 4 blocks
        wHalf2++;

    bool endIteration = false;
    for (int i = wHalf + 1; i < orH - wHalf2 - 1; i++) {
        for (int j = wHalf + 1; j < orW - wHalf2 - 1; j++) {
            endIteration = false;
            if ((blockSegmentation != 0) && (!blockSegmentation[i][j]))
                endIteration = true;
            //The round starts in tht southeastern corner
            // and goes (first) north and (then) west.
            int posCounter = 0;
//            if ((blockSegmentation != 0) && ((!blockSegmentation[i + wHalf2 + 1][j + wHalf2 + 1])))
//                    endIteration = true;
            for (int k = wHalf2; k >= -wHalf; k--) {//Going up north
                orientationValues[posCounter] = O[i + k][j + wHalf2];
                posCounter++;
                if ((blockSegmentation != 0) && ((!blockSegmentation[i + k][j + wHalf2]))) //  || (!blockSegmentation[i + k][j + wHalf2 + 1])))
                    endIteration = true;
            }
//             if ((blockSegmentation != 0) && ((!blockSegmentation[i - 1][j + wHalf2 + 1])))
//                    endIteration = true;
            for (int k = wHalf2 - 1; k >= -wHalf; k--) {//Going left to the west
                orientationValues[posCounter] = O[i - wHalf][j + k];
                posCounter++;
                if ((blockSegmentation != 0) && ((!blockSegmentation[i - wHalf][j + k]))) //  || (!blockSegmentation[i - wHalf - 1][j + k]) ) )
                    endIteration = true;
            }
//             if ((blockSegmentation != 0) && ((!blockSegmentation[i - 1][j - 1])))
//                    endIteration = true;
            for (int k = -wHalf + 1; k <= wHalf2; k++) {//Going down south
                orientationValues[posCounter] = O[i + k][j - wHalf];
                posCounter++;
                if ((blockSegmentation != 0) && ((!blockSegmentation[i + k][j - wHalf]))) //  || (!blockSegmentation[i + k][j - wHalf - 1]) ))
                    endIteration = true;
            }
//             if ((blockSegmentation != 0) && ((!blockSegmentation[i + wHalf2 + 1][j - 1])))
//                    endIteration = true;
            for (int k = -wHalf + 1; k <= wHalf2; k++) {//Going right to the east
                orientationValues[posCounter] = O[i + wHalf2][j + k];
                posCounter++;
                if ((blockSegmentation != 0) && ((!blockSegmentation[i + wHalf2][j + k]))) // || (!blockSegmentation[i + wHalf2 + 1][j + k])))
                    endIteration = true;
            }
            
          //  endIteration = false;
            if (endIteration)
                continue;

            sumDelta = 0;
            for (int k = 0; k < numPositions; k++) {
                diff = orientationValues[k + 1] - orientationValues[k];
                absdiff = (float) fabs((double) diff);
                if (absdiff < PI / 2.)
                    delta = diff;
                else if (diff <= -PI / 2.)
                    delta = PI + diff;
                else
                    delta = diff - PI;
                sumDelta = sumDelta + delta;
            }
            poincareIndex[i][j] = (1. / (2. * PI)) * sumDelta;
        }
    }

    MemUtil::freeMatrix(&orientationValues);

}


/*
 * Auxiliar functions
 */

//Returns the number of regions found

int NewPoincareSingularPoints::labelCores() {
    int label = 0;
    int nComponents = 0;
    int coreI = 0;
    int coreJ = 0;

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if ((poincareIndex[i][j] >= 0.49) && (labelImg[i][j] == 0)) {
                label++;
                labelImg[i][j] = label;
                nComponents++;
                coreI = i;
                coreJ = j;
                nComponents += fillCoreRegion(i, j, label, &coreI, &coreJ);
                blocksPerCore.push_back(nComponents);

                coreCentroids.push_back((float) coreI / (float) nComponents);
                coreCentroids.push_back((float) coreJ / (float) nComponents);
                nComponents = 0;
            }
        }
    }
    if (ImgUtil::DEBUG) {
        cout << "Hay " << label << " cores" << endl;
        for (int i = 0; i < blocksPerCore.size(); i++)
            cout << "Number of blocks = " << blocksPerCore[i] << " centroid: " << coreCentroids[2 * i] << ", " << coreCentroids[2 * i + 1] << endl;
    }
    return label;
}

int NewPoincareSingularPoints::fillCoreRegion(int posI, int posJ, int label, int *coreI, int *coreJ) {
    int nComponents = 0;
    for (int i = max(0, posI - 1); i <= min(posI + 1, orH - 1); i++) {
        for (int j = max(0, posJ - 1); j <= min(posJ + 1, orW - 1); j++) {
            if ((poincareIndex[i][j] >= 0.45) && (labelImg[i][j] == 0)) {
                labelImg[i][j] = label;
                nComponents++;
                *coreI += i;
                *coreJ += j;
                nComponents += fillCoreRegion(i, j, label, coreI, coreJ);
            }
        }
    }
    return nComponents;

}

int NewPoincareSingularPoints::labelDeltas() {
    int label = 0;
    int nComponents = 0;
    int deltaI = 0;
    int deltaJ = 0;

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {

            if ((poincareIndex[i][j] <= -0.45) && (labelImg[i][j] == 0)) {
                label--;
                labelImg[i][j] = label;
                nComponents++;
                deltaI = i;
                deltaJ = j;
                nComponents += fillDeltaRegion(i, j, label, &deltaI, &deltaJ);
                blocksPerDelta.push_back(nComponents);

                deltaCentroids.push_back((float) deltaI / (float) nComponents);
                deltaCentroids.push_back((float) deltaJ / (float) nComponents);
                nComponents = 0;
            }
        }
    }
    if (ImgUtil::DEBUG) {
        cout << "Hay " << abs(label) << " deltas" << endl;
        for (int i = 0; i < blocksPerDelta.size(); i++)
            cout << "Number of blocks = " << blocksPerDelta[i] << " centroid: " << deltaCentroids[2 * i] << ", " << deltaCentroids[2 * i + 1] << endl;
    }
    return label;
}

int NewPoincareSingularPoints::fillDeltaRegion(int posI, int posJ, int label, int *deltaI, int *deltaJ) {
    int nComponents = 0;
    for (int i = max(0, posI - 1); i <= min(posI + 1, orH - 1); i++) {
        for (int j = max(0, posJ - 1); j <= min(posJ + 1, orW - 1); j++) {
            if ((poincareIndex[i][j] <= -0.45) && (labelImg[i][j] == 0)) {
                labelImg[i][j] = label;
                nComponents++;
                *deltaI += i;
                *deltaJ += j;
                nComponents += fillDeltaRegion(i, j, label, deltaI, deltaJ);
            }
        }
    }
    return nComponents;

}

void NewPoincareSingularPoints::deleteLabelsOutsideSegmentedRegion() {

    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            //Set to 0 those labels whose 
            if (!blockSegmentation[i][j]) {
                poincareIndex[i][j] = 0;
            }
        }
    }

}

void NewPoincareSingularPoints::processSpecialRules(int rules) {

    if (rules == 0) {
        this->karuRules = false;
        this->hongRules = false;
        this->clusterRules=false;
    }
    else if (rules == 1){
        this->karuRules=true;
        this->hongRules=false;
        this->clusterRules=false;
    }
    else if (rules == 2){
        this->karuRules=false;
        this->hongRules=true;   
        this->clusterRules=false;
    }
    else if (rules == 3){
        this->karuRules=true;
        this->hongRules=true;   
        this->clusterRules=false;
    }
    else if (rules == 4){
        this->karuRules=false;
        this->hongRules=false;   
        this->clusterRules=true;
    }
    else if (rules == 5){
        this->karuRules=true;
        this->hongRules=false;   
        this->clusterRules=true;
    }
    else if (rules == 6){
        this->karuRules=false;
        this->hongRules=true;   
        this->clusterRules=true;
    }
    else if (rules == 7){
        this->karuRules=true;
        this->hongRules=true;   
        this->clusterRules=true;
    }
    else{
        cout << "Warning <Wrong rules at NewPoincareSingularPoints::processSpecialRules."<<endl;
        cout << "\t Default 'no rules' option applies."<<endl;
    }
        
    
}

void NewPoincareSingularPoints::cleanSPsWithHongRules(int minNBlock) {

    int blockW = w / orW;

    bool *deleteCore = 0;
    if (this->blocksPerCore.size() > 0)
        deleteCore = MemUtil::createVector<bool>(this->blocksPerCore.size());
    bool *deleteDelta = 0;
    if (this->blocksPerDelta.size() > 0)
        deleteDelta = MemUtil::createVector<bool>(this->blocksPerDelta.size());


    //Core deleting
    for (int i = 0; i<this->blocksPerCore.size(); i++) {

        float iCore = coreCentroids[i * 2];
        float jCore = coreCentroids[i * 2 + 1];

        if ((iCore * blockW <= rules_noCoreStripWidthHongFE * ImgUtil::fAdjust) || (jCore * blockW <= rules_noCoreStripWidthHongFE * ImgUtil::fAdjust) ||
                (iCore * blockW >= h - rules_noCoreStripWidthHongFE * ImgUtil::fAdjust) || (jCore * blockW >= w - rules_noCoreStripWidthHongFE * ImgUtil::fAdjust)) {
            //  cout << "Eliminar core " << iCore << ", " << jCore << endl;
            deleteCore[i] = true;
        } else {
            deleteCore[i] = false;
        }

    }

    //Delta deleting
    for (int i = 0; i<this->blocksPerDelta.size(); i++) {

        float iDelta = deltaCentroids[i * 2];
        float jDelta = deltaCentroids[i * 2 + 1];

        if ((iDelta * blockW <= rules_noDeltaTopStripWidthHongFE * ImgUtil::fAdjust) || (jDelta * blockW <= rules_noDeltaStripWidthHongFE * ImgUtil::fAdjust) ||
                (iDelta * blockW >= h - rules_noDeltaStripWidthHongFE * ImgUtil::fAdjust) || (jDelta * blockW >= w - rules_noDeltaStripWidthHongFE * ImgUtil::fAdjust)) {
            //  cout << "Eliminar core " << iCore << ", " << jCore << endl;
            deleteDelta[i] = true;
        } else {
            deleteDelta[i] = false;
        }
    }

    if (ImgUtil::DEBUG) {
        cout << "Phase 4" << endl;
        cout << "Cores to be deleted (there exist " << this->coreCentroids.size() / 2 << ") :" << endl;
        for (int i = 0; i<this->blocksPerCore.size(); i++) {
            cout << "\t " << deleteCore[i];

        }
        cout << endl;
        cout << "Deltas to be deleted (there exist " << this->deltaCentroids.size() / 2 << ") :" << endl;
        for (int i = 0; i<this->blocksPerDelta.size(); i++) {
            cout << "\t " << deleteDelta[i];

        }
        cout << endl;
    }

    //Close core-delta pairs deleting

    for (int i = 0; i<this->blocksPerCore.size(); i++) {

        float iCore = coreCentroids[i * 2];
        float jCore = coreCentroids[i * 2 + 1];

        for (int j = 0; j<this->blocksPerDelta.size(); j++) {

            float iDelta = deltaCentroids[j * 2];
            float jDelta = deltaCentroids[j * 2 + 1];
            
            float dist = sqrt( (iDelta - iCore) * (iDelta - iCore) + 
                                (jDelta - jCore) * (jDelta - jCore));
            
          //  cout <<" dist ( c="<<i<<"d=" <<j<<")="<<dist*blockW<<" "<<endl;
            
            if (dist*blockW <= rules_minCoreDeltaDistHongFE * ceil(sqrt(minNBlock))  * ImgUtil::fAdjust ){
                deleteCore[i]=true;
                deleteDelta[j]=true;
            }
        }

    }


    if (ImgUtil::DEBUG) {
        cout << "Phase 5" << endl;
        cout << "Cores to be deleted (there exist " << this->coreCentroids.size() / 2 << ") :" << endl;
        for (int i = 0; i<this->blocksPerCore.size(); i++) {
            cout << "\t " << deleteCore[i];

        }
        cout << endl;
        cout << "Deltas to be deleted (there exist " << this->deltaCentroids.size() / 2 << ") :" << endl;
        for (int i = 0; i<this->blocksPerDelta.size(); i++) {
            cout << "\t " << deleteDelta[i];

        }
        cout << endl;
    }

    //Delete cores & deltas
    for (int i = blocksPerCore.size() - 1; i >= 0; i--) {

        if (deleteCore[i]) {

            coreCentroids.erase(coreCentroids.begin() + i * 2 + 1);
            coreCentroids.erase(coreCentroids.begin() + i * 2);
            blocksPerCore.erase(blocksPerCore.begin() + i);
        }


    }
    //cout <<"half "<<endl;

    for (int i = blocksPerDelta.size() - 1; i >= 0; i--) {


        if (deleteDelta[i]) {
            deltaCentroids.erase(deltaCentroids.begin() + i * 2 + 1);
            deltaCentroids.erase(deltaCentroids.begin() + i * 2);
            blocksPerDelta.erase(blocksPerDelta.begin() + i);
        }
    }

    if (ImgUtil::DEBUG) {
        cout << "After first deleting" << endl;
        cout << "\t There exist " << coreCentroids.size() / 2 << " cores and " << deltaCentroids.size() / 2 << " deltas." << endl;

        for (int i = 0; i<this->coreCentroids.size(); i = i + 2) {
            cout << "\t c: (" << this->coreCentroids[i] << "," << this->coreCentroids[i + 1] << ")" << endl;
        }
        for (int i = 0; i<this->deltaCentroids.size(); i = i + 2) {
            cout << "\t d: (" << this->deltaCentroids[i] << "," << this->deltaCentroids[i + 1] << ")" << endl;
        }
    }

    MemUtil::freeMatrix(&deleteCore);
    MemUtil::freeMatrix(&deleteDelta);


    //If more than two cores or deltas, pick up the (2) closest to the center

    // First, the cores
    if (this->blocksPerCore.size() > 2) {

        //Remove all of those that are not in the first 
        bool *deleteExtraCores = MemUtil::createVector<bool>(this->blocksPerCore.size());

        //Compute the center
        float iCenter = (float) orH / (float) 2;
        float jCenter = (float) orW / (float) 2;


        //get two candidate distancesd
        float iCore = coreCentroids[0];
        float jCore = coreCentroids[1];
        float minDist = (iCenter - iCore) * (iCenter - iCore) +
                (jCenter - jCore) * (jCenter - jCore); // no need for sqrt
        int minPos = 0;

        iCore = coreCentroids[2];
        jCore = coreCentroids[3];
        float prevMinDist = (iCenter - iCore) * (iCenter - iCore) +
                (jCenter - jCore) * (jCenter - jCore); // no need for sqrt
        int prevMinPos = 1;

        deleteExtraCores[0] = false;
        deleteExtraCores[1] = false;

        if (minDist > prevMinDist) {
            float aux = minDist;
            minDist = prevMinDist;
            prevMinDist = aux; //Now minDist has the smallest distance
            int auxInt = minPos;
            minPos = prevMinPos;
            prevMinPos = auxInt; //Now minPos points to the smallest distance
        }

      //  cout << "first dists are ("<<minDist<<" and "<<prevMinDist<<"), center is ("<<iCenter<<","<<jCenter<<")"<<endl;
        
        for (int i=2; i<blocksPerCore.size(); i++){
            
            iCore=coreCentroids[i*2];
            jCore=coreCentroids[(i*2) +1];
            
            float dist = (iCenter - iCore) * (iCenter - iCore) +
                    (jCenter - jCore) * (jCenter - jCore); // no need for sqrt



            if (minDist > dist) {
                deleteExtraCores[prevMinPos] = true; //First deleting the old candidate
                prevMinDist = minDist;
                prevMinPos = minPos;
                minDist = dist;
                minPos = i;
            } else if (prevMinDist > dist) {
                deleteExtraCores[prevMinPos] = true; //First deleting the old candidate
                prevMinDist = dist;
                prevMinPos = i;
            } else {
                deleteExtraCores[i] = true;
            }
     //       cout << "new: "<<dist<<"; first dists are ("<<minDist<<" and "<<prevMinDist<<")"<<endl;
        }

        //Now deleting them
        for (int i = blocksPerCore.size() - 1; i >= 0; i--) {
            if (deleteExtraCores[i]) {
                coreCentroids.erase(coreCentroids.begin() + i * 2 + 1);
                coreCentroids.erase(coreCentroids.begin() + i * 2);
                blocksPerCore.erase(blocksPerCore.begin() + i);
            }
        }
        MemUtil::freeMatrix(&deleteExtraCores);
    }

    //Deleting the extra deltas
    if (this->blocksPerDelta.size() > 2) {

        //Remove all of those that are not in the first 
        bool *deleteExtraDeltas = MemUtil::createVector<bool>(this->blocksPerDelta.size());

        //Compute the center
        float iCenter = (float) orH / (float) 2;
        float jCenter = (float) orW / (float) 2;


        //get two candidate distancesd
        float iDelta = deltaCentroids[0];
        float jDelta = deltaCentroids[1];
        float minDist = (iCenter - iDelta) * (iCenter - iDelta) +
                (jCenter - jDelta) * (jCenter - jDelta); // no need for sqrt
        int minPos = 0;

        iDelta = deltaCentroids[2];
        jDelta = deltaCentroids[3];
        float prevMinDist = (iCenter - iDelta) * (iCenter - iDelta) +
                (jCenter - jDelta) * (jCenter - jDelta); // no need for sqrt
        int prevMinPos = 1;

        deleteExtraDeltas[0] = false;
        deleteExtraDeltas[1] = false;

        if (minDist > prevMinDist) {
            float aux = minDist;
            minDist = prevMinDist;
            prevMinDist = aux; //Now minDist has the smallest distance
            int auxInt = minPos;
            minPos = prevMinPos;
            prevMinPos = auxInt; //Now minPos points to the smallest distance
        }

        for (int i = 2; i < blocksPerDelta.size(); i++) {

            iDelta = deltaCentroids[i * 2];
            jDelta = deltaCentroids[(i * 2) + 1];

            float dist = (iCenter - iDelta) * (iCenter - iDelta) +
                    (jCenter - jDelta) * (jCenter - jDelta); // no need for sqrt

            if (minDist > dist) {
                deleteExtraDeltas[prevMinPos] = true; //First deleting the old candidate
                prevMinDist = minDist;
                prevMinPos = minPos;
                minDist = dist;
                minPos = i;
            } else if (prevMinDist > dist) {
                deleteExtraDeltas[prevMinPos] = true; //First deleting the old candidate
                prevMinDist = dist;
                prevMinPos = i;
            } else {
                deleteExtraDeltas[i] = true;
            }
        }

        for (int i = blocksPerDelta.size() - 1; i >= 0; i--) {

            if (deleteExtraDeltas[i]) {
                deltaCentroids.erase(deltaCentroids.begin() + i * 2 + 1);
                deltaCentroids.erase(deltaCentroids.begin() + i * 2);
                blocksPerDelta.erase(blocksPerDelta.begin() + i);
            }
        }
        MemUtil::freeMatrix(&deleteExtraDeltas);

    }

    if (ImgUtil::DEBUG) {
        cout << "After removing extra cores and deltas" << endl;
        cout << "\t There exist " << coreCentroids.size() / 2 << " cores and " << deltaCentroids.size() / 2 << " deltas." << endl;

        for (int i = 0; i<this->coreCentroids.size(); i = i + 2) {
            cout << "\t c: (" << this->coreCentroids[i] << "," << this->coreCentroids[i + 1] << ")" << endl;
        }
        for (int i = 0; i<this->deltaCentroids.size(); i = i + 2) {
            cout << "\t d: (" << this->deltaCentroids[i] << "," << this->deltaCentroids[i + 1] << ")" << endl;
        }
    }

}

void NewPoincareSingularPoints::cleanSPsWithKaruRules() {
    
    int blockW = w / orW;
    // clean 40 pixels from border

    //cout << "Elmininando..." << endl;
    // clean cores 80 pixels from border
    int i = blocksPerCore.size() - 1;

    bool coreDelete[blocksPerCore.size()];
    bool deltaDelete[blocksPerDelta.size()];
    memset(coreDelete, 0, blocksPerCore.size() * sizeof (bool));
    memset(deltaDelete, 0, blocksPerDelta.size() * sizeof (bool));
    int nCoresDelete = 0;

    //Deleting cores too close to the image boundaries
    while (i >= 0) {
        float iCore = coreCentroids[i * 2];
        float jCore = coreCentroids[i * 2 + 1];

        if ((iCore * blockW <= cleanCoresKaru * ImgUtil::fAdjust) || (jCore * blockW <= cleanCoresKaru * ImgUtil::fAdjust) ||
                (iCore * blockW >= h - cleanCoresKaru * ImgUtil::fAdjust) || (jCore * blockW >= w - cleanCoresKaru * ImgUtil::fAdjust)) {
            //  cout << "Eliminar core " << iCore << ", " << jCore << endl;
            coreDelete[i] = true;
            nCoresDelete++;
        }
        i--;
    }

    //Deleting deltas too close to the image boundaries
    i = blocksPerDelta.size() - 1;

    while (i >= 0) {
        float iDelta = deltaCentroids[i * 2];
        float jDelta = deltaCentroids[i * 2 + 1];

        if ((iDelta * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) || (jDelta * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) ||
                (iDelta * blockW >= h - cleanCoresDeltasKaru * ImgUtil::fAdjust) || (jDelta * blockW >= w - cleanCoresDeltasKaru * ImgUtil::fAdjust)) {
            // cout << "Eliminar delta " << iDelta << ", " << jDelta << endl;
            deltaDelete[i] = true;
        }
        i--;
    }

    // for each core, busco el delta más cercano
    if (blocksPerDelta.size() > 0) {
        i = blocksPerCore.size() - 1;

        while (i >= 0) {
            float iCore = coreCentroids[i * 2];
            float jCore = coreCentroids[i * 2 + 1];

            for (int j = blocksPerDelta.size() - 1; j >= 0; j--) {
                float iDelta = deltaCentroids[j * 2];
                float jDelta = deltaCentroids[j * 2 + 1];


                float dist = sqrt((iDelta - iCore) * (iDelta - iCore) +
                        (jDelta - jCore) * (jDelta - jCore));

                bool eliminar = false;
                if ((iDelta * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) || (jDelta * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) ||
                        (iDelta * blockW >= h - cleanCoresDeltasKaru * ImgUtil::fAdjust) || (jDelta * blockW >= w - cleanCoresDeltasKaru * ImgUtil::fAdjust) ||
                        (iCore * blockW <= cleanCoresKaru * ImgUtil::fAdjust) || (jCore * blockW <= cleanCoresKaru * ImgUtil::fAdjust) ||
                        (iCore * blockW >= h - cleanCoresKaru * ImgUtil::fAdjust) || (jCore * blockW >= w - cleanCoresKaru * ImgUtil::fAdjust)) {
                    if (dist * blockW <= cleanCoreDeltaPairOutKaru * ImgUtil::fAdjust)
                        eliminar = true;
                } else {
                    if (dist * blockW <= cleanCoreDeltaPairInKaru * ImgUtil::fAdjust)
                        eliminar = true;
                }
                if (eliminar) {
                    if (coreDelete[i] == false)
                        nCoresDelete++;
                    coreDelete[i] = true;
                    deltaDelete[j] = true;
                }
            }

            i--;
        }
    }

    bool deleteAll = false;
    for (int i = blocksPerCore.size() - 1; i >= 0; i--) {
        for (int j = blocksPerDelta.size() - 1; j >= 0; j--) {

            if ((!coreDelete[i]) && (!deltaDelete[j])) {
                float iCore = coreCentroids[i * 2];
                float iDelta = deltaCentroids[j * 2];
                float jCore = coreCentroids[i * 2 + 1];
                float jDelta = deltaCentroids[j * 2 + 1];

                if ((iDelta <= iCore) && (( blocksPerCore.size() - nCoresDelete == 1) || (abs(jDelta - jCore) * blockW < 100)   )) {
                    // cout << "iDelta = " << iDelta << " iCore = " << iCore << endl;
                    deleteAll = true;
                }

            }

        }
    }

    for (int i = blocksPerCore.size() - 1; i >= 0; i--) {
        // cout << "coreDelete = " << coreDelete[i] << endl;
        if (coreDelete[i] || deleteAll) {
            coreCentroids.erase(coreCentroids.begin() + i * 2 + 1);
            coreCentroids.erase(coreCentroids.begin() + i * 2);
            blocksPerCore.erase(blocksPerCore.begin() + i);
        }
    }
    for (int i = blocksPerDelta.size() - 1; i >= 0; i--) {

        // cout << "deltaDelete = " << deltaDelete[i] << endl;
        if (deltaDelete[i] || deleteAll) {
            deltaCentroids.erase(deltaCentroids.begin() + i * 2 + 1);
            deltaCentroids.erase(deltaCentroids.begin() + i * 2);
            blocksPerDelta.erase(blocksPerDelta.begin() + i);
        }
    }

    if (deleteAll) {
        if (ImgUtil::DEBUG)
            cout << "deltaCentroids.size() = " << deltaCentroids.size() << endl;
        deltaCentroids.clear();
        coreCentroids.clear();
        blocksPerCore.push_back(-1);
        blocksPerCore.push_back(-1);
        blocksPerCore.push_back(-1);
    }
    // cout << " end eliminar " << endl;
    // if a core está a menos de 8 pixeles en bloques de orientación, 24 en border areas
    // del delta más cercano se borran los dos!


}

void NewPoincareSingularPoints::fakeCoresAndDeltas() {
    cout << "function NewPoincareSingularPoints::fakeCoresAndDeltas" << endl;
    cout << "To be used ONLY FOR ImgUtil::DEBUGGING!!!" << endl;

    cout << "Fake cores and deltas-> Initial: " << endl;
    cout << "\t There exist " << coreCentroids.size() / 2 << " cores and " << deltaCentroids.size() / 2 << " deltas." << endl;

    for (int i = 0; i<this->coreCentroids.size(); i = i + 2) {
        cout << "\t c: (" << this->coreCentroids[i] << "," << this->coreCentroids[i + 1] << ")" << endl;
    }
    for (int i = 0; i<this->deltaCentroids.size(); i = i + 2) {
        cout << "\t d: (" << this->deltaCentroids[i] << "," << this->deltaCentroids[i + 1] << ")" << endl;
    }

    this->coreCentroids.push_back(10);
    this->coreCentroids.push_back(12);
    this->blocksPerCore.push_back(3);

    this->coreCentroids.push_back(18);
    this->coreCentroids.push_back(16);
    this->blocksPerCore.push_back(3);

    this->coreCentroids.push_back(17);
    this->coreCentroids.push_back(16);
    this->blocksPerCore.push_back(3);

    this->coreCentroids.push_back(17.5);
    this->coreCentroids.push_back(16);
    this->blocksPerCore.push_back(3);

    this->coreCentroids.push_back(4);
    this->coreCentroids.push_back(9);
    this->blocksPerCore.push_back(3);

    this->deltaCentroids.push_back(2);
    this->deltaCentroids.push_back(5);
    this->blocksPerDelta.push_back(6);

    this->deltaCentroids.push_back(11);
    this->deltaCentroids.push_back(12);
    this->blocksPerDelta.push_back(6);

    this->deltaCentroids.push_back(15);
    this->deltaCentroids.push_back(14);
    this->blocksPerDelta.push_back(3);

    cout << "Fake cores and deltas-> Final: " << endl;
    cout << "\t There exist " << coreCentroids.size() / 2 << " cores and " << deltaCentroids.size() / 2 << " deltas." << endl;

    for (int i = 0; i<this->coreCentroids.size(); i = i + 2) {
        cout << "\t c: (" << this->coreCentroids[i] << "," << this->coreCentroids[i + 1] << ")" << endl;
    }
    for (int i = 0; i<this->deltaCentroids.size(); i = i + 2) {
        cout << "\t d: (" << this->deltaCentroids[i] << "," << this->deltaCentroids[i + 1] << ")" << endl;
    }

}

void NewPoincareSingularPoints::cleanSPsMinBlock(int minNBlock) {

    //cout << "minBlock = " << minNBlock << endl;
    for (int i = blocksPerCore.size() - 1; i >= 0; i--) {
        if (blocksPerCore[i] <= minNBlock) {
            //cout << "coreDelete = " << i << endl;
            coreCentroids.erase(coreCentroids.begin() + i * 2 + 1);
            coreCentroids.erase(coreCentroids.begin() + i * 2);
            blocksPerCore.erase(blocksPerCore.begin() + i);
        }
    }
    for (int i = blocksPerDelta.size() - 1; i >= 0; i--) {

        if (blocksPerDelta[i] <= minNBlock) {
            //cout << "deltaDelete = " << i << endl;
            deltaCentroids.erase(deltaCentroids.begin() + i * 2 + 1);
            deltaCentroids.erase(deltaCentroids.begin() + i * 2);
            blocksPerDelta.erase(blocksPerDelta.begin() + i);
        }
    }


}

void NewPoincareSingularPoints::cleanSPsWithClusterRules(int minDistance){
    
    vector<Cluster> conjunto;
    vector<float> aux1, aux2;
    int i = -1;
    
    aux1.clear();
    aux2.clear();
    
    //Juntamos los singular points en un conjunto
    for (int i = 0; i < coreCentroids.size() / 2; i++){
        aux1.push_back(coreCentroids[i * 2]);
        aux1.push_back(coreCentroids[i * 2 + 1]);
        
        Cluster obj;
        obj.inicializarCluster(aux1, aux2, blocksPerCore[i], 0);
        
        conjunto.push_back(obj);
        
        aux1.clear();
    }
    for (int i = 0; i < deltaCentroids.size() / 2; i++){
        aux2.push_back(deltaCentroids[i * 2]);
        aux2.push_back(deltaCentroids[i * 2 + 1]);
        
        Cluster obj;
        obj.inicializarCluster(aux1, aux2, 0, blocksPerDelta[i]);
        
        conjunto.push_back(obj);
        
        aux2.clear();
    }
    
    float distToCenter = 0;
    float iCenter = orH / 2.;
    float jCenter = orW / 2.;
//    float iMedia, jMedia;
    float maxDistToCenter = sqrt( (iCenter) * (iCenter) + (jCenter) * (jCenter) );

    bool hayUnion = true;
    float minDist;
    int iC, jC;
    float dist;
    float distP1, distP2;
    while ((hayUnion) && (conjunto.size() > 1)) {
        hayUnion = false;
        minDist = 100000;
        iC = 0; jC = 1;
        for (int i = 0; i < conjunto.size(); i++) {
            for (int j = i + 1; j < conjunto.size(); j++) {
                dist = conjunto[i].calcularDistancia(conjunto[j]);
                
                
                if (minDist > dist) {
                    minDist = dist;
                    iC = i;
                    jC = j;
                }
            }
        }
        
        distP1 = sqrt( (conjunto[iC].mediaX - iCenter) * (conjunto[iC].mediaX - iCenter) + 
                (conjunto[iC].mediaY - jCenter) * (conjunto[iC].mediaY - jCenter) );
        distP2 = sqrt( (conjunto[jC].mediaX - iCenter) * (conjunto[jC].mediaX - iCenter) + 
                (conjunto[jC].mediaY - jCenter) * (conjunto[jC].mediaY - jCenter) );
        distToCenter = (distP1 + distP2) / 2.;
        distToCenter /= maxDistToCenter;
        
        float sigmaZhang = 0.4;
        float gauss = exp(- (distToCenter * distToCenter) / (2. * sigmaZhang * sigmaZhang)); 
        
        if (minDist <= (1. -  gauss) * minDistance ) { 
            conjunto[iC].unirCluster(conjunto[jC]);
            conjunto.erase(conjunto.begin() + jC);
            hayUnion = true;
        }
    }
    
    if (ImgUtil::DEBUG){
        for (i = 0; i < conjunto.size(); i++){
            cout << "Cluster " << i+1 << endl;
            conjunto[i].mostrarElementos();
        }
    }
    
    coreCentroids.clear();
    deltaCentroids.clear();
    blocksPerDelta.clear();
    blocksPerCore.clear();
    
    for (i = 0; i < conjunto.size(); i++){
        if (conjunto[i].numCores == conjunto[i].numDeltas + 1){
            coreCentroids.push_back(conjunto[i].mediaX);
            coreCentroids.push_back(conjunto[i].mediaY);
            blocksPerCore.push_back(abs(conjunto[i].blocks));
        }
        else if (conjunto[i].numCores == conjunto[i].numDeltas - 1){
            deltaCentroids.push_back(conjunto[i].mediaX);
            deltaCentroids.push_back(conjunto[i].mediaY);
            blocksPerDelta.push_back(abs(conjunto[i].blocks));
        }
        else if (conjunto[i].numCores == conjunto[i].numDeltas + 2){
            coreCentroids.push_back(conjunto[i].mediaX);
            coreCentroids.push_back(conjunto[i].mediaY);
            blocksPerCore.push_back(abs(conjunto[i].blocks));
        }
    }
    
    conjunto.clear();
}
