/* 
 * \file        ImageEnhancementCappelli.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     Orientation enhancement performed by Cappelli et al. on the
 *                  orientations map extracted by PCASYS
 * 
 * Header file for ImageEnhancementCappelli class
 *
 * 
 */


#ifndef ORIENTATIONENHANCEMENTCAPPELLI_H
#define	ORIENTATIONENHANCEMENTCAPPELLI_H

/**
 * \class ImageEnhancementCappelli
 *
 * The ImageEnhancementCappelli class implements the orientation image
 *  enhancement procedure in 
 * 
 * R. Cappelli, A. Lumini, D. Maio, D. Maltoni
 * "Fingerprint Classification by Directional Partitioning" 
 * IEEE Transactions on Pattern Analysis and Machine Intelligence, 
 * pp. 402-421, May, 1999
 * 
 * It tries to enhance the structural coherence of the orientations map.
 */
class OrientationEnhancementCappelli {
public:
    
    /** Default constructor */
    OrientationEnhancementCappelli();
    
     /**
     * Constructs an instance of the class provided a orientation map.
     *  The orientation map must be given in the shape of [epsilon*cos(2*theta) epsilon*sin(2*theta)],
     *      as explained in the paper.
     * \param xCosImage Cosine components of the orientations
     * \param ySinImage Sine components of the orientations
     * \param orImageH Height of the orientation map
     * \param orImageW Width of the orientation map
     */
    OrientationEnhancementCappelli(float **xCosImage, float **ySinImage, int orImageH, int orImageW);
    
    /**
     * Main enhancement procedure. It executes the steps described in the paper, but
     *      does not return the enhanced orientations map.
     * 
     * \param sigma Sigma used in the attenuation phase
     * \param gamma Gamma used in the strenghtening phase
     */
    void enhance(float sigma, float gamma);
    
    /**
     * Provides the enhanced sine components. Must be called after enhance();
     * \return 
     */
    float **getSin2Theta();
    
    /**
     * Provides the enhanced cosine components. Must be called after enhance();
     * \return 
     */
    float **getcos2Theta();
    
    /**
     * Default destructor
     */
    virtual ~OrientationEnhancementCappelli();
    
    /**
     * Reguralized the data in "data" using a rectangular, mono-weighted mask
     *  of (2*colHalfSize+1)x(2*rowHalfSize+1) elements.
     * In the corners and boundaries, smaller masks are considered.
     * Data must be a matrix of datarows x datacols elements.
     * \param data  original data
     * \param dataRows  number of rows of the original data.
     * \param dataCols number of cols of the original data.
     * \param colHalfSize half (-0.5) of the mask width.
     * \param rowHalfSize half (-0.5) of the mask height
     * \return the regularized version of the data matrix.
     */
    float **smooth(float **data,int dataRows, int dataCols, int colHalfSize,int rowHalfSize);
    
    /**
     * Calculates the attenuation values. It must be called after the dimensions of the orientation
     *  images are set. The result is stored in an internal attribute.
     * \param sigma Sigma used in the Gaussian function.
     */
    void calculateAttenuation(float sigma);
    
    /**
     * Calculates the Strenghtening rates at each position of the image.
     * \param rowHalfSize half (-0.5 units) of the height of the mask used for regularizing STR
     * \param colHalfSize half (-0.5 units) of the width of the mask used for regularizing STR
     */
    void calculateStrenghthening(int rowHalfSize,int colHalfSize);
    
private:
    
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    
    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    float **atr;        //!< Attenuation rate per position
    float **str;        //!< Strengthening rate per position
    float **smoothSTR;        //!< Strengthening rate per position
    
    float **enhancedXCos;       //!<(Enhanced) cos(2 O) value of the cosine of the orientation (x-direction)
    float **enhancedYSin;       //!<(Enhanced) sin(2 O) value of the sine of the orientation (y-direction)

};

#endif	/* ORIENTATIONENHANCEMENTCAPPELLI_H */

