/* 
 * \file        OrientationFieldMei.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for OrientationFieldMei (Mei2009) class
 *
 * 
 */


#ifndef ORIENTATIONFIELDMEI_H
#define	ORIENTATIONFIELDMEI_H


using namespace std;


/**
 * \class OrientationFieldMei
 *
 * The OrientationFieldMei class implements the orientation map extraction from
 * Mei2009 in Computer Vision and image Understanding, needed for Le PR 2012
 */
class OrientationFieldMei {
public:
    /** Default constructor */
    OrientationFieldMei();
    
    /** It constructs a new instance of the class
     * \param image Fingerprint image from which to extract the orientation map
     * \param h Height of the fingerprint image
     * \param w Width of the fingerprint image
     */
    OrientationFieldMei(unsigned char **image, int h, int w);
    
    /** Default destructor */
    ~OrientationFieldMei();
    
    /**
     * It performs the whole algorithm to compute the orientation field
     * \param W Array containing the block sizes to be used
     * \param varThresholds Array containing the variance thresholds for each scale
     * \return The orientation map
     */
    float **computeOrientationField(int *W, float *varThresholds);
    
    /**
     * It computes the gradients of the image following Mei2009 (not sobel)
     */
    void computeGradients();
    
    /**
     * It computes the first level orientation fields (see Mei2009) using the given
     * block sizes
     * \param W Array containing the block sizes to be used
     * \param coherenceThreshold Coherence threshold to consider a block as valid
     * \param varThreshold Array containing the variance thresholds for each scale
     */
    void computeFirstLevelOrientationFields(int *W, float coherenceThreshold, float *varThreshold);
    
    /**
     * It computes the first level orientation field of the given block size
     * \param W Size of the block
     * \param coherenceThreshold Coherence threshold to consider a block as valid
     * \param GBxNew Pointer to the GBx component of the orientation field (OUTPUT)
     * \param GByNew Pointer to the GBy component of the orientation field (OUTPUT)
     * \return The first level orientation field of the given block size
     */
    float **computeFirstLevelOrientationField(int W, float coherenceThreshold, float ***GBxNew, float ***GByNew);
    
    /**
     * It computes the the variance of the given first level orienation field and the corresponding
     * smooth and non smooth areas using the given threshold
     * \param W Block size of the orientation field
     * \param O Orientation field (it is modified!)
     * \param varThreshold Variance threshold to consider a block as smooth / non-smooth
     * \return The variance of each block (O is also modified!)
     */
    float **computeVarianceAndSmooth_NonSmoothAreas(int W, float **O, float varThreshold);
    
    /**
     * It combines the three first level orientation fields
     */
    void combineFirstLevelOrientationFileds();
    
    /**
     * It iteratively predicts the previously non-estimated areas of the orientation image
     */
    void predictNonEstimatedAreas();
    
    /**
     * It returns the orientation field
     * \return The orientation field
     */
    float **getOrientationField() const;
    
    /**
     * It returns the height of the orientation field
     * \return The height of the orientation field
     */
    int getOrH() const;
    
    /**
     * It returns the width of the orientation field
     * \return The width of the orientation field
     */
    int getOrW() const;
    
private:
    
    float **img;                //!< Reference to the fingerprint image data
    int h;                      //!< Height of the image
    int w;                      //!< Width of the image
    

    /* Pixel Gradients */
    float **Gx;                 //!< Pixel gradient in x
    float **Gy;                 //!< Pixel gradient in y
    float **Gmagnitude;         //!< Gradient magnitude
    
    /* Normalized Gradients */
    float **gx;                 //!< Normalized pixel gradients in x
    float **gy;                 //!< Normalized pixel gradients in y
    float **gmagnitude;         //!< Gradient magnitude
    
    /* Squared Gradients */
    float **gsx;                //!< Squared normalized pixel gradients in x
    float **gsy;                //!< Squared normalized pixel gradients in y 
    
    /* Orientations */
    float ***O;                 //!< First level orientation maps
    float *orH;                 //!< Height of each orientation map
    float *orW;                 //!< Width of each orientation map
    float *blockW;              //!< Block size of each orientation map
    
    float ***GBx;               //!< Final squared pixel gradients (see Mei2009) 
    float ***GBy;               //!< Final squared pixel gradients (see Mei2009)
    
    float ***variance;          //!< Variance of each block in each orientation map
    
    float **combinedO;          //!< The combined orientation map from the three first levels orientations
    float **combinedGBx;        //!< The corresponding Squared gradients of the combined orientation map x
    float **combinedGBy;        //!< The corresponding Squared gradients of the combined orientation map y 
};

inline float **OrientationFieldMei::getOrientationField() const { return combinedO; }
inline int OrientationFieldMei::getOrH() const { return orH[0]; }
inline int OrientationFieldMei::getOrW() const { return orW[0]; }

#endif	/* ORIENTATIONFIELDMEI_H */

