/* 
 * \file        OrientationReconstructionLi.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for OrientationReconstructionLi (Li2008) class
 *
 * 
 */

#ifndef ORIENTATIONRECONSTRUCTIONLI_H
#define	ORIENTATIONRECONSTRUCTIONLI_H


#undef max
#undef min
#include <Dense>


#include "FingerprintImg.h"

#include "ParametersLi.h"

using namespace Eigen;

class OrientationReconstructionLi {
    
public:
    OrientationReconstructionLi();
    
    ~OrientationReconstructionLi();
    
    float *run(string file, PCASYSManager *pcasys, bool segmentationPoincare = false);
    
    void computeOrientations(float **img, int h, int w);
    
    void reconstructOrientations();
    
    int reconstructPieceWiseOrientations(float ***xCosPred, float ***ySinPred, float ***Opred);
    
    void verifySingularPoints(vector<int> &SPs, MatrixXd &allCoefficients, MatrixXd &allPositions, bool atLeast1 = false);
    
    void verifySingularPoints(float **xCosPred, float **ySinPred, float **coherenceTrain, int order, MatrixXd allPositions, MatrixXd allCoefficients);
    
    VectorXd computeConstrainedWeightedLeastSquares(float **xCosPred, float **ySinPred, float **coherenceTrain, int order, MatrixXd allPositions, MatrixXd allCoefficients, float ***Onew = 0);
    
    VectorXd computeWeightedLeastSquares(float **xCosPred, float **ySinPred, float **coherenceTrain, int order); // Non-constrained

    
    VectorXd computeFirstOrderCoefficients(int posI, int posJ, float **Oold, float *D = 0);
    
    void computeFirstOrderOmegas(int posI, int posJ, MatrixXd &omega2, MatrixXd &omega4, float squared, float **Oold, bool **mask = 0);
    
    VectorXd computeCoefficientsFromOmegas(MatrixXd omega2, MatrixXd omega4, int order);
    
    float **reconstructSPZoneOrientationsFromFirstOrderCoefficients(int posI, int posJ, VectorXd coefficients, float ***xCosNew, float ***ySinNew, float squared, bool **mask = 0, int *limits = 0);
    
    float **reconstructSPZoneOrientationsFromFirstOrderCoefficientsABCD(int posI, int posJ, VectorXd coefficients, float ***xCosNew, float ***ySinNew, bool **mask, int *limits = 0);
    
    float **reconstructSPZoneOrientationsFromCoefficients(int posI, int posJ, VectorXd coefficients, int order, float ***xCosNew, float ***ySinNew, float squared);
    
    float **reconstructPieceWiseCore(int posI, int posJ, float ***xCosNew, float ***ySinNew, float *m_line, float core2 = 1);
    
    void reconstructPieceWiseUniqueCore(int posI, int posJ, float ***xCosNew, float ***ySinNew, float ***Onew);
    
    
// Se asume 1 core y un delta, el delta siempre estríctamente debajo del core!!!!!!
    int reconstructPieceWiseCoreDelta(int posIcore, int posJcore, int posIdelta, int posJdelta, float ***xCosNew, float ***ySinNew, float ***Onew);
    
    // Se asumen 2 cores y 2 deltas, los deltas siempre estríctamente debajo del core principal, que es el superior, y el cruce de sus asintotas por debajo de secundario!!!!!!
    // se usan los cores y deltas ya almacenados que deben ser 2 y 2
    int reconstructPieceWise2Core2Delta(float ***xCosNew, float ***ySinNew, float ***Onew);

        
    float **reconstructPieceWiseDelta(int posI, int posJ, float ***xCosNew, float ***ySinNew, int ***returnMask);
    
    float **reconstructDeltaFromAsymptotes(int posI, int posJ, float O1, float O2, float sign1, float sign2, float ***xCosNew, float ***ySinNew, bool ***returnMask);
    
    float **averageOrientations(float ***xCosNew, float ***ySinNew, bool **maskAverageZone);
    
    void formTrainingData(float **xCosPred, float **ySinPred, float **Opred, float **coherenceTrain);
    
    /**
     * It creates a new inclined image in which core direction points down
     * \param matrix fingerprint image
     * \param h fingerprint image's height
     * \param w fingerprint image's width
     * \param direcCore core's original direction
     * \return inclined image
     */
    float **imageAlignement(float **matrix, int h, int w, vector<float> direcCore);
    
    /**
     * It calculates the value of a pixel in the inclined image according to the original image.
     * \param img fingerprint image
     * \param vecinos all the pixels of the original image that participate in the reconstruction of one pixel in the final image
     * \param pesos weights of every pixel that participate in the reconstruction of one pixel in the final image
     * \return a final image pixel value
     */
    float calculateRotation(float **img, vector<int> vecinos, vector<float> pesos);
    
    /**
     * It calculates one pixel's neighbours' weight
     * \param indexI pixel row
     * \param indexJ pixel column
     * \param vecinos neighbours
     * \return neighbours weight vector
     */
    vector<float> extractWeights(float indexI, float indexJ, vector<int> vecinos);
    
    /**
     * It obtains all pixels that enclose a given pixel
     * \param indexI pixel row
     * \param indexJ pixel column
     * \param h original image's height
     * \param w original image's width
     * \return neighbour pixel vector
     */
    vector<int> extractNeighbours(float indexI, float indexJ, int h, int w);
    
    /**
     * It returns the core which is closer to the existing deltas
     * \return 
     */
    int extractCenterCore();
     
    /**
     * It reduces the image dimension according to the center location
     * \param img fingerprint image
     * \param h image's height
     * \param w image's width
     * \param centro vector with core coordinates
     * \param hC new image's height
     * \param wC new image's width
     * \return the reduced image
     */
    float **imageCropping(float **img, int h, int w, vector<int> centro, int hC, int wC);
    
    /**
     * Parting from two images it compares a part of them and give a result of the similarity
     * \param Oorig original orientation image
     * \param Omodel orientation image to compare with
     * \param coherence coherence between images
     * \param orH image's height
     * \param orW image's width
     * \param smallH height of the piece of the image to compare
     * \param smallW width of the piece of the image to compare
     * \param centralH central height
     * \param centralW central width
     * \return a value that represents the similarity
     */
    float orientationDifference(float **Oorig, float **Omodel, float **coherence, int orH, int orW, int smallH, int smallW, int centralH, int centralW);
    
    /**
     * It creates the feature vector
     * \param coefficients
     * \param order
     * \param coreMag
     * \param deltaMag
     * \return 
     */
    float *createFeatureVector(VectorXd coefficients, int order, float coreMag, float deltaMag);
    
    /**
     * It calculates the coherence mean of an image
     * \param coherence coherence image
     * \param orH image's height
     * \param orW image's width
     * \return coherence mean
     */
    float meanCoherence(float **coherence, int orH, int orW);
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:

    float **O;
    float **xCos;
    float **ySin;
    float **coherence;
    
    int blockW;
    int orH;
    int orW;
    
    vector<int> cores;
    vector<int> deltas;
    
    vector<float> coreAngles;
    
    int dimFV;                          //!< Dimension of feature vector
};



inline int OrientationReconstructionLi::getDimFV() const { return dimFV;}

#endif	/* ORIENTATIONRECONSTRUCTIONLI_H */

