/* 
 * \file        ParametersCappelli.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Parameters file for Cappelli fingerprint classifciation by directional image
 * partitioning algorithm (DynamicMasks)
 *
 * 
 */

#ifndef PARAMETERSCAPPELLI_H
#define	PARAMETERSCAPPELLI_H

#include "ImgUtil.h"

/* Parameters specified in the paper */

/* for the preprocessing of the orientation map */
const float sigmaCappelli = 9.6;                        //!< Sigma for the attenuation
const float lambdaCappelli = 112;                       //!< Lambda value for the weighting factor

/* For the computation of the cost */
const float C0Capelli = 10.;                            //!< Penalty proportional to the number of regions
const float muCappelli = 1.6;                           //!< Weight for the cost function

/* Maximum global displacement */
const int displacementLowerBoundCappelli = -8;                  //!< Minimun global displacement
const int displacementUpperBoundCappelli = 8;                   //!< Maximum global displacement

/* Maximum global rotation */
const float rotationLowerBoundCappelli = -20. * PI / 180.;      //!< Minimum global rotation
const float rotationUpperBoundCappelli = 20. * PI / 180.;       //!< Maximum global rotation
const float rotationStepCappelli = 5. * PI / 180.;              //!< Rotation step size



/* Masks definition, manually extrated from the paper */
/* The distances are in proportion, from [0 1] */
/* Arch */
const float p1p2archCappelli = 0.3;                             //!< Distance from p1 to p2 (arch)
const float p2p3archCappelli = 0.4;                             //!< Distance from p2 to p3 (arch) 
const float toptop5archCappelli = 0.25;                         //!< Distance from the top to p5 (arch) 
const float lefttop5archCappelli = 0.43;                        //!< Distance from the left to p5 (arch) 
const float p5p7archCappelli = 0.35;                            //!< Distance from p5 to p7 (arch) 
const float p9p11archCappelli = 0.18;                           //!<  Distance from p9 to p11 (arch)

const float relationP1P2archCappelli = 37.;                     //!< Angle relatin between region 1 and 2 (arch)
const float relationP2P3archCappelli = relationP1P2archCappelli;        //!< Angle relatin between region 2 and 3 (arch)

/* Common for Left/right loop, tented arc and whorl */
const float p1p2lrtwCappelli = 0.2;                             //!< Distance from p1 to p2 (left, right, tented and whorl)
const float p2p3lrtwCappelli = 0.3;                             //!< Distance from p2 to p3 (left, right, tented and whorl) 
// Hence, p3p4 = 0.3 and p4p5 = 0.2

/* Common for left/right loop and tented arc */
const float p3p6lrtCappelli = 0.37;                             //!< Distance from p3 to p6 (left, right and tented)
const float p7p10lrtCappelli = 0.29;                            //!< Distance from p7 to p10 (left, right and tented)

/* Common for left/right loop */
const float p6mobilityXlrCappelli = 0.13;                       //!< Horizontal mobility window for p6 (left and right loop)
const float p6mobilityYlrCappelli = 0.14;                       //!< Vertical mobility window for p6 (left and right loop)


const float p8mobilityXlrCappelli = 0.2;                        //!< Horizontal mobility window for p8 (left and right loop)
const float p8mobilityYlrCappelli = 0.14;                       //!< Vertical mobility window for p8 (left and right loop)



/* Left Loop */
const float p10p11leftCappelli = 0.76;                          //!< Distance from p10 to p11 (left loop)
const float relationP1P6leftCappelli = 36.;                     //!< Angle relatin between region 1 and 6 (left loop)

/* Right Loop */
const float p10p11rightCappelli = 1. - p10p11leftCappelli;              //!< Distance from p10 to p11 (right loop) 
const float relationP4P5rightCappelli = -relationP1P6leftCappelli;      //!< Angle relatin between region 4 and 5 (right loop)

/* Tented Arc */

const float p8mobilityXtCappelli = p6mobilityXlrCappelli;               //!< Horizontal mobility window for p8 (tented arch)
const float p8mobilityYtCappelli = p6mobilityYlrCappelli;               //!< Vertical mobility window for p8 (tented arch)

const float relationP1P4tentedCappelli = -38.;                  //!< Angle relatin between region 1 and 4 (tented arch)

/* Whorl */
const float p3p6whorlCappelli = 0.4;                            //!< Distance from p3 to p6 (whorl)
const float p7p9whorlCappelli = 0.21;                           //!< Distance from p7 to p9 (whorl)
const float p12p14whorlCappelli = 0.22;                         //!< Distance from p12 to p14 (whorl)
const float bottomtop9Cappelli = 0.29;                          //!< Distance from the bottom to p9 (whorl)


const float p6mobilityXwhorlCappelli = 0.085;                   //!< Horizontal mobility window for p6 (whorl)
const float p6mobilityYwhorlCappelli = 0.14;                    //!< Vertical mobility window for p6 (whorl)


const float p10mobilityXwhorlCappelli = p6mobilityXwhorlCappelli;       //!< Horizontal mobility window for p10 (whorl)
const float p10mobilityYwhorlCappelli = 2. * p6mobilityYwhorlCappelli;  //!< Vertical mobility window for p10 (whorl)


const float p9mobilityXwhorlCappelli = p8mobilityXlrCappelli;           //!< Horizontal mobility window for p9 (whorl)
const float p9mobilityYwhorlCappelli = p8mobilityYlrCappelli;           //!< Vertical mobility window for p9 (whorl)


const float p11mobilityXwhorlCappelli = p9mobilityXwhorlCappelli;       //!< Horizontal mobility window for p11 (whorl)
const float p11mobilityYwhorlCappelli = p9mobilityYwhorlCappelli;       //!< Vertical mobility window for p11 (whorl)

const float relationP1P9whorlCappelli = -30.;                   //!< Angle relatin between region 1 and 9 (whorl)
const float relationP4P9whorlCappelli = -relationP1P9whorlCappelli;     //!< Angle relatin between region 4 and 9 (whorl)

#endif	/* PARAMETERSCAPPELLI_H */

