/* 
 * \file        PseudoridgeTracer.cpp
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of PseudoridgeTracer class
 *
 * 
 */

#include "MemUtil.hpp"
#include "ImgUtil.h"

#include "PseudoridgeTracer.h"

PseudoridgeTracer::PseudoridgeTracer() {
}

PseudoridgeTracer::PseudoridgeTracer(float **orientationsImage,int oriH, int oriW, int oriBlockSize){

    this->orH=oriH;
    this->orW=oriW;
    
    this->oriImage=MemUtil::copy2DMatrix<float>(orientationsImage, orH,orW);
    
    this->centerI=-1;
    this->centerJ=-1;
    this->blockSize=oriBlockSize;
    
    this->blockwiseSegmentation=0;
    
    this->numDiscreteDirections=-1;
    
    this->discreteOriImage=0;
    
    this->pseudoRImage=0;
    this->endOne=0;
    this->endTwo=0;
    
    this->oneTurned=false;
    this->twoTurned=false;
    
}



PseudoridgeTracer::~PseudoridgeTracer() {
    
    MemUtil::freeMatrix(&oriImage);
    MemUtil::freeMatrix(&pseudoRImage);
    MemUtil::freeMatrix(&discreteOriImage);
    MemUtil::freeMatrix(&blockwiseSegmentation);
    MemUtil::freeMatrix(&endOne);
    MemUtil::freeMatrix(&endTwo);
}


void PseudoridgeTracer::tracePseudoridges(int numOfDirections,int maxPRidgeLength, int turningRules){
    
    int w = orW*blockSize;
    int h = orH*blockSize;
    
    if (ImgUtil::DEBUG) {
        cout << "orH = " << orH << " orW = " << orW << "\nblockSize = " << blockSize << endl;    
        cout << "PR:\n\th = " << (float)orW*blockSize << endl << "w = " << (float)orH*blockSize << endl;    
    }
    
    this->numDiscreteDirections=numOfDirections;
    
    if (this->discreteOriImage==0){
        discretizeOrientations(this->numDiscreteDirections);
    }
    
    this->pseudoRImage=MemUtil::create2DMatrix<int>(h,w);
    
    

    for (int launchingSide=1; launchingSide<=2; launchingSide++){
   
        int firstDir=0;
        
        int posDiscreteOriI = centerI / this->blockSize;
        int posDiscreteOriJ = centerJ / this->blockSize;
            
        //Selecting the launching side
        if (launchingSide==1){
            // going to the left
            firstDir=discreteOriImage[posDiscreteOriI][posDiscreteOriJ];
        }
        else{
            // going to the right
            firstDir=(discreteOriImage[posDiscreteOriI][posDiscreteOriJ]+8)%16;
        }
        
        // Tracing vars
        int pointsCounter=1;
        int accumulatedTurn=0; //number of sectors
        
        //Calculate First Position
        int nextPosJ=evolveJ(centerJ,firstDir);
        int nextPosI=evolveI(centerI,firstDir);
        int dir=firstDir;
        
        pseudoRImage[nextPosI][nextPosJ]=1;
        
        //Whether the ridge is within the image limits
        bool inLimits=true; 
        bool turned=false;
        
        //Current position of the trace
        int currentPosJ=-1;
        int currentPosI=-1;
        
        
        while ( (pointsCounter<maxPRidgeLength) && (~turned) && inLimits){
            
            currentPosJ=nextPosJ;
            currentPosI=nextPosI;
            int prevDir=dir;
            
            pseudoRImage[currentPosI][currentPosJ]=1;
            
            posDiscreteOriI=(int)ceil(currentPosI/(int)this->blockSize);
            posDiscreteOriJ=(int)ceil(currentPosJ/(int)this->blockSize);
            int initialDir = (discreteOriImage[posDiscreteOriI][posDiscreteOriJ]+8)%16;
            
            
            //Pick-up the closest direction of the current orientation
            float  prodOfVectorsA=cos((prevDir-initialDir)*PI/8.);
            float prodOfVectorsB=cos((prevDir-((initialDir+8)%16))*PI/8.);
            
            if (acos(prodOfVectorsA)<acos(prodOfVectorsB)){
                dir=initialDir;
            }else{
                dir=(initialDir+8)%16;
            }
            
            nextPosJ=evolveJ(currentPosJ,dir);
            nextPosI=evolveI(currentPosI,dir);
            
            //Sum up the accumulated turn
            if (turningRules==TURNINGRULES_HONG){
                if (dir-prevDir<-8){
                    accumulatedTurn=(accumulatedTurn+dir-prevDir+16);
                }else if(dir-prevDir>8){
                    accumulatedTurn=accumulatedTurn-(-dir+prevDir+16);
                }else{
                    accumulatedTurn=accumulatedTurn+dir-prevDir;
                }
                
                turned=false;
            }
            else if (turningRules==TURNINGRULES_ZHANG){
                
                
                if (dir-prevDir<-8){
                    accumulatedTurn=(accumulatedTurn+dir-prevDir+16);
                }else if(dir-prevDir>8){
                    accumulatedTurn=accumulatedTurn-(-dir+prevDir+16);
                }else{
                    accumulatedTurn=accumulatedTurn+dir-prevDir;
                }
                
                turned=(abs(accumulatedTurn)<numDiscreteDirections);
                
            }
            
            
            //Check whether you are in limits
            if ((min(nextPosJ,nextPosI)<0) || (nextPosJ>= orW * blockSize) || (nextPosI>= orH * blockSize)){
                inLimits=false;
            }
            else if (this->blockwiseSegmentation!=0){
                if ((this->blockwiseSegmentation[posDiscreteOriI][posDiscreteOriJ]==false)){
                    inLimits=false;
                }
            }
            
                
            
            pointsCounter=pointsCounter+1;
            
            
        }
        
        
        if (launchingSide==1){
            //cout << "Accumulated turn in the first Pridge="<<accumulatedTurn*(2*PI)/numDiscreteDirections<<"."<<endl;
            endOne= MemUtil::createVector<int>(2);
            endOne[1]=currentPosJ;
            endOne[0]=currentPosI;
            oneTurned=turned;
            oneExited=~inLimits;
            oneAccTurn=accumulatedTurn*(2.*PI)/numDiscreteDirections;
        }else{
            //cout << "Accumulated turn in the second Pridge="<<accumulatedTurn*(2*PI)/numDiscreteDirections<<"."<<endl;
            endTwo= MemUtil::createVector<int>(2);
            endTwo[1]=currentPosJ;
            endTwo[0]=currentPosI;
            twoTurned=turned;
            twoExited=~inLimits;
            twoAccTurn=accumulatedTurn*(2.*PI)/numDiscreteDirections;
        }
        
    }
        
    return;
}


void PseudoridgeTracer::discretizeOrientations(int numOfOrientations){
    
    this->discreteOriImage=MemUtil::create2DMatrix<int>(orH,orW);
    
    //Orientations are assumed to be [-pi/2,pi/2]
    //numOfOrientations should (MUST!) be divisible by 4
    
    for (int i=0;i<orH;i++){
        for (int j=0; j<orW; j++){
            
            //Dirnum is the one initially representing -pi
            int dirNum= 0.75*numOfOrientations;
            //(PI/numOfOrientation) = half of one of the (numOfOrientations) sections amplitude
            float angleLimit=-(PI/2.)+(PI/(float)numOfOrientations);
            while(angleLimit<oriImage[i][j]){
                dirNum=(dirNum+1)%numOfOrientations;
                angleLimit+=+((2.*PI)/(float)numOfOrientations);
            }
            
            this->discreteOriImage[i][j]=dirNum;
        }
    }
    
    
    return;
            
}

int PseudoridgeTracer::evolveI(int oldI, int dir){
    
    if (ImgUtil::DEBUG && (this->numDiscreteDirections!=16)){
        cout << "Warning, PseudoridgeTracer::evolveJ only prepared for 16 discrete directions"<<endl;
    }
    
    int newI;
    
    switch (dir){
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:{
            newI=oldI-2;
            break;
        }
        case 1:
        case 7:{
            newI=oldI-1;
            break;
        }
        case 9:
        case 15:{
            newI=oldI+1;
            break;
        }
        case 10:
        case 11:
        case 12:
        case 13:
        case 14:{
            newI=oldI+2;
            break;
        }
        default:{
            newI=oldI;
            break;
        }
    }
    
    return newI;
    
}

int PseudoridgeTracer::evolveJ(int oldJ,int dir){
    
    if (ImgUtil::DEBUG && (this->numDiscreteDirections!=16)){
        cout << "PseudoridgeTracer: Warning, PseudoridgeTracer::evolveI only prepared for 16 discrete directions"<<endl;
    }
    
    int newJ;
    
    switch (dir){
        case 2:
        case 1:
        case 0:
        case 15:
        case 14:{
            newJ=oldJ+2;
            break;
        }
        case 3:
        case 13:{
            newJ=oldJ+1;
            break;
        }
        case 5:
        case 11:{
            newJ=oldJ-1;
            break;
        }
        case 6:
        case 7:
        case 8:
        case 9:
        case 10:{
            newJ=oldJ-2;
            break;
        }
        default:{
            newJ=oldJ;
            break;
        }
    }
    
    return newJ;
    
}

void PseudoridgeTracer::setCenter(int posI, int posJ){
    this->centerI=posI;
    this->centerJ=posJ;
}


void PseudoridgeTracer::setBlockwiseSegmentation(bool **segmentation){
    
    this->blockwiseSegmentation=MemUtil::copy2DMatrix<bool>(segmentation, orH,orW);

    
    return;
}


int PseudoridgeTracer::getNumberOfTurns(){
     
    // Combining both turns
    float totalTurn= abs(oneAccTurn-twoAccTurn);
    int numOfTurns= floor(abs(totalTurn/(5.*PI/6.))*1.01);
   
    //cout << "Number of turns ("<<oneAccTurn<<","<<twoAccTurn<<")->"<<totalTurn<<"->"<<numOfTurns<<endl;
    return numOfTurns;//How many complete half-way turns
    
}

/*
 * INFORMATION RETRIEVAL
 * 
 */

int PseudoridgeTracer::getCenterI(){return this->centerI; }
int PseudoridgeTracer::getCenterJ(){return this->centerJ;}
int **PseudoridgeTracer::getPseudoridge(){return this->pseudoRImage;}
int *PseudoridgeTracer::getFirstEnd(){return this->endOne;}
int *PseudoridgeTracer::getSecondEnd(){return this->endTwo;}
bool PseudoridgeTracer::firstPRidgeTurned(){return this->oneTurned;}
bool PseudoridgeTracer::secondPRidgeTurned(){return this->twoTurned;}

bool PseudoridgeTracer::firstPRidgeExited(){return this->oneExited;}
bool PseudoridgeTracer::secondPRidgeExited(){return this->oneExited;}

 
