/* 
 * \file        SPAnglesNyongesa.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for SPAnglesNyongesa (Nyongesa2004) class
 *
 * 
 */


#ifndef SPANGLESNYONGESA_H
#define	SPANGLESNYONGESA_H


#include "PCASYSManager.h"

class SPAnglesNyongesa {
public:
    /* Default constructor */
    SPAnglesNyongesa();
    
    /**
     * Constructor of the class
     * @param image Image to be processed in the feature extraction
     * @param h Height of the image
     * @param w Widht of the image
     */
    SPAnglesNyongesa(float **image, int h, int w);
    
    /* Default destructor */
    ~SPAnglesNyongesa();

    /**
     * Executes the singular points Nyongesa method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * It computes the feature vector following Nyongesa method
     * \return The feature vector
     */
    float *computeFeatures();
    
    /**
     * It binarizes the fingerprint image
     * \param W The size of the window to be used in the binarization
     */
    void binarizeImage(int W);
    
    /**
     * It creates the orientation map following Nyongesa method (merthe98)
     * \param blockW the size of the block for the orientation computation
     * \param Wsmooth Size of the window to smooth the orientation map after its computation
     */
    void createOrientationImage(int blockW, int Wsmooth); 
    
    /**
     * It computes the orientation map using Nyongesa method starting from the
     * pixel orientation, creating a block orientation map from them, and smoothing 
     * this map at the end
     * \param blockW Size of the window for the pixel orientation computation
     * \param blockW2 Size of the block for the orientation map computatioon
     * \param Wsmooth Size of the window for the smoothing phase
     */
    void createOrientationImagePerPixel(int blockW, int blockW2, int Wsmooth);
    
    /**
     * It computes the overall orientation in a overallWxoverallW centered window
     * \param overallW Size of the window for the overall orientation computation
     * \return The overal orientation in the window
     */
    float computeOverallOrientation(int overallW);
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:
    float **img;                //!< Reference to the fingerprint image data
    
    bool **binarizedImg;           //!< Binarized image
    
    int h;                      //!< Height of the image
    int w;                      //!< Width of the image
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    int blockW;         //!< Block size used to compute the orientation map
    
    
    float **O;          //!< Orientation of each block (in radians)
    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    
    int dimFV;                          //!< Dimension of feature vector
};

inline int SPAnglesNyongesa::getDimFV() const { return dimFV;}

#endif	/* SPANGLESNYONGESA_H */

