/* 
 * \file        SPClassificationKaru.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of SPClassificationKaru (Karu96) class
 *
 * 
 */

#include "ImgUtil.h"
#include "MemUtil.hpp"

#include "FingerprintImg.h"
#include "NewPoincareSingularPoints.h"
#include "SPClassificationKaru.h"

#include "ParametersKaru.h"
#include "ParametersPoincare.h"


SPClassificationKaru::SPClassificationKaru() {
    img = 0;
    h = -1;
    w = -1;
    
    ySin = 0;
    xCos = 0;
    O = 0;
    pixelO = 0;
    segmentation = 0;
}



SPClassificationKaru::SPClassificationKaru(float **image, int h, int w) {
    
    img = MemUtil::copy2DMatrix<float>(image, h, w);
        
    this->h = h;
    this->w = w;
    
    ySin = 0;
    xCos = 0;
    O = 0;
    pixelO = 0;
    segmentation = 0;
}

SPClassificationKaru::~SPClassificationKaru() {
    
    MemUtil::freeMatrix(&img);    
    MemUtil::freeMatrix(&O);    
    MemUtil::freeMatrix(&xCos);    
    MemUtil::freeMatrix(&ySin);    
    MemUtil::freeMatrix(&pixelO);
    MemUtil::freeMatrix(&segmentation);
    
}

float *SPClassificationKaru::run(string file, PCASYSManager *pcasys){
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){    
        fi.setPCASYSManager(pcasys);   
        fi.readFingerprintImage(file);    
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }

    h = fi.getH();
    w = fi.getW();
   
    img = MemUtil::copy2DMatrix<float>(fi.getSegmentedImageFloat(), h, w);

    float *fv = computeFeatures();
    dimFV = 3;
    
    return fv;
    
}
void SPClassificationKaru::setOrientationData(float **xCos, float **ySin, int orH, int orW, int blockW) {
    this->orH = orH;
    this->orW = orW;
    this->blockW = blockW;
    this->xCos = MemUtil::create2DMatrix<float>(orH, orW);
    this->ySin = MemUtil::create2DMatrix<float>(orH, orW);
    this->O = MemUtil::create2DMatrix<float>(orH, orW);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            this->xCos[i][j] = xCos[i][j];
            this->ySin[i][j] = ySin[i][j];
            
            O[i][j] = 0.5 * atan2(ySin[i][j], xCos[i][j]);
        }
    }
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationSinCosCopiedPCASYS";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
}

void SPClassificationKaru::setImageData(float **img, int h, int w) {
    this->img = MemUtil::copy2DMatrix(img, h, w);
    this->h = h;
    this->w = w;
}

void SPClassificationKaru::setSegmentation(bool **segmentation) {
    this->segmentation = MemUtil::copy2DMatrix<bool>(segmentation, orH, orW);
}

float SPClassificationKaru::computeCoresAndDeltas(vector<float> *cores, vector<float> *deltas, bool findOnlyCore) {
    bool poincareSegmentation = true;
    if (segmentation == 0) {
        segmentation = MemUtil::create2DMatrix<bool>(orH, orW, true);
    }
    else {
        h = blockW * orH;
        w = blockW * orW;
        poincareSegmentation = false;
    }
    
    float centerI = ((float)orH - 1.) / 2.;
    float centerJ = ((float)orW - 1.) / 2.;
    // dividir xCos e ySin por la distancia al centro
    float quality = 0;
    float maxDistToCenter =  sqrt( centerI * centerI + centerJ * centerJ );
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            if (( (i + 1) * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) || ( i * blockW >= (blockW * orH) - cleanCoresDeltasKaru * ImgUtil::fAdjust) ||
                    ( (j + 1) * blockW <= cleanCoresDeltasKaru * ImgUtil::fAdjust) || ( j * blockW >= (blockW * orW) - cleanCoresDeltasKaru * ImgUtil::fAdjust)  ) {
                    O[i][j] = 0;
                    xCos[i][j] = 1;
                    ySin[i][j] = 0;
                    segmentation[i][j] = false;
                    quality += 1;
                }
            else {
            
                float distToCenter = sqrt( (centerI - i) * (centerI - i) + (centerJ - j) * (centerJ - j) ) / maxDistToCenter;
                quality += sqrt(xCos[i][j] * xCos[i][j] + ySin[i][j] * ySin[i][j]);
                xCos[i][j] *= (1. - distToCenter);
                ySin[i][j] *= (1. - distToCenter);
                //segmentation[i][j] = true;
            }
            
            if (( (i ) * blockW >= cleanCoresDeltasKaru * ImgUtil::fAdjust) && ( (i - 1) * blockW <= (blockW * orH) - cleanCoresDeltasKaru * ImgUtil::fAdjust) &&
                    ( (j ) * blockW >= cleanCoresDeltasKaru * ImgUtil::fAdjust) && ( (j  - 1) * blockW <= (blockW * orW) - cleanCoresDeltasKaru * ImgUtil::fAdjust)  ) {
                
                //segmentation[i][j] = true;
            }
        }
    }
    quality /= (orH * orW);
    
     
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationSinCosSmoothDistToCenter";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    if (TEST_1){
        string fileName = ImgUtil::makeTestFileName("OrientationMap");
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName);
    }
    
    
    
    NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(O, xCos, ySin, orH, orW);
    if (poincareSegmentation) {
        npsp->generateSegmentation(img, h, w);
        npsp->combineWithSegmentation(segmentation);
    }
    else {
        npsp->setBlockSegmentation(segmentation);
        npsp->setHeight(h);
        npsp->setWidth(w);
    }
        
    int nSmooth = npsp->computeCoresAndDeltas(curveWKaru, smoothWPoincareKaru, poincareModeKaru, rulesToBeAppliedKaru, minNBlocksKaru, normalizeKaru); 
    int smoothSize = smoothWPoincareKaru;
    while( (npsp->getnCores() != npsp->getnDeltas()) ||  ((npsp->getnCores() > 2) || (npsp->getnDeltas() > 2)) ) { 
        
        if ((npsp->getnCores() == 2) && (npsp->getnDeltas() < 2))
                break;
        if (findOnlyCore && (npsp->getnCores() <= 2))
            break;
        
        nSmooth = npsp->computeCoresAndDeltas(curveWKaru, smoothSize, poincareModeKaru, rulesToBeAppliedKaru, minNBlocksKaru, normalizeKaru); 
        
        if (nSmooth >= maxSmoothPoincare)
            smoothSize+=2;
    }
    
    MemUtil::freeMatrix(&O);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    O = MemUtil::copy2DMatrix(npsp->getOrientationMap(), orH, orW);
    xCos = MemUtil::copy2DMatrix(npsp->getXCos(), orH, orW);
    ySin = MemUtil::copy2DMatrix(npsp->getYSin(), orH, orW);
    
    for (int i = 0; i < npsp->getCoreCentroids().size(); i++) {
        cores->push_back(npsp->getCoreCentroids()[i]);
    }
    for (int i = 0; i < npsp->getDeltaCentroids().size(); i++)
        deltas->push_back(npsp->getDeltaCentroids()[i]);
    
    
    delete npsp;
    
    
    return quality + nSmooth;
}

float *SPClassificationKaru::computeFeatures() {
    
    vector<float> cores;
    vector<float> deltas;
    
    createOrientationImage(orienationMaskSizeKaru , (int)(blockWKaru * ImgUtil::fAdjust), smoothWKaru); // block size = 9x9 per pixel 8x8 block orientation y 3x3 smooth
    float quality = computeCoresAndDeltas(&cores, &deltas);
    int nSmooth = floor(quality);
    quality -= nSmooth;

    
    string type = "-1";   
        
    int nCores = cores.size() / 2;
    int nDeltas = deltas.size() / 2;
    
    if (TEST_1){
        string fileName = ImgUtil::makeTestFileName("SingularPoints");
        ImgUtil::writeSPImage(img, h, w, deltas, cores, blockW, fileName); 
    }
    
    if ( ((nCores == 0) && (nDeltas == 0))) // (nSmooth > maxSmoothPoincare) ||
        type = "A";
    else if ((nCores == 2) && (nDeltas <= 2))  // antes && ==
        type = "W";
    else { // loop or tented
        float iCore = cores[0];
        float jCore = cores[1];
        float iDelta = ceil(deltas[0]);
        float jDelta = round(deltas[1]);
        
        float beta = atan2(iCore - iDelta, jDelta - jCore);
        float beta2 = beta;
        if (beta < - PI / 2.)
            beta += PI;
        else if (beta > PI / 2.)
            beta -= PI;
        
        int n = 0;
        int i = round(iCore)+1;
        if (round(iCore) >= iDelta)
            i--;
        int j = round(jCore);
        if ( (sin(beta) <= sin(PI / 4.)) && (sin(beta) > 0))
            j--;
        else if ((sin(beta) > sin(-PI / 4.)) && (sin(beta) < 0))
            j++;
        float iF = i;
        float jF = j;
        float avSum = 0;
        while ( (i != iDelta) || (j != jDelta) ) {
            
            avSum += abs(sin(O[i][j] - beta));
            iF -= sin(beta2);
            jF += cos(beta2);
            if (i != iDelta)
                i = ceil(iF);
            if (j != jDelta)
                if (beta > 0)
                    j = floor(jF);
                else
                    j = ceil(jF);
            iF = i;
            jF = j;
            n++;
        }
        if (n != 0)
            avSum /= (float)n;
        if (abs(avSum) < thresholdTentedArcKaru)
            type = "T";
        else {
            type = "Loop";
            
            i = round(iCore)+1;
            j = round(jCore);
            iF = i;
            jF = j;
            n = 0;
            while ( (i > 0) && (j > 0) && (i < orH - 1) && (j < orW - 1) && (n < 50) && (abs(jCore-j)<20)) {
                iF += abs(sin(O[i][j]));
                if (O[i][j] > 0)
                    jF -= cos(O[i][j]);
                else
                    jF += cos(O[i][j]);
                i = round(iF);
                j = round(jF);
                n++;
            }
            float dif = (i - iCore) * (jDelta - jCore) - (j - jCore) * (iDelta - iCore);
            if (dif < 0)
                type = "R";
            else
                type = "L";
        }
    }
    float *featureVector = MemUtil::createVector<float>(3);
    featureVector[1] = nSmooth;
    featureVector[2] = quality;
    if (type == "W")
        featureVector[0] = 0;
    else if (type == "A")
        featureVector[0] = 1;
    else if (type == "T")
        featureVector[0] = 2;
    else if (type == "R")
        featureVector[0] = 3;
    else if (type == "L")
        featureVector[0] = 4;
    else
        featureVector[0] = -1;
    
    
    
//    delete npsp;
    
    
    return featureVector;
}



void SPClassificationKaru::createOrientationImage(int blockW, int Wsmooth, int Wsmooth2) {

    this->blockW = Wsmooth;
 
    orH = h / Wsmooth;
    orW = w / Wsmooth;
 
    pixelO = MemUtil::create2DMatrix<float>(h, w);

    O = MemUtil::create2DMatrix<float>(orH, orW);
  
    xCos = MemUtil::create2DMatrix<float>(orH, orW);

    ySin = MemUtil::create2DMatrix<float>(orH, orW);

    int masc[blockW][blockW];
    int halfBlock = blockW / 2;
    for (int i = 0; i < blockW; i++) {
        for (int j = 0; j < blockW; j++) {
            masc[i][j] = -1;
            if ((i == halfBlock) && (j == halfBlock))
                masc[i][j] = -1;
            else if ((i == halfBlock) && (j % 2 == 0))
                masc[i][j] = 0;
            else if ((j == halfBlock) && (i % 2 == 0))
                masc[i][j] = 4;
            else if ((i == j) && (i % 2 == 0))
                masc[i][j] = 6;
            else if ((i + j == blockW - 1) && (i % 2 == 0))
                masc[i][j] = 2;
            else if (i == 2 * j - 4)
                masc[i][j] = 5;
            else if (i == -2 * j + 12)
                masc[i][j] = 3;
            else if (j == 2 * i - 4)
                masc[i][j] = 7;
            else if (j == -2 * i + 12)
                masc[i][j] = 1;
        }
    }
   
    /* Compute Orientations */
    for (int i = halfBlock; i < h - halfBlock; i++) {
        for (int j = halfBlock; j < w - halfBlock; j++) {
            float s[8] = {0, 0, 0, 0, 0, 0, 0, 0};
            for (int u = i - halfBlock; u < i + halfBlock + 1; u++) {
                for (int v = j - halfBlock; v < j + halfBlock + 1; v++) {
                    int valMasc = masc[(u - i + halfBlock)][(v - j + halfBlock)];
                    if (valMasc != -1) 
                            s[valMasc] += img[u][v];        
                }
            }
            
            int sp = s[0];
            int sq = s[0];
            int p = 0;
            int q = 0;
            float tot  = 0;
//            bool iguales = false;
            for (int k = 0; k < 8; k++) {
                if (s[k] < sp) {
                    sp = s[k];
                    p = k;
//                    iguales = false;
                }
                else if (s[k] > sq) {
                    sq = s[k];
                    q = k;
//                    iguales = false;
                }
//                else if ((s[k] == sp) || (s[k] == sq))
//                    iguales = true;
                tot += s[k];
            }
            if (4. * img[i][j] + sq + sp > 3./8. * tot) // en el paper está mal la fórmula, ver PCASYS
                pixelO[i][j] = q * PI / 8.;     // valley
            else
                pixelO[i][j] = p * PI / 8.;     // ridge
            
//            if (iguales)
//                pixelO[i][j] = -1;
        }
    }

    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {            
            for (int u = i * Wsmooth; u < (i + 1) * Wsmooth; u++) {
                for (int v = j * Wsmooth; v < (j + 1) * Wsmooth; v++) {
//                    if (pixelO[i][j] != -1) {
                        xCos[i][j] += cos(2. * pixelO[u][v]);
                        ySin[i][j] += sin(2. * pixelO[u][v]);
//                    }                    
                }
            }
            xCos[i][j] /= (float)(Wsmooth * Wsmooth);
            ySin[i][j] /= (float)(Wsmooth * Wsmooth);
                    
            O[i][j] = 0.5 * atan2(ySin[i][j], xCos[i][j]);
            
            if (O[i][j] > PI / 2.)
                O[i][j] -= PI;            
            float mag = sqrt(xCos[i][j] * xCos[i][j] + ySin[i][j] * ySin[i][j]);
            xCos[i][j] = cos(2.0 * O[i][j]) * mag;
            ySin[i][j] = sin(2.0 * O[i][j]) * mag;            
        }
    }
    if (ImgUtil::DEBUG) {
        string fileName2; 
        fileName2 = "OrientationSinCos";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    /* Smooth orientations */
    ImgUtil::smoothOrientations(O, &xCos, &ySin, orH, orW, Wsmooth2, false);
    
    if (ImgUtil::DEBUG) {        
        string fileName2 = "OrientationSinCosSmooth";
        ImgUtil::writeOrImg(xCos, ySin, orW, orH, fileName2);
    }
    
    
    
}