/* 
 * \file        SPAnglesNyongesa.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for SPClassificationKaru (Karu96) class
 *
 * 
 */

#ifndef SPCLASSIFICATIONKARU_H
#define	SPCLASSIFICATIONKARU_H

#include "PCASYSManager.h"

class SPClassificationKaru {
public:
    /* Default constructor */
    SPClassificationKaru();
    
    /**
     * It constructs a new instance of the class
     * \param image Fingerpritn image to be processed
     * \param h Height of the image
     * \param w Width of the image
     */
    SPClassificationKaru(float **image, int h, int w);
    
    /* Default destructor */
    ~SPClassificationKaru();

    /**
     * Executes the singular points Karu method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
    
    /**
     * It sets the orientation data for the computation of the cores and deltas using
     * a different orientation map from the one computed in Karu
     * \param xCos x-component of the orientation map (cos(2 theta))
     * \param ySin y-component of the orientation map (sin(2 theta))
     * \param orH Height of the orientation map
     * \param orW Widht of the orientation map
     * \param blockW Block size of the orientation map blocks
     */
    void setOrientationData(float **xCos, float **ySin, int orH, int orW, int blockW);
    
    /**
     * It sets the image data for the computation of the cores and deltas using
     * a different orientation map from the one computed in Karu
     * \param img The image to be processed
     * \param h Height of the image
     * \param w Widht of the image
     */
    void setImageData(float **img, int h, int w);
    
    void setSegmentation(bool **segmentation);

    /**
     * It computes the cores and deltas of the fingerprint using the karu method
     * \param cores Cores found in the image
     * \param deltas Deltas found in the image
     * \param findOnlyCore Whether to stop despite no deltas have been found (for its use outside karu class)
     * \return The fingerprint quality and the number of smooths needed (their sum)
     */
    float computeCoresAndDeltas(vector<float> *cores, vector<float> *deltas, bool findOnlyCore = false);
    
    /**
     * It computes the feature vector
     * \return The feature vector
     */
    float *computeFeatures();
    
    /**
     * It creates the orientation map with the slits sums method
     * @param blockW    Size of the masc for the slits sums
     * @param Wsmooth   Size of the block for the computation of the orientation map
     * @param Wsmooth2  Size of the window for the smoothing of the orientation map
     */    
    void createOrientationImage(int blockW, int Wsmooth, int Wsmooth2); 
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:
    float **img;                //!< Reference to the fingerprint image data
    
    
    
    int h;                      //!< Height of the image
    int w;                      //!< Width of the image
    
    int orH;            //!< Height of the orientation map
    int orW;            //!< Width of the orientation map
    int blockW;         //!< Block size used to compute the orientation map
    
    float **pixelO;
    float **O;          //!< Orientation of each block (in radians)
    float **xCos;       //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;       //!< sin(2 O) value of the sine of the orientation (y-direction)
    bool **segmentation; //!< segmentation of the image
    
    int dimFV;                          //!< Dimension of feature vector
};

inline int SPClassificationKaru::getDimFV() const { return dimFV;}

#endif	/* SPCLASSIFICATIONKARU_H */

