/* 
 * \file        SymVarReferencePointLe.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of SymVarReferencePointLe (Le2012) class
 *
 * 
 */


#include <cmath>
#include <vector>


#include "SymVarReferencePointLe.h"
#include "FingerprintImg.h"
#include "OrientationFieldMei.h"
#include "NewPoincareSingularPoints.h"
#include "ImgUtil.h"
#include "MemUtil.hpp"
#include "ParametersLe.h"


SymVarReferencePointLe::SymVarReferencePointLe() {
    orH = -1;
    orW = -1;
    O = 0;
    
    coreX = -1;
    coreY = -1;
    referenceDirection = -1;
    
    sym2 = 0;           
    VORIV = 0;           
    VSCOME = 0;
}

SymVarReferencePointLe::SymVarReferencePointLe(float **Omap, int orH, int orW) {
    this->orH = orH;
    this->orW = orW;
    O = MemUtil::create2DMatrix<float>(orH, orW);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            O[i][j] = Omap[i][j];
            if (O[i][j] > PI /2.) 
                O[i][j] -= PI;
        }
    }
    
    coreX = -1;
    coreY = -1;
    referenceDirection = -1;
    sym2 = 0;           
    VORIV = 0;           
    VSCOME = 0;
}

SymVarReferencePointLe::~SymVarReferencePointLe() {
    
    MemUtil::freeMatrix(&O);

    MemUtil::freeMatrix(&sym2);
    
    MemUtil::freeMatrix(&VORIV);

    MemUtil::freeMatrix(&VSCOME);

}

float *SymVarReferencePointLe::run(string file, PCASYSManager *pcasys, bool segmentar){
    FingerprintImg fi;
    float **fingercode;

    if (ImgUtil::fAdjust != 1){ 
        fi.setPCASYSManager(pcasys);  
        fi.readFingerprintImage(file);      
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    OrientationFieldMei *ofm = new OrientationFieldMei(fi.getSegmentedImage(), fi.getH(), fi.getW());

    int *W = MemUtil::createVector<int>(3);
    W[0] = wLeMei[0];// * ImgUtil::fAdjust;
    W[1] = wLeMei[1];// * ImgUtil::fAdjust;
    W[2] = wLeMei[2];// * ImgUtil::fAdjust;
    float *varThreshold = MemUtil::createVector<float>(3);
    varThreshold[0] = varThresholdLeMei[0];
    varThreshold[1] = varThresholdLeMei[1];
    varThreshold[2] = varThresholdLeMei[2];

    ofm->computeOrientationField(W, varThreshold);

    bool **masc = MemUtil::create2DMatrix<bool>(ofm->getOrH(), ofm->getOrW());
    float **Omap = ofm->getOrientationField();

    for (int i = 0; i < ofm->getOrH(); i++) {
        for (int j = 0; j < ofm->getOrW(); j++) {
            if (Omap[i][j] == -1)
                masc[i][j] = false;
            else
                masc[i][j] = true;
        }
    }
    
    if (segmentar){
        NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(fi.getSegmentedImageFloat(),  fi.getH(), fi.getW());
        npsp->setOrH(ofm->getOrH());
        npsp->setOrW(ofm->getOrW());
        npsp->generateSegmentation(); //fi.getSegmentedImageFloat(),  fi.getH(), fi.getW());
        npsp->combineWithSegmentation(masc);
        MemUtil::freeMatrix(&masc);
        masc = MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(), ofm->getOrH(), ofm->getOrW());
        delete npsp;
        
        for (int i = 0; i < ofm->getOrH(); i++) {
            for (int j = 0; j < ofm->getOrW(); j++) {
                if (masc[i][j] == false)
                    Omap[i][j] = -1;
            }
        }
    }
    
    orH = ofm->getOrH();
    orW = ofm->getOrW();
    O = MemUtil::create2DMatrix<float>(orH, orW);
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            O[i][j] = Omap[i][j];
            if (O[i][j] > PI /2.) 
                O[i][j] -= PI;
        }
    }
    
    computeReferencePoint();

    fi.setCore(getCoreX() * W[0] + W[0] / 2, getCoreY() * W[0] + W[0] / 2); //CENTER_DISPLACEMENT_ROWS); // no desplazar

    /* Correct the rotation of the fingerprint */
    int R = getReferenceDirection();

    int auxGSX = gaborSupportXLe * ImgUtil::fAdjust;
    if (auxGSX % 2 == 0)
        auxGSX--;
    
    int auxGSY = gaborSupportYLe * ImgUtil::fAdjust;
    if (auxGSY % 2 == 0)
        auxGSY--;
    
    fingercode = fi.getFingerCode(centerDisplacementLe * ImgUtil::fAdjust, defNumSectorsLe, numBandsLe, bandWithLe * ImgUtil::fAdjust, numGaborFiltersLe,
            V0Le, M0Le, gaborFrecuencyLe / ImgUtil::fAdjust, gaborDXLe * ImgUtil::fAdjust, gaborDYLe * ImgUtil::fAdjust, auxGSX, auxGSY, R, 1, masc, ofm->getOrH());

    dimFV = fi.getNFilters() * fi.getNumSectors();
    
    float *fv = MemUtil::createVector<float>(dimFV);
    for (int i = 0; i < fi.getNFilters(); i++) {
        for (int j = 0; j < fi.getNumSectors(); j++) {
            fv[i * fi.getNumSectors() + j] = fingercode[i][j];
        }
    }

    MemUtil::freeMatrix(&fingercode);
    MemUtil::freeMatrix(&masc);
    MemUtil::freeMatrix(&W);
    MemUtil::freeMatrix(&varThreshold);    
    delete ofm;

    cout << endl;
    
    
    MemUtil::freeMatrix(&fingercode);    

    return fv;
}

void SymVarReferencePointLe::computeReferencePoint() {
    
    int Ws2 = Ws2Le; //7;
    float sigmaC = sigmaCLe; //1.5;
    computeSymmetryValues(Ws2, sigmaC);
    
    int N = NLe; //4;
    int K = KLe; //5;
    computeVORIVfeature(N, K);
    
    computeVSCOME();
    
    int nOrientations = defNumSectorsLe; // 16;     // k = numDefSectors en Le parameters!!
    int initialLength = initialLengthLe; //4;
    int width = widthLe; //5;
    int maxLength = maxLengthLe; // 15;
    computeReferenceOrientation(nOrientations, initialLength, width, maxLength);
    
}

void SymVarReferencePointLe::computeSymmetryValues(int Ws2, float sigmaC) {
    
    sym2 = MemUtil::create2DMatrix<float>(orH,orW);
    float **Fx = MemUtil::create2DMatrix<float>(orH,orW);
    float **Fy = MemUtil::create2DMatrix<float>(orH,orW);
  
    int halfWs = Ws2 / 2;
    float sum = 0;
    
    float **g = MemUtil::create2DMatrix<float>(Ws2,Ws2);
    // g initialization
    for (int i = 0; i < Ws2; i++) 
        for (int j = 0; j < Ws2; j++) {
            float x = i - halfWs;
            float y = j - halfWs;
            g[i][j] = exp(- (x * x + y * y) / (2. * sigmaC * sigmaC));
            sum += g[i][j];
        }
  
    // Uso v, u y x, y por seguir la notación del paper
    for (int u = halfWs; u < orH; u++) {
        for (int v = halfWs; v < orW - halfWs; v++) {
            int m = (Ws2 * Ws2); 
                   
            for (int x = -halfWs; x < halfWs + 1; x++) { // halfWs + 1
                for (int y = -halfWs; y < halfWs + 1; y++) { //v + halfWs + 1
                    if ((y < 0)) {
                   
                        if (((O[u][v] != -1) && (O[u + y][v + x] != -1))){ //&& ((x < 0) ||  (y < 0) )) { //(x < 0) || 
                        
                        float r = sqrt(x * x + y * y);
                        float ybarra = (2. * x * y) / r;        // están cambiados!
                        float xbarra = (x * x - y * y) / r;
                        
                        float ang = O[u + y][v + x];// -PI / 2.; //-
                        Fx[u][v] += g[x + halfWs][y + halfWs]  *
                                 ( (xbarra * cos(2. * ang) ) - 
                                    (ybarra * sin(2. * ang)) );
                        Fy[u][v] += g[x + halfWs][y + halfWs]  *
                                 ( (ybarra * cos(2. * ang) ) + 
                                    (xbarra * sin(2. * ang)) );
                        
                    }
                    else 
                        m = 1000;
                }
                }
            }
            if (m > 0)
                sym2[u][v] = (1. / (float) (m)) * sqrt(Fx[u][v] * Fx[u][v] + Fy[u][v] * Fy[u][v]);
        }
    }
    
        
    if (ImgUtil::DEBUG) {
        string fileName = "SymVarSymmetryValues";
        ImgUtil::writeGrayScaleImage(sym2, orH, orW, fileName, 255 * 2);
    }
    
    MemUtil::freeMatrix(&g);
    MemUtil::freeMatrix(&Fx);
    MemUtil::freeMatrix(&Fy);

}


void SymVarReferencePointLe::computeVORIVfeature(int N, int K) {
    
    VORIV = MemUtil::create2DMatrix<float>(orH,orW);
    
    int halfK = K / 2;
    
    for (int u = N-1; u < orH; u++) {
        for (int v = halfK; v < orW - halfK; v++) {
            
            
     //   u = 5; 
    //    v = 25;
            int n = 0;
            bool fin = false;
            for (int k = -halfK; k < halfK + 1; k++) {
                if (fin) break;
                float VORIVaux = 0;
                for (int i = 0; i < N - 1; i++) {
                    if ( (O[u - i][v + k] != -1) && (O[u - i - 1][v + k] != -1 )) {
                        float fx = O[u - i][v + k] - O[u - i - 1][v + k] ;
                        
                        if (abs(fx) <= (PI / 2.)) 
                            fx = fx;
                        else if (fx > (PI / 2.))
                            fx = -(PI - fx);
                        else if (fx < (-PI / 2.))
                            fx = PI + fx;

                        VORIVaux += fx;
                        n++;
                    }
                    else {
                        VORIVaux = 0;
                        VORIV[u][v] = 0;
                        n = 0;
                        fin = true;
                        break;
                    }
                }
        //       if (O[u][v + k] > 0)//PI / 2.) 
                if (k < 0)
                    VORIV[u][v] += VORIVaux;
                else if (k > 0)
                    VORIV[u][v] -= VORIVaux;
                else {
                    if (VORIVaux < 0)
                             VORIV[u][v] += VORIVaux;
                        else
                             VORIV[u][v] -= VORIVaux;
                }
            }
            if (n != 0)
                VORIV[u][v] /= (float) (PI * PI * 2 * 2);// * K * N);K * (N-1)
            else
                VORIV[u][v] = 0;
            
        }
    }
    
    if (ImgUtil::DEBUG) {
        float **VORIVprint = MemUtil::create2DMatrix<float>(orH,orW);
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                VORIVprint[i][j] = VORIV[i][j] < 0 ? 0 : VORIV[i][j];
            }
        }
        
        string fileName = "VORIVfeature";
        ImgUtil::writeGrayScaleImage(VORIV, orH, orW, fileName, 255 * 3);
        
        fileName = "VORIVfeaturePrint";
        ImgUtil::writeGrayScaleImage(VORIVprint, orH, orW, fileName, 255 * 3);
        
        MemUtil::freeMatrix(&VORIVprint);
    }   
}

void SymVarReferencePointLe::computeVSCOME() {
    
    VSCOME = MemUtil::create2DMatrix<float>(orH,orW);
    
    float max = -10000;
    int i = -1;
    int j = -1;
    for (int u = 0; u < orH; u++) {
        for (int v = 0; v < orW; v++) {           
            VSCOME[u][v] = 0.5 * (VORIV[u][v] + sym2[u][v]);
            if (VSCOME[u][v] > max) {
                max = VSCOME[u][v];
                i = u;
                j = v;
            }
        }
    }
    
    coreX = i;
    coreY = j;
    if (ImgUtil::DEBUG) {
        cout << "CONVEX CORE in " << i << ", " << j << endl;
        string fileName = "VSCOME";
        ImgUtil::writeGrayScaleImage(VSCOME, orH, orW, fileName, 255 * 3);
        
        if (ImgUtil::DEBUG) {
            VSCOME[i][j] = 1./3;
            string fileName = "CORE";
            ImgUtil::writeGrayScaleImage(VSCOME, orH, orW, fileName, 255 * 3);
        }
        
        float **VSCOMEprint = MemUtil::create2DMatrix<float>(orH,orW);
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                VSCOMEprint[i][j] = VSCOME[i][j] < 0 ? 0 : VSCOME[i][j];
            }
        }
        
        fileName = "VSCOMEprint";
        ImgUtil::writeGrayScaleImage(VSCOMEprint, orH, orW, fileName, 255 * 3);
        
        MemUtil::freeMatrix(&VSCOMEprint);
    }
}
    

void SymVarReferencePointLe::computeReferenceOrientation(int nOrientations, int initialLength, int width, int maxLength) {
    
    
    int L = initialLength;
    int W = width;
      
    vector<int> candidates;   
    float *var = MemUtil::createVector<float>(nOrientations);
    
    do {
        candidates.clear();
        float min = 1000;
        int minIndex = -1;
        for (int k = 0; k < nOrientations; k++) {
            float thetaK = (float)k * 2. * PI / (float) nOrientations;

            float ang = thetaK + 0.5 * PI; // el +PI/2 es para que se inici el 0 donde siempre

            int M = 0;
            for (int w = 0; w < W; w++) {
                for (int l = 0; l < L; l++) {

                    int u = (int) (coreX + ((float)l) * cos(ang) + (w - (float)W / 2.) * sin(ang));
                    int v = (int) (coreY + ((float)l) * sin(ang) + ((float)W / 2. - w) * cos(ang));

                    if ((u >= 0) && (v >= 0) && (u < orH) && (v < orW)) {
                        VSCOME[u][v] = k * 255. / (float) nOrientations;
                        var[k] += abs(sin( O[u][v] - thetaK ));
                        M++;
                    }
                }
            }
            if (M != 0) 
                var[k] /= (float) M;

            if (var[k] < min) {
                min = var[k];
                minIndex = k;
            }
        }
        
        bool insert = false;
        int localMin = 0;
        for (int k = 0; k < nOrientations; k++) {
            if ( (var[k] < var[minIndex] + 0.1) ) { // así entra también minIndex
                insert = true;
                if (var[k] < var[localMin])
                    localMin = k;
                // si k - 1 está dentro
            }
            else if (insert) {
                insert = false;
                candidates.push_back(localMin);
                localMin = k;
            }
        }
        if (insert)
            candidates.push_back(localMin);
        /* Si hay candidatas seguidas, quedarse con la mínima */
        L++;
        if (ImgUtil::DEBUG)
            cout << "hay " << candidates.size() << " candidates " << endl;
    } while ( (candidates.size() > 2) && (L <= maxLength));
    
    int mean = 0;
    for (int i = 0; i < candidates.size(); i++) {
        if (ImgUtil::DEBUG)
            cout << "candidate = " << candidates[i] << endl;
        mean += candidates[i];
    }
    mean /= candidates.size();
    
    int dif1 = 0;
    int dif2 = 0;
    int mean2 = (mean + nOrientations / 2) % nOrientations;
    
    for (int i = 0; i < candidates.size(); i++) {
        dif1 += min(abs(mean - candidates[i]), nOrientations - abs(mean - candidates[i]));
        dif2 += min(abs(mean2 - candidates[i]), nOrientations - abs(mean2 - candidates[i]));
    }
    if (dif2 < dif1)
        mean = mean2;
    
    if (ImgUtil::DEBUG)
        cout << "reference direction = " << mean << " en angulo = " << (float)mean * 2. * PI / (float) nOrientations * 180. / PI << endl;
    
    MemUtil::freeMatrix(&var);
    
    referenceDirection = mean; // to obtain the angle in radians: (float)mean * 2. * PI / (float) nOrientations
}