/* 
 * \file        SymVarReferencePointLe.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for SymVarReferencePointLe (Le2012) class
 *
 * 
 */


#ifndef SYMVARREFERENCEPOINTLE_H
#define	SYMVARREFERENCEPOINTLE_H

#include "PCASYSManager.h"

using namespace std;

/**
 * \class SymVarReferencePointLe
 *
 * The SymVarReferencePointLe class implements Fingerprint reference point detection 
 * based on symmetry and variation algorithm from Le PR 2012
 */
class SymVarReferencePointLe {
public:
    /** Default constructur */
    SymVarReferencePointLe();
    
    /**
     * Constructs an instance of the class with an orientation map to be used
     * for the extraction of the convex core point
     * \param Omap Orientation map of the fingerprint
     * \param orH Height of the orientation map
     * \param orW Width of the orientation map
     */
    SymVarReferencePointLe(float **Omap, int orH, int orW);
    
    /** Default destructor */
    ~SymVarReferencePointLe();

    /**
     * Executes the variance reference point method from the fingerprint in file
     * \param file fingerprint
     * \param segmentar indicates if Poincare's segmentation must be done
     * \return fingerprint feature vector
     */
    float *run(string file, PCASYSManager *pcasys, bool segmentar = false);
    
    /**
     * It computes the convex core point (reference point) in the fingerprint
     * orientation map
     */
    void computeReferencePoint();

    /**
     * It computes the symmetry values based on complex filters of second order
     * \param Ws2 Window size of the complex filter
     * \param sigmaC Sigma value for the gaussian in the convolution
     */
    void computeSymmetryValues(int Ws2, float sigmaC);
    
    /**
     * It computes the VORIV feature described in the paper (similar to poincare)
     * \param N Length of the line segment considered
     * \param K Size of the neighborhood considered
     */
    void computeVORIVfeature(int N, int K);
    
    /**
     * It computes the VSCOME of the orientation map, which consists of the mean
     * of the VORIV feature and the Symmetry values
     */
    void computeVSCOME();
    
    /**
     * It computes the direction of the reference point. It implements the algorithm
     * in the paper: Fingerprint Reference-Point Detection, EURASIP Journal on Applied 
     * Signal Processing 2005:4, 498–509
     * \param nOrientations Number of orientations (directions ) to be considered
     * \param initialLength Initial length of the window for the search
     * \param width Width of the window for the search
     * \param maxLength Maximum lenght of the window for the search
     */
    void computeReferenceOrientation(int nOrientations, int initialLength, int width, int maxLength);
    
    /** It returns the core point position Row
     * \return the core point position Row
     */
    int getCoreX() const;
    
    /** It returns the core point position Column
     * \return the core point position Column
     */
    int getCoreY() const;
    
    /** It returns the reference direction
     * \return the reference direction
     */
    int getReferenceDirection() const;
    
    /**
     * It returns the feature vector's size
     * \return feature vector's size
     */
    int getDimFV() const;
    
private:

    float **O;                  //!< Orientation map of the fingerprint to find the reference point
    int orH;                    //!< Height of the orientation map
    int orW;                    //!< Width of the orientation map
                        
    float **sym2;               //!< Symmetry values responses from the complex filter
    float **VORIV;              //!< VORIV feature values
    float **VSCOME;             //!< VSCOME values, that is, the combination of sym2 and VORIV
    
    int coreX;                  //!< Core point position Rows
    int coreY;                  //!< Core point position Columns
    
    int referenceDirection;     //!< Direction of the reference point
    
    int dimFV;                          //!< Dimension of feature vector
};

inline int SymVarReferencePointLe::getDimFV() const { return dimFV;}

inline int SymVarReferencePointLe::getCoreX() const{ return coreX;}

inline int SymVarReferencePointLe::getCoreY() const{ return coreY;}

inline int SymVarReferencePointLe::getReferenceDirection() const{ return referenceDirection;}

#endif	/* SYMVARREFERENCEPOINTLE_H */

