/* 
 * File:   WangCoreDeltaDetector.cpp
 * Author: clopez
 * 
 * Created on 25 de junio de 2012, 11:54
 */

 /* 
 * \file        WangCoreDeltaDetector.cpp
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of WangCoreDeltaDetector class
 *
 * 
 */

#include "WangCoreDeltaDetector.h"
#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "ParametersWangCoreDeltaDetector.h"

WangCoreDeltaDetector::WangCoreDeltaDetector(){}
WangCoreDeltaDetector::WangCoreDeltaDetector(float **coh, float **ori, int hei, int wid, int bSize,bool overlapping){
    
    this->h=hei;
    this->w=wid;
    this->orH=(hei/bSize);
    this->orW=(wid/bSize);
    
    this->blockSize=bSize;
    
    if (overlapping){
        this->orH-=1;
        this->orW-=1;
    }
    
    this->coherence=MemUtil::create2DMatrix<float>(hei,wid);
    this->oriImage=MemUtil::create2DMatrix<float>(this->orH,this->orW);
    
    for (int i=0;i<hei;i++){
        for (int j=0;j<wid; j++){
            this->coherence[i][j]=coh[i][j];
        }
    }

    for (int i=0;i<this->orH;i++){
        for (int j=0;j<this->orW; j++){
            this->oriImage[i][j]=ori[i][j];
        }
    }
            
    this->existCoreDelta=false;
    this->bwiseSegmentation=0;
    this->pixelSegmentation=0;
}

WangCoreDeltaDetector::~WangCoreDeltaDetector(){
    if (this->coherence>0){MemUtil::freeMatrix(&coherence);}
    if (this->oriImage>0){MemUtil::freeMatrix(&oriImage);}
    if (this->Pplus>0){MemUtil::freeMatrix(&Pplus);}
    if (this->recNeigh>0){MemUtil::freeMatrix(&recNeigh);}
    if (this->recNeigh>0){MemUtil::freeMatrix(&recNeigh);}
    if (this->bwiseSegmentation>0){MemUtil::freeMatrix(&bwiseSegmentation);}
}

//Basic functions
void WangCoreDeltaDetector::findCoreDeltas(){

    //cout << "one"<<endl;
    computePplus(PplusWindow_WangCDDetector);
    
    int centerI,centerJ;
    //cout << "two"<<endl;
    findPMaxValue(&centerI,&centerJ);
    
    if (centerI>=0){
        centerI=(centerI+0.5)*this->blockSize;
        centerJ=(centerJ+0.5)*this->blockSize;
        findLocalMaxima(centerI,centerJ);
        return;
    }
    else{
        this->existCoreDelta=false;
        this->coredeltaI=-1;
        this->coredeltaJ=-1;
    }
    
    
    

}

bool WangCoreDeltaDetector::foundCoreDelta(){
    return this->existCoreDelta;
}

int *WangCoreDeltaDetector::getCoreDeltaPosition(){
    int *pos=MemUtil::createVector<int>(2);
    pos[0]=this->coredeltaI;
    pos[1]=this->coredeltaJ;
    return pos;
}

void WangCoreDeltaDetector::setKaruRulesForSPCleaning(bool flag){
    this->karuRules=flag;
}

void WangCoreDeltaDetector::setPixelSegmentation(bool** segm){
    this->pixelSegmentation=MemUtil::copy2DMatrix<bool>(segm, h, w);
}

void WangCoreDeltaDetector::setBlockwiseSegmentation(bool** segm){
    this->bwiseSegmentation=MemUtil::copy2DMatrix<bool>(segm, orH, orW);
}

//Auxiliars
void WangCoreDeltaDetector::computePplus(int curveW){

    
    
   float sumDelta;
    float delta, diff, absdiff;

    //cout << "computing pplus with curveW = " << curveW << ", noSpmargin="<< noSPmargin_WangCoreDeltaDetector * ImgUtil::fAdjust<<endl;

    //Image where the poincare index is stored
    this->Pplus = MemUtil::create2DMatrix<float>(orH, orW);


    int numPositions = (4 * curveW) - 4;
    float *orientationValues = MemUtil::createVector<float>(numPositions + 1); //Window values
    int wHalf = (curveW - 1) / 2;

    // For usual poincare
    int wHalf2 = wHalf;
    if (curveW % 2 == 0) // For poincare with 4 blocks
        wHalf2++;

    bool endIteration = false;
    
    

    int fromI=((noSPmargin_WangCoreDeltaDetector * ImgUtil::fAdjust)/this->blockSize);
    
    //cout << "about to start" << endl;
    
    for (int i = fromI; i < orH - ((noSPmargin_WangCoreDeltaDetector * ImgUtil::fAdjust)/this->blockSize); i++) {
        for (int j = ((noSPmargin_WangCoreDeltaDetector * ImgUtil::fAdjust)/this->blockSize); j < orW - ((noSPmargin_WangCoreDeltaDetector * ImgUtil::fAdjust)/this->blockSize); j++) {
            
            
            endIteration = false;
            ////if ((blockSegmentation != 0) && (!blockSegmentation[i][j]))
            ////    endIteration = true;
            //The round starts in tht southeastern corner
            // and goes (first) north and (then) west.
            
            //cout <<"(i,j)=("<<i<<","<<j<<")."<<endl;
            
            int posCounter = 0;
            for (int k = wHalf2; k >= -wHalf; k--) {//Going up north
                orientationValues[posCounter] = oriImage[i + k][j + wHalf2];
                posCounter++;
                ////if ((blockSegmentation != 0) && (!blockSegmentation[i + k][j + wHalf2]))
                ////    endIteration = true;
            }
            for (int k = wHalf2 - 1; k >= -wHalf; k--) {//Going left to the west
                orientationValues[posCounter] = oriImage[i - wHalf][j + k];
                posCounter++;
                ////if ((blockSegmentation != 0) && (!blockSegmentation[i - wHalf][j + k]))
                ////    endIteration = true;
            }
            for (int k = -wHalf + 1; k <= wHalf2; k++) {//Going down south
                orientationValues[posCounter] = oriImage[i + k][j - wHalf];
                posCounter++;
                ////if ((blockSegmentation != 0) && (!blockSegmentation[i + k][j - wHalf]))
                ////    endIteration = true;
            }
            for (int k = -wHalf + 1; k <= wHalf2; k++) {//Going right to the east
                orientationValues[posCounter] = oriImage[i + wHalf2][j + k];
                posCounter++;
                ////if ((blockSegmentation != 0) && (!blockSegmentation[i + wHalf2][j + k]))
                ////    endIteration = true;
            }
            
          //  endIteration = false;
            if (endIteration)
                continue;

            sumDelta = 0;
            for (int k = 0; k < numPositions; k++) {
                diff = orientationValues[k + 1] - orientationValues[k];
                absdiff = (float) fabs((double) diff);
                if (diff>0){
                    if (absdiff < PI / 2.)
                        delta = diff;
                    else if (diff <= -PI / 2.)
                        delta = PI + diff;
                    else
                        delta = diff - PI;
                    sumDelta = sumDelta + delta;
                }
            }
            Pplus[i][j] = (1. / (2. * PI)) * sumDelta;
        }
    }

    MemUtil::freeMatrix(&orientationValues);

    if (this->bwiseSegmentation>0){
        for (int i=0; i<orH; i++){
            for (int j=0; j<orW; j++){
                if (this->bwiseSegmentation[i][j]==false){
                    this->Pplus[i][j]=0;
                }
            }
        }
    }
    
    if (this->karuRules){
        int margin=(karuSideMargin_WangCDDetector * ImgUtil::fAdjust)/this->blockSize;
        int topMargin=(karuUpperMargin_WangCDDetector * ImgUtil::fAdjust)/this->blockSize;
        
        for (int i=0; i<orH; i++){
            for (int j=0; j<orW; j++){
                
                if (i<topMargin){ this->Pplus[i][j]=0;}
                else if (i>(orH-margin)){this->Pplus[i][j]=0;}
                else if (j<margin){this->Pplus[i][j]=0;}
                else if (j>(orW-margin)){this->Pplus[i][j]=0;}
                
            }
        }
        
    }
    
}

void WangCoreDeltaDetector::findPMaxValue(int *posI,int *posJ){

    float maxValue=Pplus[0][0];
    (*posI)=0;
    (*posJ)=0;
    
    for (int i=0; i<this->orH; i++){
        for (int j=0; j<this->orW; j++){
            if (Pplus[i][j]>maxValue){
                maxValue=Pplus[i][j];
                (*posI)=i;
                (*posJ)=j;
            }
        }
    }
    
    if (maxValue<minPPlusValue_WangCDDetector){
        (*posI)=-1;
        (*posJ)=-1;
    }
        
    
}

void WangCoreDeltaDetector::findLocalMaxima(int centerI, int centerJ){
    
    int candMinI=centerI-candidatesScope_WangCDDetector * ImgUtil::fAdjust;
    int candMinJ=centerJ-candidatesScope_WangCDDetector * ImgUtil::fAdjust;
    int candMaxI=centerI+candidatesScope_WangCDDetector * ImgUtil::fAdjust;
    int candMaxJ=centerJ+candidatesScope_WangCDDetector * ImgUtil::fAdjust;
    
    
    float minCoherence=1000;
    this->coredeltaI=-1;
    this->coredeltaJ=-1;
    this->existCoreDelta=false;
    
    //cout<<"here"<<endl;
    
    for (int i=candMinI;i<=candMaxI; i++){
        
        //cout<<"row "<<i<<endl;
        
        for (int j=candMinJ;j<=candMaxJ; j++){
            
            bool localMin=true;
            
            
            for (int varI=-localMinOutRad_WangCDDetector * ImgUtil::fAdjust; varI<=localMinOutRad_WangCDDetector * ImgUtil::fAdjust; varI++){
                for (int varJ=-localMinOutRad_WangCDDetector * ImgUtil::fAdjust; varJ<=localMinOutRad_WangCDDetector * ImgUtil::fAdjust; varJ++){
                    
                    float d=sqrt(varI*varI+varJ*varJ);
                    
                    if ( (d<localMinOutRad_WangCDDetector * ImgUtil::fAdjust) && ((localMinInRad_WangCDDetector * ImgUtil::fAdjust)<d)){
                        if (this->coherence[i][j]>(this->coherence[i+varI][j+varJ]-localMinDiff_WangCDDetector)){
                            localMin=false;
                        }
                            
                    }
                        
                    
                }
            }
            
            if (localMin && (this->coherence[i][j]<minCoherence)){
                minCoherence=this->coherence[i][j];
                this->coredeltaI=i;
                this->coredeltaJ=j;
                this->existCoreDelta=true;
            }
            
            
            
        }
    } 

}
