/* 
 * \file        XuejunTanDFE.h
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Implementation file for XuejunTanFE class
 *
 * 
 */
#include "XuejunTanFE.h"
#include "ParametersXuejunTanFE.h"
#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "FingerprintImg.h"


XuejunTanFE::XuejunTanFE() {
}

XuejunTanFE::XuejunTanFE(float **image, int imageH, int imageW) {
    
    int iWX = ((int)(imageWidth_XuejunTan * ImgUtil::fAdjust) % 2 != 0) ? (int)(imageWidth_XuejunTan * ImgUtil::fAdjust - 1) : (int)(imageWidth_XuejunTan * ImgUtil::fAdjust);
    int iHX = ((int)(imageHeight_XuejunTan * ImgUtil::fAdjust) % 2 != 0) ? (int)(imageHeight_XuejunTan * ImgUtil::fAdjust - 1) : (int)(imageHeight_XuejunTan * ImgUtil::fAdjust);
    
    //Copying the image
    this->img=MemUtil::create2DMatrix<float>(iHX,iWX);  
    //Calculating the trimming (if necessary)
    int offsetI,offsetJ;
    
    if (imageH >= iHX){ 
        offsetI = round((imageH-iHX)/2); 
    }
    else{ 
        cout<< "XuejunTanFE::XuejunTanFE: Warning- The fingerprint image is too small, it must contain at least " << iHX << "x" << iWX << " pixels."<<endl;
    }
    
    if (imageW >= iWX){ 
        offsetJ = round((imageW-iWX)/2); 
    }
    else{ 
        cout<< "XuejunTanFE::XuejunTanFE: Warning- The fingerprint image is too small, it must contain at least " << iHX << "x" << iWX << " pixels."<<endl;

    }

    //Filling the image in
    for (int i=0; i<iHX; i++){
        for (int j=0; j<iWX; j++){
            this->img[i][j]=image[offsetI+i][offsetJ+j];
        }
    }
    
    if (ImgUtil::DEBUG){
        string filename="FPrint_trimmedXTFE";
        ImgUtil::writeGrayScaleImage(this->img,iHX,iWX,filename,1);
    }
    
    this->h=iHX;
    this->w=iWX;
    
    this->orH = h / (int)(blockSize_XuejunTan * ImgUtil::fAdjust);
    this->orW = w/ (int)(blockSize_XuejunTan * ImgUtil::fAdjust);
    
    //Initializing the remaining vars
    this->blockSize=-1;
    this->features=0;
    this->oriImage=0;
    this->xCos=0;
    this->ySin=0;
}

XuejunTanFE::~XuejunTanFE() {
    MemUtil::freeMatrix(&img);
    MemUtil::freeMatrix(&oriImage);
    MemUtil::freeMatrix(&xCos);
    MemUtil::freeMatrix(&ySin);
    MemUtil::freeMatrix(&features);
    //cout << "XuejunTanFE: Warning, features NOT freed."<<endl;
    
}

float *XuejunTanFE::run(string file, PCASYSManager *pcasys){
    
    
    /*
    * Reading
    */

    FingerprintImg fi;
    
    if (ImgUtil::fAdjust != 1){

        fi.setPCASYSManager(pcasys); 
        fi.readFingerprintImage(file);      
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
        
		
    /*
    * Computing
    */
		 
    initializeValuesFromImage(fi.getSegmentedImageFloat(),fi.getH(),fi.getW());
    computeFeatures();

    
    /*
    * Preparing data to be returned
    */
    dimFV = orH * orW * 16;
    float *returnedFeats=MemUtil::createVector<float>(dimFV);

    int cont = 0;
    
    for (int i=0; i<orH; i++){
        for (int j=0; j<orW; j++){
            for (int k=0; k<16; k++){
                    returnedFeats[cont++]=this->features[i][j][k];
            }
        }
    }

    return returnedFeats;//New copy in memory
}

void XuejunTanFE::computeFeatures(){
    
    //Initializing basic variables
    initializeFeatures();
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Features initialized."<<endl;}
    
    //Primitive Feature 1: Orientation
    computeOrientations(blockSize_XuejunTan * ImgUtil::fAdjust);
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Feature 1 completed."<<endl;}

    //Primitives 2,4,6,8,10: Mean, std, min, max and median (size 3)
    
    calculateNeighbourhoodInformation(3,1,2);
    
    //Primitives 3,5,7,9,11: Mean, std, min, max and median (size 5)
    
    calculateNeighbourhoodInformation(5,2,2);
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Features 2-11 completed."<<endl;}
    
    //Primitives 12-13: Sobel gradient components on the orientation image (first vert, then hor)
    computeOrientationEdges();
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Features 12-13 completed."<<endl;}
    
    //Primitive 14: Binary orientation
    computeBinaryOrientations();
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Feature 14 completed."<<endl;}
    
    //Primitive 15: Binary orientation
    computeSineAndCosine();
    if (ImgUtil::DEBUG){cout <<"XuejunTanFE::computeFeatures: Feature 15 completed."<<endl;}
    
    return;
    
}

float ***XuejunTanFE::getFeatures(){
    
    return this->features;
    
}

int XuejunTanFE::getFeaturesHeight(){return this->orH;}
int XuejunTanFE::getFeaturesWidth(){return this->orW;}


void XuejunTanFE::initializeValuesFromImage(float **image, int imageH, int imageW){
        
    int iWX = ((int)(imageWidth_XuejunTan * ImgUtil::fAdjust) % 2 != 0) ? imageWidth_XuejunTan * ImgUtil::fAdjust - 1 : imageWidth_XuejunTan * ImgUtil::fAdjust;
    int iHX = ((int)(imageHeight_XuejunTan * ImgUtil::fAdjust) % 2 != 0) ? imageHeight_XuejunTan * ImgUtil::fAdjust - 1 : imageHeight_XuejunTan * ImgUtil::fAdjust;
    
    //Copying the image
    this->img=MemUtil::create2DMatrix<float>(iHX,iWX);  
    //Calculating the trimming (if necessary)
    int offsetI,offsetJ;
    
    if (imageH >= iHX){ 
        offsetI = round((imageH-iHX)/2); 
    }
    else{ 
        cout<< "XuejunTanFE::XuejunTanFE: Warning- The fingerprint image is too small, it must contain at least " << iHX << "x" << iWX << " pixels."<<endl;
    }
    
    if (imageW >= iWX){ 
        offsetJ = round((imageW-iWX)/2); 
    }
    else{ 
        cout<< "XuejunTanFE::XuejunTanFE: Warning- The fingerprint image is too small, it must contain at least " << iHX << "x" << iWX << " pixels."<<endl;

    }

    //Filling the image in
    for (int i=0; i<iHX; i++){
        for (int j=0; j<iWX; j++){
            this->img[i][j]=image[offsetI+i][offsetJ+j];
        }
    }
    
    if (ImgUtil::DEBUG){
        string filename="FPrint_trimmedXTFE";
        ImgUtil::writeGrayScaleImage(this->img,iHX,iWX,filename,1);
    }
    
    this->h=iHX;
    this->w=iWX;
    
    this->orH = h / (int)(blockSize_XuejunTan * ImgUtil::fAdjust);
    this->orW = w/ (int)(blockSize_XuejunTan * ImgUtil::fAdjust);
    
    //Initializing the remaining vars
    this->blockSize=-1;
    this->features=0;
    this->oriImage=0;
    this->xCos=0;
    this->ySin=0;
}

/**
 * Internal functions
 */

void XuejunTanFE::initializeFeatures(){
    
    this->features=MemUtil::create3DMatrix<float>(orH,orW,16,0);
    
    
}

void XuejunTanFE::computeOrientations(int blockW) {
        
    
    this->blockSize=blockW;
    float **xCos = MemUtil::create2DMatrix<float>(orH, orW);
    float **ySin = MemUtil::create2DMatrix<float>(orH, orW);
    this->oriImage = MemUtil::create2DMatrix<float>(orH, orW);
    
    float **Vx = MemUtil::create2DMatrix<float>(orH, orW);
    float **Vy = MemUtil::create2DMatrix<float>(orH, orW);
    
    float **gradX = MemUtil::create2DMatrix<float>(h, w);
    float **gradY = MemUtil::create2DMatrix<float>(h, w);
    
    
    
    /* compute SOBEL */
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] - 
                   img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]) ;
            
            gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                    img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]) ;
            
            //gradMagnitude[i][j] = sqrt(gradX[i][j] * gradX[i][j] + gradY[i][j] * gradY[i][j]);
            
        }
    }
    
    
    float Gsx, Gsy;
    /* Compute Orientations */
    for (int i = 0; i < orH; i++) {
        for (int j = 0; j < orW; j++) {
            
            for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                    Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                    Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                    Vx[i][j] += Gsx;
                    Vy[i][j] += Gsy;
                    
                    

                }
            }
            
            oriImage[i][j]= 0.5 * atan2(Vx[i][j], Vy[i][j]);
            oriImage[i][j] += (oriImage[i][j] <= 0) ? 0.5 * PI : -0.5 * PI; 
            
            xCos[i][j] = cos(2.0 * oriImage[i][j]);
            ySin[i][j] = sin(2.0 * oriImage[i][j]);
        }
    }
    
    
    
    
    
    for (int i=0;i<orH; i++){
        for (int j=0; j<orW; j++){
            //this->features[i][j][1]=oriImage[i][j];
            //Mapping [-pi/2,pi/2]->[0,180] (note that this is counterCW to CW)
            //this->features[i][j][0]= (((-oriImage[i][j])+(PI/2))/PI)*180;
            if (oriImage[i][j] > 0){
                this->features[i][j][0]= 90+ ( ((PI/2)-oriImage[i][j])/(PI/2))*90;
            }
            else{
                this->features[i][j][0]=-((2*oriImage[i][j])/PI)*90;
            }
        }
    }
    
    
    MemUtil::freeMatrix(&Vx);
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&Vy);
    MemUtil::freeMatrix(&gradY);
    
}

void XuejunTanFE::calculateNeighbourhoodInformation(int neighSize,int firstFeatPos,int numSizes){
    
    int halfNeighSize=(int)(neighSize-1)/2;
    
    for (int i=0; i<this->orH; i++){
    
        int fromI=max(0,i-halfNeighSize);
        int toI=min(this->orH-1,i+halfNeighSize);
        
        for (int j=0; j<this->orW; j++){
        
            int fromJ=max(0,j-halfNeighSize);
            int toJ=min(this->orW-1,j+halfNeighSize);
            
            int totalPositions=(toI-fromI+1)*(toJ-fromJ+1);
            
            float* values=MemUtil::createVector<float>(totalPositions);
            
            int numPos=0;
            for (int u=fromI; u<=toI; u++){
                for (int v=fromJ; v<=toJ; v++){
                    values[numPos]=this->features[u][v][0];
                    numPos++;
                }
            }
            
            
            
            sortVector(values,totalPositions);
            
            
            //Mean (1st pos);
            float acc=sumVector(values,totalPositions);
            
            if (ImgUtil::DEBUG && (i==2) && (j==2)){
                cout << "Vector: "<<endl;
                for (int jdf=0; jdf<totalPositions; jdf++){
                    cout << values[jdf] <<"\t";
                }
                cout <<endl;
                
                cout <<"\t Positions: ["<<totalPositions<<"], acc=["<<acc<<"], mean=["<<acc/totalPositions<<"]."<< endl;
            }
            
            features[i][j][firstFeatPos]=acc/totalPositions;
            
            
            //Standard deviation (2nd pos);
            features[i][j][firstFeatPos+(1*numSizes)]=stdVector(values,features[i][j][firstFeatPos],totalPositions);
            
            
            //Min (3rd position)
            features[i][j][firstFeatPos+(2*numSizes)]=values[0];
            
            
            //Max (4th position)
            features[i][j][firstFeatPos+(3*numSizes)]=values[totalPositions-1];
                        
            //Median (5th position)
            if (totalPositions%2==1){
                features[i][j][firstFeatPos+(4*numSizes)]=values[(totalPositions-1)/(int)2];
                
            }
            else{
                features[i][j][firstFeatPos+(4*numSizes)]=(values[(totalPositions/2)-1]+values[(totalPositions/2)])/2;
                
            }
            
            MemUtil::freeMatrix(&values);
            
        }
    }
    
}

void XuejunTanFE::computeOrientationEdges(){
    
    //First/Last rows/cols are replicated (aka extended boundaries condition)
    
    //Features 12 (vertical) and 13 (horizontal)
    for (int i = 0; i < orH ; i++) {
        
        int prevI=max(0,i-1);
        int nextI=min(orH-1,i+1);
        
        for (int j = 0; j < orW ; j++) {
            
            int prevJ=max(0,j-1);
            int nextJ=min(orW-1,j+1);
            
            this->features[i][j][11]= -(features[nextI][prevJ][0] + 2.0 * features[nextI][j][0] + features[nextI][nextJ][0] -
                    features[prevI][prevJ][0] - 2.0 * features[prevI][j][0] - features[prevI][nextJ][0]);
            
            this->features[i][j][12]= (features[nextI][nextJ][0] + 2.0 * features[i][nextJ][0] + features[prevI][nextJ][0] - 
                   features[nextI][prevJ][0] - 2.0 * features[i][prevJ][0] - features[prevI][prevJ][0]) ;

        }
        
    }
}

void XuejunTanFE::computeBinaryOrientations(){
    
    //Feature 14:
    //According to the text, this must be done this way.
    //However, the images indicate that they might be calculated pixelwise and then aggregated by blocks.
    for (int i = 0; i < orH ; i++) {
        for (int j = 0; j < orW ; j++) {
            if (this->features[i][j][0]>90){
                this->features[i][j][13]=1;
            }else{
                this->features[i][j][13]=0;
            }
        }
    }
    
    //Otherwise, we can tray to aggregate by blocks:
    
 
}

void XuejunTanFE::computeSineAndCosine(){
    
    //Features 15,16: Sin and cosine components of the orientation.
    //Note that the original data is in DEGREES (CW), and must be converted to RADIANS (counterCW) for using sin/cos
    
    for (int i = 0; i < orH ; i++) {
        for (int j = 0; j < orW ; j++) {
            
            //This cannot be done this way becaouse the orientations span changes
            //float angInRads=oriImage[i][j];
            //Recalculation of the orientaion from degrees to radians
            float angInRads=(-this->features[i][j][0]/180)*PI;
            
            this->features[i][j][14]=sin(angInRads);
            this->features[i][j][15]=cos(angInRads);

            
        }
    }
}


/**
 * Auxiliar functions
 */

void XuejunTanFE::sortVector(float *originalVector,int numElements){

    for (int i=0; i<numElements;i++){
        for (int j=i; j<numElements; j++){
            
            if (originalVector[i]>originalVector[j]){
                float aux =originalVector[i];
                originalVector[i]=originalVector[j];
                originalVector[j]=aux;
            }

        }
    }


}

float XuejunTanFE::sumVector(float *originalVector,int numElements){
    float acc=0;
    for (int i=0;i<numElements;i++){
        acc+=originalVector[i];
    }
    return acc;
}

float XuejunTanFE::stdVector(float *originalVector,float mean, int numElements){
    
    float acc=0;
    for (int i=0;i<numElements;i++){
        acc+= (originalVector[i]-mean)*(originalVector[i]-mean);
    }
    return sqrt(acc/numElements);
    
}

/**
 * Representation functions
 */

float **XuejunTanFE::normalizedFeaturesMosaic(int *sizeI, int *sizeJ){
    
    
    //Calculate dimensions and declare the mosaic
    
    *sizeI= (4*orH)+(3*mosaicMargin_XuejunTan);
    *sizeJ= (4*orW)+(3*mosaicMargin_XuejunTan);
    
    float **mosaic=MemUtil::create2DMatrix<float>(*sizeI,*sizeJ,0);
    
    //Set normalization factors
    float *normFactorsMax=new float[16];
    float *normFactorsMin=new float[16];
    
    //->First 11 features
    for (int i=0; i<11; i++){ 
        normFactorsMin[i]=0;
        normFactorsMax[i]=180;
    }
    
    //->Features 12 and 13 (sobel)

    for (int feat=11; feat<13; feat++){
        
        normFactorsMin[feat]=this->features[0][0][feat];
        normFactorsMax[feat]=this->features[0][0][feat];
        
        for (int i = 0; i < this->orH; i++){
            for (int j = 0; j < this->orW; j++){
                //Feature checking
                if (this->features[i][j][feat] > normFactorsMax[feat]){ 
                    normFactorsMax[feat]= this->features[i][j][feat];
                }
                if (this->features[i][j][feat] < normFactorsMin[feat]){ 
                    normFactorsMin[feat]= this->features[i][j][feat];
                }   

            }
        }
    }
    
    //-> binary image
    normFactorsMin[13]=0;
    normFactorsMax[13]=1;
    
    //-> sine/cosine
    normFactorsMin[14]=-1;
    normFactorsMax[14]=0;
    normFactorsMin[15]=-1;
    normFactorsMax[15]=1;
    
    //Fill the mosaic
    
    for (int feat=0; feat<16; feat++){
        
        //Read the maximum and the minimum
        float featMin=normFactorsMin[feat];
        float featMax=normFactorsMax[feat];
        
        
        
        //calculate the startI and startJ position
        int startI= ceil(feat/4) * (orH+mosaicMargin_XuejunTan);
        int startJ= (feat%4) * (orW+mosaicMargin_XuejunTan);
        
        //fill the mosaic in
        
        for (int i = 0; i < this->orH; i++){
            for (int j = 0; j < this->orW; j++){
                
                //The value is normalized
                mosaic[startI+i][startJ+j]= (features[i][j][feat]-featMin)/(featMax-featMin);
                
            }
        }
    }
    
    return mosaic;
    
}











