/* 
 * \file        XuejunTanDFE.h
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for XuejunTanFE class
 *
 * 
 */

#ifndef XUEJUNTANFE_H
#define	XUEJUNTANFE_H

#include "PCASYSManager.h"

using namespace std;

/**
 * @class XuejunTanFE
 *
 * The XuejunTanFE class implements the feature extraction technique
 *      used in
 *  
 * Xuejun Tan, Bir Bhanu, Yingqiang Lin
 * "Fingerprint Classification based on Learned Features" 
 * IEEE Trans. on Systems, man and cybernetics- Part C
 * pp. 287-300, August 2005
 * 
 * 
 */
class XuejunTanFE {
public:
    
    /**
     * Default constructor
     */
    XuejunTanFE();
    
    /**
     * Basic constructor. Note that, if the original image is larger than 384x416, it will be 
     *          trimmed downto such dimensions.
     * @param image     Grayscale image to be used
     * @param imageH    Height of the image
     * @param imageW    Width of the image
     */
    XuejunTanFE(float **image, int imageH, int imageW);
    
    /**
     * Destroyer
     */
    virtual ~XuejunTanFE();
    
    /**
     * Executes the Xuejun Tan FE method from the fingerprint in file
     * \param file fingerprint
     * \param pcasys pcasys manager
     * \return feature vector
     */
    float *run(string file, PCASYSManager *pcasys);
	
	/**
     * Assess initial values to the class when it's called from run() method
     * @param image     Grayscale image to be used
     * @param imageH    Height of the image
     * @param imageW    Width of the image
     */
    void initializeValuesFromImage(float **image, int imageH, int imageW);
	
    /**
     * Compute all the features according to the rules provided in the paper
     */
    void computeFeatures();
    
    /**
     * Retrieve the feature matrix (comes in a 3D matrix with the feature number acting as third dimension)
     * @return The features 3D matrix
     */
    float ***getFeatures();
    
    /**
     * Height of the features matrix
     * @return The number of rows (first dim) of the feature matrix
     */
    int getFeaturesHeight();
    
    /**
     * Width of the features matrix
     * @return The number of cols (second dim) of the feature matrix
     */
    int getFeaturesWidth();
    
    //
    //Internal functions
    //
    
    /**
     * Initializes the features matrix
     */
    void initializeFeatures();
    
    /**
     * Performs the basic orientations computation and stores the first feature (orientation in degrees)
     * @param blockW        Side of the blocks used for orientation aggregation
     */
    void computeOrientations(int blockW);
    
    /**
     * Calculate the basic features from orientation neighbourhoods (mean, std, min, max, median)
     * @param neighSize     Size of the considered neighbourhood
     * @param firstFeatPos  Position of the feature matrix where the first feature is to be stored
     * @param numSizes      Number of sizes to be considered in the problem. The features will be stored at positions (firstFeatPos, firstFeatPos+numSizes, firstFeatPos+2*numSizes, etc.)
     */
    void calculateNeighbourhoodInformation(int neighSize,int firstFeatPos,int numSizes);
    
    /**
     * Convolve the orientations map with horizontal and vertical Sobel masks. Also store the results in the corresponding feature matrix positions (12th and 13th)
     */
    void computeOrientationEdges();
    
    /**
     * Binarize the orientation map and store its result in the 14th position of the feature matrix.
     */
    void computeBinaryOrientations();
    
    /**
     * Compute the sin and cosine of the orientation at each position. This information takes up positions 15th and 16th of the feature matrix.
     * @param originalVector
     * @param numElements
     */
    void computeSineAndCosine();
    
    //Auxiliar functions
    /**
     * Sorts a vector of floats
     * @param originalVector    Original vector (modified in the call!) 
     * @param numElements       Number of elements contained in the vector.
     */
    void sortVector(float *originalVector,int numElements);
    
    /**
     * Sums the elements of a vector 
     * @param originalVector    Original vector
     * @param numElements       Number of elements it contains
     * @return                  The sum of the elements
     */
    float sumVector(float *originalVector,int numElements);
    
    /**
     * Calculates the standard deviation of the elements in a vector
     * @param originalVector    Original vector
     * @param mean              Mean of the elements in the vector
     * @param numElements       Number of elements it contains
     * @return                  The standard deviation
     */
    float stdVector(float *originalVector,float mean, int numElements);
    
    //
    //Information/Debugging retrieval functions
    //
    
    /**
     * Retrieve a mosaic where all the features are displayed as in the paper (normalized & tiled in 4x4 configuration)
     * @param sizeI     Height of the mosaic (in pixels)
     * @param sizeJ     Width of the mosaic (in pixels)
     * @return          The 2D mosaic
     */
    float **normalizedFeaturesMosaic(int *sizeI, int *sizeJ);
    
	/**
     * Returns a 1D array with the features, which length is stored in length
     * @param length    Length of the feature vector
     * @return          The feature vector
     */
    float *getFeatureStream(int *length);
    
    /**
     * It returns the feature vector's dimension
     * \return feature vector's dimension
     */
    int getDimFV() const;
    
private:
    
    float **img;                //!< Fingerprint image
    int w;                      //!< Width of the image
    int h;                      //!< Height of the image
    
    float ***features;
    
    float **oriImage;           //!< Orientation map
    int orW;                    //!< Width of the orientation map
    int orH;                    //!< Height of the orientation map
    int blockSize;              //!< Size of the blocks used for the computation of the orientations map
    
    float **xCos;               //!< cos(2 O) value of the cosine of the orientation (x-direction)
    float **ySin;               //!< sin(2 O) value of the sine of the orientation (y-direction)
  
    int dimFV;                  //!< Length of the feature vector
};

inline int XuejunTanFE::getDimFV() const { return dimFV;}

#endif	/* XUEJUNTANFE_H */

